@@ -17,14 +17,17 @@ from qibo.backends import (
     construct_backend,
     get_backend,
     get_device,
-    get_precision,
+    get_dtype,
     get_threads,
+    get_transpiler,
+    get_transpiler_name,
     list_available_backends,
     matrices,
     set_backend,
     set_device,
-    set_precision,
+    set_dtype,
     set_threads,
+    set_transpiler,
 )
 from qibo.config import (
     get_batch_size,
