@@ -164,6 +164,20 @@ def RY(symplectic_matrix, q, nqubits, theta):
         return RY_3pi_2(symplectic_matrix, q, nqubits)
 
 
+def GPI2(symplectic_matrix, q, nqubits, phi):
+    if phi % (2 * np.pi) == 0:
+        return RX(symplectic_matrix, q, nqubits, np.pi / 2)
+
+    if (phi / np.pi - 1) % 2 == 0:
+        return RX(symplectic_matrix, q, nqubits, -np.pi / 2)
+
+    if (phi / (np.pi / 2) - 1) % 4 == 0:
+        return RY(symplectic_matrix, q, nqubits, np.pi / 2)
+
+    # theta == 3*pi/2 + 2*n*pi
+    return RY(symplectic_matrix, q, nqubits, -np.pi / 2)
+
+
 def SWAP(symplectic_matrix, control_q, target_q, nqubits):
     """Decomposition --> CNOT-CNOT-CNOT"""
     r, x, z = _get_rxz(symplectic_matrix, nqubits)
@@ -388,10 +402,12 @@ def _rowsum(symplectic_matrix, h, i, nqubits, determined=False):
 def _determined_outcome(state, q, nqubits):
     state[-1, :] = 0
     idx = (state[:nqubits, q].nonzero()[0] + nqubits).astype(np.uint)
+    if len(idx) == 0:
+        return state, state[-1, -1]
     state = _pack_for_measurements(state, nqubits)
     state = _rowsum(
         state,
-        2 * nqubits * np.ones(idx.shape, dtype=np.uint),
+        _dim_xz(nqubits) * np.ones(idx.shape, dtype=np.uint),
         idx,
         _packed_size(nqubits),
         True,
@@ -411,7 +427,7 @@ def _random_outcome(state, p, q, nqubits):
         state = _rowsum(
             state,
             h.astype(np.uint),
-            p * np.ones(h.shape[0], dtype=np.uint),
+            np.uint(p) * np.ones(h.shape[0], dtype=np.uint),
             _packed_size(nqubits),
             False,
         )
@@ -427,7 +443,14 @@ def _random_outcome(state, p, q, nqubits):
 @cache
 def _dim(nqubits):
     """Returns the dimension of the symplectic matrix for a given number of qubits."""
-    return 2 * nqubits + 1
+    return _dim_xz(nqubits) + 1
+
+
+@cache
+def _dim_xz(nqubits):
+    """Returns the dimension of the symplectic matrix (only the de/stabilizers generators part,
+    without the phases and scratch row) for a given number of qubits."""
+    return 2 * nqubits
 
 
 @cache
@@ -440,8 +463,8 @@ def _packbits(array, axis):
     return np.packbits(array, axis=axis)
 
 
-def _unpackbits(array, axis):
-    return np.unpackbits(array, axis=axis)
+def _unpackbits(array, axis, count):
+    return np.unpackbits(array, axis=axis, count=count)
 
 
 def _pack_for_measurements(state, nqubits):
@@ -452,23 +475,16 @@ def _pack_for_measurements(state, nqubits):
     return np.hstack((x, z, r[:, None]))
 
 
-@cache
-def _pad_size(n):
-    """Returns the size of the pad added to an array of original dimension `n` after unpacking."""
-    return 8 - (n % 8)
-
-
 def _unpack_for_measurements(state, nqubits):
     """Unpacks the symplectc matrix that was packed for measurements."""
-    xz = _unpackbits(state[:, :-1], axis=1)
-    padding_size = _pad_size(nqubits)
-    x, z = xz[:, :nqubits], xz[:, nqubits + padding_size : -padding_size]
+    xz = _unpackbits(state[:, :-1], axis=1, count=_dim_xz(nqubits))
+    x, z = xz[:, :nqubits], xz[:, nqubits:]
     return np.hstack((x, z, state[:, -1][:, None]))
 
 
 def _init_state_for_measurements(state, nqubits, collapse):
     if collapse:
-        return _unpackbits(state, axis=0)[: _dim(nqubits)]
+        return _unpackbits(state, axis=0, count=_dim_xz(nqubits))[: _dim(nqubits)]
     else:
         return state.copy()
 
@@ -491,14 +507,14 @@ def M(state, qubits, nqubits, collapse=False):
     return sample
 
 
-def cast(x, dtype=None, copy=False):
+def cast(x, dtype=None, copy: bool = False):
     if dtype is None:
         dtype = "complex128"
     if isinstance(x, np.ndarray):
         return x.astype(dtype, copy=copy)
     elif sparse.issparse(x):  # pragma: no cover
         return x.astype(dtype, copy=copy)
-    return np.array(x, dtype=dtype, copy=copy)
+    return np.asarray(x, dtype=dtype, copy=copy if copy else None)
 
 
 def _clifford_pre_execution_reshape(state):
@@ -523,7 +539,7 @@ def _clifford_post_execution_reshape(state, nqubits: int):
     Returns:
         (np.array) The unpacked and reshaped state.
     """
-    state = _unpackbits(state, axis=0)[: _dim(nqubits)]
+    state = _unpackbits(state, axis=0, count=_dim(nqubits))[: _dim(nqubits)]
     return state
 
 
