@@ -1,4 +1,5 @@
 import abc
+from typing import Optional, Union
 
 from qibo.config import raise_error
 
@@ -9,7 +10,6 @@ class Backend(abc.ABC):
         self.name = "backend"
         self.platform = None
 
-        self.precision = "double"
         self.dtype = "complex128"
         self.matrices = None
 
@@ -25,20 +25,66 @@ class Backend(abc.ABC):
     def __repr__(self):
         if self.platform is None:
             return self.name
-        else:
-            return f"{self.name} ({self.platform})"
+
+        return f"{self.name} ({self.platform})"
+
+    @property
+    @abc.abstractmethod
+    def qubits(self) -> Optional[list[Union[int, str]]]:  # pragma: no cover
+        """Return the qubit names of the backend. If :class:`SimulationBackend`, return None."""
+        raise_error(NotImplementedError)
+
+    @property
+    @abc.abstractmethod
+    def connectivity(
+        self,
+    ) -> Optional[list[tuple[Union[int, str], Union[int, str]]]]:  # pragma: no cover
+        """Return the available qubit pairs of the backend. If :class:`SimulationBackend`, return None."""
+        raise_error(NotImplementedError)
+
+    @property
+    @abc.abstractmethod
+    def natives(self) -> Optional[list[str]]:  # pragma: no cover
+        """Return the native gates of the backend. If :class:`SimulationBackend`, return None."""
+        raise_error(NotImplementedError)
 
     @abc.abstractmethod
-    def set_precision(self, precision):  # pragma: no cover
-        """Set complex number precision.
+    def set_dtype(self, dtype: str):  # pragma: no cover
+        """Set data type of arrays created using the backend.
+
+        .. note::
+            The data types ``float32`` and ``float64`` are intended to be used when the circuits
+            to be simulated only contain gates with real-valued matrix representations.
+            Using one of the aforementioned data types with circuits that contain complex-valued
+            matrices will raise a casting error.
+
+        .. note::
+            List of gates that always admit a real-valued matrix representation:
+            :class:`qibo.gates.I`, :class:`qibo.gates.X`, :class:`qibo.gates.Z`,
+            :class:`qibo.gates.H`, :class:`qibo.gates.Align`, :class:`qibo.gates.RY`,
+            :class:`qibo.gates.CNOT`, :class:`qibo.gates.CZ`, :class:`qibo.gates.CRY`,
+            :class:`qibo.gates.SWAP`, :class:`qibo.gates.FSWAP`, :class:`qibo.gates.GIVENS`,
+            :class:`qibo.gates.RBS`, :class:`qibo.gates.TOFFOLI`, :class:`qibo.gates.CCZ`,
+            and :class:`qibo.gates.FanOut`.
+
+        .. note::
+            The following parametrized gates can have real-valued matrix representations
+            depending on the values of their parameters:
+            :class:`qibo.gates.RX`, :class:`qibo.gates.RZ`, :class:`qibo.gates.U1`,
+            :class:`qibo.gates.U2`, :class:`qibo.gates.U3`, :class:`qibo.gates.CRX`,
+            :class:`qibo.gates.CRZ`, :class:`qibo.gates.CU1`, :class:`qibo.gates.CU2`,
+            :class:`qibo.gates.CU3`, :class:`qibo.gates.fSim`, :class:`qibo.gates.GeneralizedfSim`,
+            :class:`qibo.gates.RXX`, :class:`qibo.gates.RYY`, :class:`qibo.gates.RZZ`,
+            :class:`qibo.gates.RZX`, and :class:`qibo.gates.GeneralizedRBS`.
 
         Args:
-            precision (str): 'single' or 'double'.
+            dtype (str): the options are the following: ``complex128``, ``complex64``,
+                ``float64``, and ``float32``.
         """
         raise_error(NotImplementedError)
 
     @abc.abstractmethod
-    def set_device(self, device):  # pragma: no cover
+    def set_device(self, device: str):  # pragma: no cover
         """Set simulation device.
 
         Args:
@@ -47,7 +93,7 @@ class Backend(abc.ABC):
         raise_error(NotImplementedError)
 
     @abc.abstractmethod
-    def set_threads(self, nthreads):  # pragma: no cover
+    def set_threads(self, nthreads: int):  # pragma: no cover
         """Set number of threads for CPU simulation.
 
         Args:
@@ -56,17 +102,22 @@ class Backend(abc.ABC):
         raise_error(NotImplementedError)
 
     @abc.abstractmethod
-    def cast(self, x, copy=False):  # pragma: no cover
+    def cast(self, x, dtype=None, copy: bool = False):  # pragma: no cover
         """Cast an object as the array type of the current backend.
 
         Args:
             x: Object to cast to array.
-            copy (bool): If ``True`` a copy of the object is created in memory.
+            dtype (str or type, optional): data type of ``x`` after casting.
+                Options are ``"complex128"``, ``"complex64"``, ``"float64"``,
+                or ``"float32"``. If ``None``, defaults to ``Backend.dtype``.
+                Defaults to ``None``.
+            copy (bool, optional): If ``True`` a copy of the object is created in memory.
+                Defaults to ``False``.
         """
         raise_error(NotImplementedError)
 
     @abc.abstractmethod
-    def issparse(self, x):  # pragma: no cover
+    def is_sparse(self, x):  # pragma: no cover
         """Determine if a given array is a sparse tensor."""
         raise_error(NotImplementedError)
 
@@ -84,18 +135,18 @@ class Backend(abc.ABC):
         raise_error(NotImplementedError)
 
     @abc.abstractmethod
-    def zero_state(self, nqubits):  # pragma: no cover
+    def zero_state(self, nqubits: int):  # pragma: no cover
         """Generate :math:`|000 \\cdots 0 \\rangle` state vector as an array."""
         raise_error(NotImplementedError)
 
     @abc.abstractmethod
-    def zero_density_matrix(self, nqubits):  # pragma: no cover
+    def zero_density_matrix(self, nqubits: int):  # pragma: no cover
         """Generate :math:`|000\\cdots0\\rangle\\langle000\\cdots0|` density matrix as an array."""
         raise_error(NotImplementedError)
 
     @abc.abstractmethod
     def identity_density_matrix(
-        self, nqubits, normalize: bool = True
+        self, nqubits: int, normalize: bool = True
     ):  # pragma: no cover
         """Generate density matrix
 
@@ -109,22 +160,22 @@ class Backend(abc.ABC):
         raise_error(NotImplementedError)
 
     @abc.abstractmethod
-    def plus_state(self, nqubits):  # pragma: no cover
+    def plus_state(self, nqubits: int):  # pragma: no cover
         """Generate :math:`|+++\\cdots+\\rangle` state vector as an array."""
         raise_error(NotImplementedError)
 
     @abc.abstractmethod
-    def plus_density_matrix(self, nqubits):  # pragma: no cover
+    def plus_density_matrix(self, nqubits: int):  # pragma: no cover
         """Generate :math:`|+++\\cdots+\\rangle\\langle+++\\cdots+|` density matrix as an array."""
         raise_error(NotImplementedError)
 
     @abc.abstractmethod
-    def matrix(self, gate):  # pragma: no cover
+    def matrix(self, gate: "qibo.gates.abstract.Gate"):  # pragma: no cover
         """Convert a :class:`qibo.gates.Gate` to the corresponding matrix."""
         raise_error(NotImplementedError)
 
     @abc.abstractmethod
-    def matrix_parametrized(self, gate):  # pragma: no cover
+    def matrix_parametrized(self, gate: "qibo.gates.abstract.Gate"):  # pragma: no cover
         """Equivalent to :meth:`qibo.backends.abstract.Backend.matrix` for parametrized gates."""
         raise_error(NotImplementedError)
 
@@ -134,71 +185,77 @@ class Backend(abc.ABC):
         raise_error(NotImplementedError)
 
     @abc.abstractmethod
-    def apply_gate(self, gate, state, nqubits):  # pragma: no cover
+    def apply_gate(self, gate, state, nqubits: int):  # pragma: no cover
         """Apply a gate to state vector."""
         raise_error(NotImplementedError)
 
     @abc.abstractmethod
-    def apply_gate_density_matrix(self, gate, state, nqubits):  # pragma: no cover
+    def apply_gate_density_matrix(self, gate, state, nqubits: int):  # pragma: no cover
         """Apply a gate to density matrix."""
         raise_error(NotImplementedError)
 
     @abc.abstractmethod
-    def apply_gate_half_density_matrix(self, gate, state, nqubits):  # pragma: no cover
+    def apply_gate_half_density_matrix(
+        self, gate, state, nqubits: int
+    ):  # pragma: no cover
         """Apply a gate to one side of the density matrix."""
         raise_error(NotImplementedError)
 
     @abc.abstractmethod
-    def apply_channel(self, channel, state, nqubits):  # pragma: no cover
+    def apply_channel(self, channel, state, nqubits: int):  # pragma: no cover
         """Apply a channel to state vector."""
         raise_error(NotImplementedError)
 
     @abc.abstractmethod
-    def apply_channel_density_matrix(self, channel, state, nqubits):  # pragma: no cover
+    def apply_channel_density_matrix(
+        self, channel, state, nqubits: int
+    ):  # pragma: no cover
         """Apply a channel to density matrix."""
         raise_error(NotImplementedError)
 
     @abc.abstractmethod
     def collapse_state(
-        self, state, qubits, shot, nqubits, normalize=True
+        self, state, qubits, shot, nqubits: int, normalize: bool = True
     ):  # pragma: no cover
         """Collapse state vector according to measurement shot."""
         raise_error(NotImplementedError)
 
     @abc.abstractmethod
     def collapse_density_matrix(
-        self, state, qubits, shot, nqubits, normalize=True
+        self, state, qubits, shot, nqubits: int, normalize: bool = True
     ):  # pragma: no cover
         """Collapse density matrix according to measurement shot."""
         raise_error(NotImplementedError)
 
     @abc.abstractmethod
-    def reset_error_density_matrix(self, gate, state, nqubits):  # pragma: no cover
+    def reset_error_density_matrix(self, gate, state, nqubits: int):  # pragma: no cover
         """Apply reset error to density matrix."""
         raise_error(NotImplementedError)
 
     @abc.abstractmethod
-    def thermal_error_density_matrix(self, gate, state, nqubits):  # pragma: no cover
+    def thermal_error_density_matrix(
+        self, gate, state, nqubits: int
+    ):  # pragma: no cover
         """Apply thermal relaxation error to density matrix."""
         raise_error(NotImplementedError)
 
     @abc.abstractmethod
     def execute_circuit(
-        self, circuit, initial_state=None, nshots=None
+        self, circuit, initial_state=None, nshots: int = None
     ):  # pragma: no cover
         """Execute a :class:`qibo.models.circuit.Circuit`."""
         raise_error(NotImplementedError)
 
     @abc.abstractmethod
     def execute_circuits(
-        self, circuits, initial_states=None, nshots=None
+        self, circuits, initial_states=None, nshots: int = None
     ):  # pragma: no cover
-        """Execute multiple :class:`qibo.models.circuit.Circuit`s in parallel."""
+        """Execute multiple :class:`qibo.models.circuit.Circuit` in parallel."""
         raise_error(NotImplementedError)
 
     @abc.abstractmethod
     def execute_circuit_repeated(
-        self, circuit, nshots, initial_state=None
+        self, circuit: "qibo.models.circuit.Circuit", nshots: int, initial_state=None
     ):  # pragma: no cover
         """Execute a :class:`qibo.models.circuit.Circuit` multiple times.
 
@@ -209,33 +266,43 @@ class Backend(abc.ABC):
 
     @abc.abstractmethod
     def execute_distributed_circuit(
-        self, circuit, initial_state=None, nshots=None
+        self, circuit, initial_state=None, nshots: int = None
     ):  # pragma: no cover
         """Execute a :class:`qibo.models.circuit.Circuit` using multiple GPUs."""
         raise_error(NotImplementedError)
 
     @abc.abstractmethod
     def calculate_symbolic(
-        self, state, nqubits, decimals=5, cutoff=1e-10, max_terms=20
+        self,
+        state,
+        nqubits: int,
+        decimals: int = 5,
+        cutoff: float = 1e-10,
+        max_terms: int = 20,
     ):  # pragma: no cover
         """Dirac representation of a state vector."""
         raise_error(NotImplementedError)
 
     @abc.abstractmethod
     def calculate_symbolic_density_matrix(
-        self, state, nqubits, decimals=5, cutoff=1e-10, max_terms=20
+        self,
+        state,
+        nqubits: int,
+        decimals: int = 5,
+        cutoff: float = 1e-10,
+        max_terms: int = 20,
     ):  # pragma: no cover
         """Dirac representation of a density matrix."""
         raise_error(NotImplementedError)
 
     @abc.abstractmethod
-    def calculate_probabilities(self, state, qubits, nqubits):  # pragma: no cover
+    def calculate_probabilities(self, state, qubits, nqubits: int):  # pragma: no cover
         """Calculate probabilities given a state vector."""
         raise_error(NotImplementedError)
 
     @abc.abstractmethod
     def calculate_probabilities_density_matrix(
-        self, state, qubits, nqubits
+        self, state, qubits, nqubits: int
     ):  # pragma: no cover
         """Calculate probabilities given a density matrix."""
         raise_error(NotImplementedError)
@@ -246,7 +313,7 @@ class Backend(abc.ABC):
         raise_error(NotImplementedError)
 
     @abc.abstractmethod
-    def sample_shots(self, probabilities, nshots):  # pragma: no cover
+    def sample_shots(self, probabilities, nshots: int):  # pragma: no cover
         """Sample measurement shots according to a probability distribution."""
         raise_error(NotImplementedError)
 
@@ -256,12 +323,12 @@ class Backend(abc.ABC):
         raise_error(NotImplementedError)
 
     @abc.abstractmethod
-    def samples_to_binary(self, samples, nqubits):  # pragma: no cover
+    def samples_to_binary(self, samples, nqubits: int):  # pragma: no cover
         """Convert samples from decimal representation to binary."""
         raise_error(NotImplementedError)
 
     @abc.abstractmethod
-    def samples_to_decimal(self, samples, nqubits):  # pragma: no cover
+    def samples_to_decimal(self, samples, nqubits: int):  # pragma: no cover
         """Convert samples from binary representation to decimal."""
         raise_error(NotImplementedError)
 
@@ -272,28 +339,20 @@ class Backend(abc.ABC):
 
     @abc.abstractmethod
     def update_frequencies(
-        self, frequencies, probabilities, nsamples
+        self, frequencies, probabilities, nsamples: int
     ):  # pragma: no cover
         raise_error(NotImplementedError)
 
     @abc.abstractmethod
-    def sample_frequencies(self, probabilities, nshots):  # pragma: no cover
+    def sample_frequencies(self, probabilities, nshots: int):  # pragma: no cover
         """Sample measurement frequencies according to a probability distribution."""
         raise_error(NotImplementedError)
 
     @abc.abstractmethod
-    def partial_trace(self, state, qubits, nqubits):  # pragma: no cover
-        """Trace out specific qubits of a state vector."""
-        raise_error(NotImplementedError)
-
-    @abc.abstractmethod
-    def partial_trace_density_matrix(self, state, qubits, nqubits):  # pragma: no cover
-        """Trace out specific qubits of a density matrix."""
-        raise_error(NotImplementedError)
-
-    @abc.abstractmethod
-    def calculate_norm(self, state, order=2):  # pragma: no cover
-        """Calculate norm of a state vector.
+    def calculate_vector_norm(
+        self, state, order: Union[int, float, str] = 2
+    ):  # pragma: no cover
+        """Calculate norm of an :math:`1`-dimensional array.
 
         For specifications on possible values of the parameter ``order``
         for the ``tensorflow`` backend, please refer to
@@ -304,9 +363,12 @@ class Backend(abc.ABC):
         raise_error(NotImplementedError)
 
     @abc.abstractmethod
-    def calculate_norm_density_matrix(self, state, order="nuc"):  # pragma: no cover
-        """Calculate norm of a density matrix. Default is the ``nuclear`` norm.
+    def calculate_matrix_norm(
+        self, state, order: Union[int, float, str] = "nuc"
+    ):  # pragma: no cover
+        """Calculate norm of a :math:`2`-dimensional array.
 
+        Default is the ``nuclear`` norm.
         If ``order="nuc"``, it returns the nuclear norm of ``state``,
         assuming ``state`` is Hermitian (also known as trace norm).
         For specifications on the other  possible values of the
@@ -328,20 +390,43 @@ class Backend(abc.ABC):
         raise_error(NotImplementedError)
 
     @abc.abstractmethod
-    def calculate_eigenvalues(self, matrix, k=6):  # pragma: no cover
+    def calculate_eigenvalues(
+        self, matrix, k: int = 6, hermitian: bool = True
+    ):  # pragma: no cover
         """Calculate eigenvalues of a matrix."""
         raise_error(NotImplementedError)
 
     @abc.abstractmethod
-    def calculate_eigenvectors(self, matrix, k=6):  # pragma: no cover
+    def calculate_eigenvectors(
+        self, matrix, k: int = 6, hermitian: bool = True
+    ):  # pragma: no cover
         """Calculate eigenvectors of a matrix."""
         raise_error(NotImplementedError)
 
     @abc.abstractmethod
-    def calculate_matrix_exp(
-        self, matrix, a, eigenvectors=None, eigenvalues=None
+    def calculate_expectation_state(
+        self, hamiltonian, state, normalize: bool
     ):  # pragma: no cover
-        """Calculate matrix exponential of a matrix.
+        """Calculate expectation value of a state vector given the observable matrix."""
+        raise_error(NotImplementedError)
+
+    @abc.abstractmethod
+    def calculate_expectation_density_matrix(
+        self, hamiltonian, state, normalize: bool
+    ):  # pragma: no cover
+        """Calculate expectation value of a density matrix given the observable matrix."""
+        raise_error(NotImplementedError)
+
+    @abc.abstractmethod
+    def calculate_matrix_exp(
+        self,
+        matrix,
+        phase: Union[float, int, complex] = 1,
+        eigenvectors=None,
+        eigenvalues=None,
+    ):  # pragma: no cover
+        """Calculate the exponential :math:`e^{\\theta \\, A}` of a matrix :math:`A`
+        and ``phase`` :math:`\\theta`.
 
         If the eigenvectors and eigenvalues are given the matrix diagonalization is
         used for exponentiation.
@@ -349,44 +434,66 @@ class Backend(abc.ABC):
         raise_error(NotImplementedError)
 
     @abc.abstractmethod
-    def calculate_expectation_state(
-        self, hamiltonian, state, normalize
+    def calculate_matrix_log(
+        self, matrix, base: Union[float, int] = 2, eigenvectors=None, eigenvalues=None
     ):  # pragma: no cover
-        """Calculate expectation value of a state vector given the observable matrix."""
+        """Calculate the logarithm :math:`\\log_{b}(A)` with a ``base`` :math:`b`
+        of a matrix :math:`A`.
+
+        If the eigenvectors and eigenvalues are given the matrix diagonalization is
+        used for exponentiation.
+        """
         raise_error(NotImplementedError)
 
     @abc.abstractmethod
-    def calculate_expectation_density_matrix(
-        self, hamiltonian, state, normalize
+    def calculate_matrix_power(
+        self, matrix, power: Union[float, int], precision_singularity: float = 1e-14
     ):  # pragma: no cover
-        """Calculate expectation value of a density matrix given the observable matrix."""
+        """Calculate the (fractional) ``power`` :math:`\\alpha` of ``matrix`` :math:`A`,
+        i.e. :math:`A^{\\alpha}`.
+
+        .. note::
+            For the ``pytorch`` backend, this method relies on a copy of the original tensor.
+            This may break the gradient flow. For the GPU backends (i.e. ``cupy`` and
+            ``cuquantum``), this method falls back to CPU whenever ``power`` is not
+            an integer.
+        """
         raise_error(NotImplementedError)
 
     @abc.abstractmethod
-    def calculate_hamiltonian_matrix_product(
-        self, matrix1, matrix2
+    def calculate_matrix_sqrt(
+        self, matrix, precision_singularity: float = 1e-14
     ):  # pragma: no cover
-        """Multiply two matrices."""
+        """Calculate the square root of ``matrix`` :math:`A`, i.e. :math:`A^{1/2}`.
+
+        .. note::
+            For the ``pytorch`` backend, this method relies on a copy of the original tensor.
+            This may break the gradient flow. For the GPU backends (i.e. ``cupy`` and
+            ``cuquantum``), this method falls back to CPU.
+        """
         raise_error(NotImplementedError)
 
     @abc.abstractmethod
-    def calculate_hamiltonian_state_product(self, matrix, state):  # pragma: no cover
-        """Multiply a matrix to a state vector or density matrix."""
+    def calculate_singular_value_decomposition(self, matrix):  # pragma: no cover
+        """Calculate the Singular Value Decomposition of ``matrix``."""
         raise_error(NotImplementedError)
 
     @abc.abstractmethod
-    def assert_allclose(self, value, target, rtol=1e-7, atol=0.0):  # pragma: no cover
+    def calculate_jacobian_matrix(
+        self, circuit, parameters, initial_state=None, return_complex: bool = True
+    ):  # pragma: no cover
+        """Calculate the Jacobian matrix of ``circuit`` with respect to varables ``params``."""
         raise_error(NotImplementedError)
 
-    def assert_circuitclose(self, circuit, target_circuit, rtol=1e-7, atol=0.0):
+    @abc.abstractmethod
+    def assert_allclose(
+        self, value, target, rtol: float = 1e-7, atol: float = 0.0
+    ):  # pragma: no cover
+        raise_error(NotImplementedError)
+
+    def assert_circuitclose(
+        self, circuit, target_circuit, rtol: float = 1e-7, atol: float = 0.0
+    ):
         value = self.execute_circuit(circuit)._state
         target = self.execute_circuit(target_circuit)._state
         self.assert_allclose(value, target, rtol=rtol, atol=atol)
-
-    @abc.abstractmethod
-    def test_regressions(self, name):  # pragma: no cover
-        """Correct outcomes for tests that involve random numbers.
-
-        The outcomes of such tests depend on the backend.
-        """
-        raise_error(NotImplementedError)
