@@ -1,9 +1,12 @@
+"""Module defining the Numpy backend."""
+
 import collections
 import math
+from typing import Union
 
 import numpy as np
 from scipy import sparse
-from scipy.linalg import block_diag
+from scipy.linalg import block_diag, fractional_matrix_power, logm
 
 from qibo import __version__
 from qibo.backends import einsum_utils
@@ -33,16 +36,30 @@ class NumpyBackend(Backend):
             np.complex128,
         )
 
-    def set_precision(self, precision):
-        if precision != self.precision:
-            if precision == "single":
-                self.precision = precision
-                self.dtype = "complex64"
-            elif precision == "double":
-                self.precision = precision
-                self.dtype = "complex128"
-            else:
-                raise_error(ValueError, f"Unknown precision {precision}.")
+    @property
+    def qubits(self):
+        return None
+
+    @property
+    def connectivity(self):
+        return None
+
+    @property
+    def natives(self):
+        return None
+
+    def set_dtype(self, dtype):
+        if dtype not in ("complex128", "complex64", "float64", "float32"):
+            raise_error(
+                ValueError,
+                f"Unknown ``dtype`` ``{dtype}``."
+                + "``dtype`` must be one of the following options: 'complex128', 'complex64',"
+                + "'float64', 'float32'",
+            )
+
+        if dtype != self.dtype:
+            self.dtype = dtype
+
             if self.matrices:
                 self.matrices = self.matrices.__class__(self.dtype)
 
@@ -56,28 +73,37 @@ class NumpyBackend(Backend):
         if nthreads > 1:
             raise_error(ValueError, "numpy does not support more than one thread.")
 
-    def cast(self, x, dtype=None, copy=False):
+    def cast(self, x, dtype=None, copy: bool = False):
         if dtype is None:
             dtype = self.dtype
+
         if isinstance(x, self.tensor_types):
             return x.astype(dtype, copy=copy)
-        elif self.issparse(x):
+
+        if self.is_sparse(x):
             return x.astype(dtype, copy=copy)
-        return np.array(x, dtype=dtype, copy=copy)
 
-    def issparse(self, x):
-        from scipy import sparse
+        return np.asarray(x, dtype=dtype, copy=copy if copy else None)
 
+    def is_sparse(self, x):
         return sparse.issparse(x)
 
     def to_numpy(self, x):
-        if self.issparse(x):
+        if self.is_sparse(x):
             return x.toarray()
         return x
 
     def compile(self, func):
         return func
 
+    def zeros(self, shape, dtype=None):
+        if dtype is None:  # pragma: no cover
+            dtype = self.dtype
+        return self.np.zeros(shape, dtype=dtype)
+
+    def random_choice(self, a, **kwargs):
+        return self.np.random.choice(a, **kwargs)
+
     def zero_state(self, nqubits):
         state = self.np.zeros(2**nqubits, dtype=self.dtype)
         state[0] = 1
@@ -108,14 +134,30 @@ class NumpyBackend(Backend):
         """Convert a gate to its matrix representation in the computational basis."""
         name = gate.__class__.__name__
         _matrix = getattr(self.matrices, name)
-        if callable(_matrix):
+        if name == "I":
             _matrix = _matrix(2 ** len(gate.target_qubits))
+        elif name == "Align":
+            _matrix = _matrix(0, 2)
+        elif callable(_matrix):
+            return self.matrix_parametrized(gate)
+
         return self.cast(_matrix, dtype=_matrix.dtype)
 
     def matrix_parametrized(self, gate):
         """Convert a parametrized gate to its matrix representation in the computational basis."""
         name = gate.__class__.__name__
-        _matrix = getattr(self.matrices, name)(*gate.parameters)
+        _matrix = getattr(self.matrices, name)
+        if name == "GeneralizedRBS":
+            _matrix = _matrix(
+                qubits_in=gate.init_args[0],
+                qubits_out=gate.init_args[1],
+                theta=gate.init_kwargs["theta"],
+                phi=gate.init_kwargs["phi"],
+            )
+        elif name == "FanOut":
+            _matrix = _matrix(*gate.init_args)
+        else:
+            _matrix = _matrix(*gate.parameters)
         return self.cast(_matrix, dtype=_matrix.dtype)
 
     def matrix_fused(self, fgate):
@@ -156,7 +198,6 @@ class NumpyBackend(Backend):
         return self.cast(matrix.toarray())
 
     def apply_gate(self, gate, state, nqubits):
-        state = self.cast(state)
         state = self.np.reshape(state, nqubits * (2,))
         matrix = gate.matrix(self)
         if gate.is_controlled_by:
@@ -250,7 +291,7 @@ class NumpyBackend(Backend):
 
     def apply_channel(self, channel, state, nqubits):
         probabilities = channel.coefficients + (1 - np.sum(channel.coefficients),)
-        index = self.sample_shots(probabilities, 1)[0]
+        index = int(self.sample_shots(probabilities, 1)[0])
         if index != len(channel.gates):
             gate = channel.gates[index]
             state = self.apply_gate(gate, state, nqubits)
@@ -309,13 +350,16 @@ class NumpyBackend(Backend):
         return self.np.reshape(state, shape)
 
     def reset_error_density_matrix(self, gate, state, nqubits):
-        from qibo.gates import X
+        from qibo.gates import X  # pylint: disable=C0415
+        from qibo.quantum_info.linalg_operations import (  # pylint: disable=C0415
+            partial_trace,
+        )
 
         state = self.cast(state)
         shape = state.shape
         q = gate.target_qubits[0]
         p_0, p_1 = gate.init_kwargs["p_0"], gate.init_kwargs["p_1"]
-        trace = self.partial_trace_density_matrix(state, (q,), nqubits)
+        trace = partial_trace(state, (q,), backend=self)
         trace = self.np.reshape(trace, 2 * (nqubits - 1) * (2,))
         zero = self.zero_density_matrix(1)
         zero = self.np.tensordot(trace, zero, 0)
@@ -333,11 +377,15 @@ class NumpyBackend(Backend):
         return self.np.reshape(state, shape)
 
     def depolarizing_error_density_matrix(self, gate, state, nqubits):
+        from qibo.quantum_info.linalg_operations import (  # pylint: disable=C0415
+            partial_trace,
+        )
+
         state = self.cast(state)
         shape = state.shape
         q = gate.target_qubits
         lam = gate.init_kwargs["lam"]
-        trace = self.partial_trace_density_matrix(state, q, nqubits)
+        trace = partial_trace(state, q, backend=self)
         trace = self.np.reshape(trace, 2 * (nqubits - len(q)) * (2,))
         identity = self.identity_density_matrix(len(q))
         identity = self.np.reshape(identity, 2 * len(q) * (2,))
@@ -363,7 +411,6 @@ class NumpyBackend(Backend):
         return state
 
     def execute_circuit(self, circuit, initial_state=None, nshots=1000):
-
         if isinstance(initial_state, type(circuit)):
             if not initial_state.density_matrix == circuit.density_matrix:
                 raise_error(
@@ -371,7 +418,8 @@ class NumpyBackend(Backend):
                     f"""Cannot set circuit with density_matrix {initial_state.density_matrix} as
                       initial state for circuit with density_matrix {circuit.density_matrix}.""",
                 )
-            elif (
+
+            if (
                 not initial_state.accelerators == circuit.accelerators
             ):  # pragma: no cover
                 raise_error(
@@ -379,10 +427,9 @@ class NumpyBackend(Backend):
                     f"""Cannot set circuit with accelerators {initial_state.density_matrix} as
                       initial state for circuit with accelerators {circuit.density_matrix}.""",
                 )
-            else:
-                return self.execute_circuit(initial_state + circuit, None, nshots)
+            return self.execute_circuit(initial_state + circuit, None, nshots)
         elif initial_state is not None:
-            initial_state = self.cast(initial_state)
+            initial_state = self.cast(initial_state, dtype=initial_state.dtype)
             valid_shape = (
                 2 * (2**circuit.nqubits,)
                 if circuit.density_matrix
@@ -398,10 +445,14 @@ class NumpyBackend(Backend):
         if circuit.repeated_execution:
             if circuit.measurements or circuit.has_collapse:
                 return self.execute_circuit_repeated(circuit, nshots, initial_state)
-            else:
-                raise RuntimeError(
-                    "Attempting to perform noisy simulation with `density_matrix=False` and no Measurement gate in the Circuit. If you wish to retrieve the statistics of the outcomes please include measurements in the circuit, otherwise set `density_matrix=True` to recover the final state."
-                )
+
+            raise_error(
+                RuntimeError,
+                "Attempting to perform noisy simulation with `density_matrix=False` "
+                + "and no Measurement gate in the Circuit. If you wish to retrieve the "
+                + "statistics of the outcomes please include measurements in the circuit, "
+                + "otherwise set `density_matrix=True` to recover the final state.",
+            )
 
         if circuit.accelerators:  # pragma: no cover
             return self.execute_distributed_circuit(circuit, initial_state, nshots)
@@ -410,21 +461,21 @@ class NumpyBackend(Backend):
             nqubits = circuit.nqubits
 
             if circuit.density_matrix:
-                if initial_state is None:
-                    state = self.zero_density_matrix(nqubits)
-                else:
-                    # cast to proper complex type
-                    state = self.cast(initial_state)
+                state = (
+                    self.zero_density_matrix(nqubits)
+                    if initial_state is None
+                    else self.cast(initial_state)
+                )
 
                 for gate in circuit.queue:
                     state = gate.apply_density_matrix(self, state, nqubits)
 
             else:
-                if initial_state is None:
-                    state = self.zero_state(nqubits)
-                else:
-                    # cast to proper complex type
-                    state = self.cast(initial_state)
+                state = (
+                    self.zero_state(nqubits)
+                    if initial_state is None
+                    else self.cast(initial_state)
+                )
 
                 for gate in circuit.queue:
                     state = gate.apply(self, state, nqubits)
@@ -437,19 +488,19 @@ class NumpyBackend(Backend):
                         state, circuit.measurements, backend=self, nshots=nshots
                     )
                     return circuit._final_state
-                else:
-                    circuit._final_state = QuantumState(state, backend=self)
-                    return circuit._final_state
 
-            else:
-                if circuit.measurements:
-                    circuit._final_state = CircuitResult(
-                        state, circuit.measurements, backend=self, nshots=nshots
-                    )
-                    return circuit._final_state
-                else:
-                    circuit._final_state = QuantumState(state, backend=self)
-                    return circuit._final_state
+                circuit._final_state = QuantumState(state, backend=self)
+                return circuit._final_state
+
+            if circuit.measurements:
+                circuit._final_state = CircuitResult(
+                    state, circuit.measurements, backend=self, nshots=nshots
+                )
+                return circuit._final_state
+
+            circuit._final_state = QuantumState(state, backend=self)
+
+            return circuit._final_state
 
         except self.oom_error:
             raise_error(
@@ -481,8 +532,11 @@ class NumpyBackend(Backend):
             and not circuit.measurements
             and not circuit.density_matrix
         ):
-            raise RuntimeError(
-                "The circuit contains only collapsing measurements (`collapse=True`) but `density_matrix=False`. Please set `density_matrix=True` to retrieve the final state after execution."
+            raise_error(
+                RuntimeError,
+                "The circuit contains only collapsing measurements (`collapse=True`) but "
+                + "`density_matrix=False`. Please set `density_matrix=True` to retrieve "
+                + "the final state after execution.",
             )
 
         results, final_states = [], []
@@ -631,7 +685,7 @@ class NumpyBackend(Backend):
         self.np.random.seed(seed)
 
     def sample_shots(self, probabilities, nshots):
-        return self.np.random.choice(
+        return self.random_choice(
             range(len(probabilities)), size=nshots, p=probabilities
         )
 
@@ -639,10 +693,10 @@ class NumpyBackend(Backend):
         return self.cast(shots, dtype=shots[0].dtype)
 
     def samples_to_binary(self, samples, nqubits):
-        ### This is faster just staying @ NumPy.
-        qrange = np.arange(nqubits - 1, -1, -1, dtype=np.int32)
-        samples = self.to_numpy(samples)
-        return np.mod(np.right_shift(samples[:, None], qrange), 2)
+        qrange = self.cast(
+            np.arange(nqubits - 1, -1, -1, dtype=np.int32), dtype=self.np.int32
+        )
+        return self.np.mod(self.np.right_shift(samples[:, None], qrange), 2)
 
     def samples_to_decimal(self, samples, nqubits):
         ### This is faster just staying @ NumPy.
@@ -668,7 +722,7 @@ class NumpyBackend(Backend):
         from qibo.config import SHOT_BATCH_SIZE
 
         nprobs = probabilities / self.np.sum(probabilities)
-        frequencies = self.np.zeros(len(nprobs), dtype=self.np.int64)
+        frequencies = self.zeros(len(nprobs), dtype=self.np.int64)
         for _ in range(nshots // SHOT_BATCH_SIZE):
             frequencies = self.update_frequencies(frequencies, nprobs, SHOT_BATCH_SIZE)
         frequencies = self.update_frequencies(
@@ -679,81 +733,57 @@ class NumpyBackend(Backend):
         )
 
     def apply_bitflips(self, noiseless_samples, bitflip_probabilities):
-        noiseless_samples = self.cast(noiseless_samples, dtype=noiseless_samples.dtype)
-        fprobs = self.cast(bitflip_probabilities, dtype="float64")
         sprobs = self.cast(np.random.random(noiseless_samples.shape), dtype="float64")
-        flip_0 = self.cast(sprobs < fprobs[0], dtype=noiseless_samples.dtype)
-        flip_1 = self.cast(sprobs < fprobs[1], dtype=noiseless_samples.dtype)
+        flip_0 = self.cast(
+            sprobs < bitflip_probabilities[0], dtype=noiseless_samples.dtype
+        )
+        flip_1 = self.cast(
+            sprobs < bitflip_probabilities[1], dtype=noiseless_samples.dtype
+        )
         noisy_samples = noiseless_samples + (1 - noiseless_samples) * flip_0
         noisy_samples = noisy_samples - noiseless_samples * flip_1
         return noisy_samples
 
-    def partial_trace(self, state, qubits, nqubits):
-        state = self.cast(state)
-        state = self.np.reshape(state, nqubits * (2,))
-        axes = 2 * [list(qubits)]
-        rho = self.np.tensordot(state, self.np.conj(state), axes)
-        shape = 2 * (2 ** (nqubits - len(qubits)),)
-        return self.np.reshape(rho, shape)
-
-    def partial_trace_density_matrix(self, state, qubits, nqubits):
-        state = self.cast(state)
-        state = self.np.reshape(state, 2 * nqubits * (2,))
-
-        order = tuple(sorted(qubits))
-        order += tuple(i for i in range(nqubits) if i not in qubits)
-        order += tuple(i + nqubits for i in order)
-        shape = 2 * (2 ** len(qubits), 2 ** (nqubits - len(qubits)))
-
-        state = self.np.transpose(state, order)
-        state = self.np.reshape(state, shape)
-        return self.np.einsum("abac->bc", state)
-
-    def calculate_norm(self, state, order=2):
+    def calculate_vector_norm(self, state, order=2):
         state = self.cast(state)
         return self.np.linalg.norm(state, order)
 
-    def calculate_norm_density_matrix(self, state, order="nuc"):
+    def calculate_matrix_norm(self, state, order="nuc"):
         state = self.cast(state)
         return self.np.linalg.norm(state, ord=order)
 
     def calculate_overlap(self, state1, state2):
-        return self.np.abs(self.np.sum(np.conj(self.cast(state1)) * self.cast(state2)))
+        return self.np.abs(
+            self.np.sum(self.np.conj(self.cast(state1)) * self.cast(state2))
+        )
 
     def calculate_overlap_density_matrix(self, state1, state2):
         return self.np.trace(
             self.np.matmul(self.np.conj(self.cast(state1)).T, self.cast(state2))
         )
 
-    def calculate_eigenvalues(self, matrix, k=6):
-        if self.issparse(matrix):
+    def calculate_eigenvalues(self, matrix, k: int = 6, hermitian: bool = True):
+        if self.is_sparse(matrix):
             log.warning(
                 "Calculating sparse matrix eigenvectors because "
                 "sparse modules do not provide ``eigvals`` method."
             )
             return self.calculate_eigenvectors(matrix, k=k)[0]
-        return np.linalg.eigvalsh(matrix)
+        if hermitian:
+            return np.linalg.eigvalsh(matrix)
+        return np.linalg.eigvals(matrix)
 
-    def calculate_eigenvectors(self, matrix, k=6):
-        if self.issparse(matrix):
+    def calculate_eigenvectors(self, matrix, k: int = 6, hermitian: bool = True):
+        if self.is_sparse(matrix):
             if k < matrix.shape[0]:
                 from scipy.sparse.linalg import eigsh
 
                 return eigsh(matrix, k=k, which="SA")
             else:  # pragma: no cover
                 matrix = self.to_numpy(matrix)
-        return np.linalg.eigh(matrix)
-
-    def calculate_matrix_exp(self, a, matrix, eigenvectors=None, eigenvalues=None):
-        if eigenvectors is None or self.issparse(matrix):
-            if self.issparse(matrix):
-                from scipy.sparse.linalg import expm
-            else:
-                from scipy.linalg import expm
-            return expm(-1j * a * matrix)
-        expd = self.np.diag(self.np.exp(-1j * a * eigenvalues))
-        ud = self.np.transpose(np.conj(eigenvectors))
-        return self.np.matmul(eigenvectors, self.np.matmul(expd, ud))
+        if hermitian:
+            return np.linalg.eigh(matrix)
+        return np.linalg.eig(matrix)
 
     def calculate_expectation_state(self, hamiltonian, state, normalize):
         statec = self.np.conj(state)
@@ -770,18 +800,76 @@ class NumpyBackend(Backend):
             ev /= norm
         return ev
 
-    # TODO: remove this method
-    def calculate_hamiltonian_matrix_product(self, matrix1, matrix2):
-        return matrix1 @ matrix2
+    def calculate_matrix_exp(
+        self,
+        matrix,
+        phase: Union[float, int, complex] = 1,
+        eigenvectors=None,
+        eigenvalues=None,
+    ):
+        if eigenvectors is None or self.is_sparse(matrix):
+            if self.is_sparse(matrix):
+                from scipy.sparse.linalg import expm
+            else:
+                from scipy.linalg import expm
 
-    # TODO: remove this method
-    def calculate_hamiltonian_state_product(self, matrix, state):
-        if len(tuple(state.shape)) > 2:
+            return expm(phase * matrix)
+
+        expd = self.np.exp(phase * eigenvalues)
+        ud = self.np.transpose(self.np.conj(eigenvectors))
+
+        return (eigenvectors * expd) @ ud
+
+    def calculate_matrix_log(self, matrix, base=2, eigenvectors=None, eigenvalues=None):
+        if eigenvectors is None:
+            # to_numpy and cast needed for GPUs
+            matrix_log = logm(self.to_numpy(matrix)) / float(np.log(base))
+            matrix_log = self.cast(matrix_log, dtype=matrix_log.dtype)
+
+            return matrix_log
+
+        # log = self.np.diag(self.np.log(eigenvalues) / float(np.log(base)))
+        log = self.np.log(eigenvalues) / float(np.log(base))
+        ud = self.np.transpose(np.conj(eigenvectors))
+
+        return (eigenvectors * log) @ ud
+
+    def calculate_matrix_power(
+        self,
+        matrix,
+        power: Union[float, int],
+        precision_singularity: float = 1e-14,
+    ):
+        if not isinstance(power, (float, int)):
             raise_error(
-                ValueError,
-                f"Cannot multiply Hamiltonian with rank-{len(tuple(state.shape))} tensor.",
+                TypeError,
+                f"``power`` must be either float or int, but it is type {type(power)}.",
             )
-        return matrix @ state
+
+        if power < 0.0:
+            # negative powers of singular matrices via SVD
+            determinant = self.np.linalg.det(matrix)
+            if abs(determinant) < precision_singularity:
+                return _calculate_negative_power_singular_matrix(
+                    matrix, power, precision_singularity, self.np, self
+                )
+
+        return fractional_matrix_power(matrix, power)
+
+    def calculate_matrix_sqrt(self, matrix):
+        return self.calculate_matrix_power(matrix, power=0.5)
+
+    def calculate_singular_value_decomposition(self, matrix):
+        return self.np.linalg.svd(matrix)
+
+    def calculate_jacobian_matrix(
+        self, circuit, parameters=None, initial_state=None, return_complex: bool = True
+    ):
+        raise_error(
+            NotImplementedError,
+            "This method is only implemented in backends that allow automatic differentiation, "
+            + "e.g. ``PytorchBackend`` and ``TensorflowBackend``.",
+        )
 
     def assert_allclose(self, value, target, rtol=1e-7, atol=0.0):
         if isinstance(value, CircuitResult) or isinstance(value, QuantumState):
@@ -792,21 +880,14 @@ class NumpyBackend(Backend):
         target = self.to_numpy(target)
         np.testing.assert_allclose(value, target, rtol=rtol, atol=atol)
 
-    def test_regressions(self, name):
-        if name == "test_measurementresult_apply_bitflips":
-            return [
-                [0, 0, 0, 0, 2, 3, 0, 0, 0, 0],
-                [0, 0, 0, 0, 2, 3, 0, 0, 0, 0],
-                [0, 0, 0, 0, 0, 1, 0, 0, 0, 0],
-                [0, 0, 0, 0, 2, 0, 0, 0, 0, 0],
-            ]
-        elif name == "test_probabilistic_measurement":
-            return {0: 249, 1: 231, 2: 253, 3: 267}
-        elif name == "test_unbalanced_probabilistic_measurement":
-            return {0: 171, 1: 148, 2: 161, 3: 520}
-        elif name == "test_post_measurement_bitflips_on_circuit":
-            return [
-                {5: 30},
-                {5: 18, 4: 5, 7: 4, 1: 2, 6: 1},
-                {4: 8, 2: 6, 5: 5, 1: 3, 3: 3, 6: 2, 7: 2, 0: 1},
-            ]
+
+def _calculate_negative_power_singular_matrix(
+    matrix, power: Union[float, int], precision_singularity: float, engine, backend
+):
+    """Calculate negative power of singular matrix."""
+    U, S, Vh = backend.calculate_singular_value_decomposition(matrix)
+    # cast needed because of different dtypes in `torch`
+    S = backend.cast(S)
+    S_inv = engine.where(engine.abs(S) < precision_singularity, 0.0, S**power)
+
+    return engine.linalg.inv(Vh) @ backend.np.diag(S_inv) @ engine.linalg.inv(U)
