@@ -24,8 +24,7 @@ class Channel(Gate):
     def controlled_by(self, *q):
         raise_error(ValueError, f"Noise channel cannot be controlled on qubits {q}.")
 
-    def on_qubits(self, qubit_map):  # pragma: no cover
-        # future TODO
+    def on_qubits(self, qubit_map: dict):  # pragma: no cover
         raise_error(
             NotImplementedError,
             "`on_qubits` method is not available for the `Channel` gate.",
@@ -48,8 +47,11 @@ class Channel(Gate):
         of the Kraus channel :math:`\\{K_{\\alpha}\\}_{\\alpha}`.
 
         .. math::
-            \\mathcal{E} = \\sum_{\\alpha} \\, |K_{\\alpha}\\rangle\\rangle
-                \\langle\\langle K_{\\alpha}|
+            \\mathcal{E} = \\sum_{\\alpha} \\, |K_{\\alpha})(K_{\\alpha}| \\, ,
+
+        where :math:`|K_{\\alpha})` is the vectorization of the Kraus operator
+        :math:`K_{\\alpha}`.
+        For a definition of vectorization, see :func:`qibo.quantum_info.vectorization`.
 
         Args:
             nqubits (int, optional): total number of qubits to be considered
@@ -62,7 +64,7 @@ class Channel(Gate):
                 vectorization is done block-wise. Defaut is ``"row"``.
             backend (:class:`qibo.backends.abstract.Backend`, optional):
                 backend to be used in the execution. If ``None``,
-                it uses :class:`qibo.backends.GlobalBackend`.
+                it uses the current backend.
                 Defaults to ``None``.
 
         Returns:
@@ -97,7 +99,9 @@ class Channel(Gate):
             kraus_op.append(gate)
             kraus_op = kraus_op.matrix(backend)
             kraus_op = vectorization(kraus_op, order=order, backend=backend)
-            super_op += coeff * np.outer(kraus_op, np.conj(kraus_op))
+            super_op = super_op + coeff * backend.np.outer(
+                kraus_op, backend.np.conj(kraus_op)
+            )
             del kraus_op
 
         return super_op
@@ -116,7 +120,7 @@ class Channel(Gate):
                 it raises ``NotImplementedError``. Defaut is ``"row"``.
             backend (:class:`qibo.backends.abstract.Backend`, optional):
                 backend to be used in the execution. If ``None``,
-                it uses :class:`qibo.backends.GlobalBackend`.
+                it uses the current backend.
                 Defaults to ``None``.
 
         Returns:
@@ -155,7 +159,7 @@ class Channel(Gate):
                 Pauli elements in the basis. Default is "IXYZ".
             backend (:class:`qibo.backends.abstract.Backend`, optional): backend
                 to be used in the execution. If ``None``, it uses
-                :class:`qibo.backends.GlobalBackend`.
+                the current backend.
                 Defaults to ``None``.
 
         Returns:
@@ -176,7 +180,9 @@ class Channel(Gate):
             nqubits, normalize, pauli_order=pauli_order, backend=backend
         )
 
-        super_op = unitary @ super_op @ np.transpose(np.conj(unitary))
+        super_op = (
+            unitary @ super_op @ backend.np.transpose(backend.np.conj(unitary), (1, 0))
+        )
 
         return super_op
 
@@ -309,6 +315,11 @@ class KrausChannel(Channel):
         self.coefficients = len(self.gates) * (1,)
         self.coefficient_sum = 1
         self._all_unitary_operators = True if all(unitary_check) else False
+        self._qubits = qubits
+
+    def on_qubits(self, qubit_map: dict):
+        _qubits = _get_new_qubit_map(self, self._qubits, qubit_map)
+        return self.__class__(_qubits, self.gates)
 
 
 class UnitaryChannel(KrausChannel):
@@ -363,10 +374,15 @@ class UnitaryChannel(KrausChannel):
             )
 
         self.init_args = [probabilities, self.gates]
+        self._qubits = qubits
 
     def apply(self, backend, state, nqubits):
         return backend.apply_channel(self, state, nqubits)
 
+    def on_qubits(self, qubit_map: dict):
+        _qubits = _get_new_qubit_map(self, self._qubits, qubit_map)
+        return self.__class__(_qubits, list(zip(self.init_args[0], self.init_args[1])))
+
 
 class PauliNoiseChannel(UnitaryChannel):
     """Multi-qubit noise channel that applies Pauli operators with given probabilities.
@@ -443,6 +459,10 @@ class PauliNoiseChannel(UnitaryChannel):
         self.init_args = qubits
         self.init_kwargs = dict(operators)
 
+    def on_qubits(self, qubit_map: dict):
+        _qubits = _get_new_qubit_map(self, self.init_args, qubit_map)
+        return self.__class__(_qubits, list(self.init_kwargs.items()))
+
 
 class DepolarizingChannel(PauliNoiseChannel):
     """:math:`n`-qubit Depolarizing quantum error channel,
@@ -494,6 +514,10 @@ class DepolarizingChannel(PauliNoiseChannel):
     def apply_density_matrix(self, backend, state, nqubits):
         return backend.depolarizing_error_density_matrix(self, state, nqubits)
 
+    def on_qubits(self, qubit_map: dict):
+        _qubits = _get_new_qubit_map(self, self.init_args[0], qubit_map)
+        return self.__class__(_qubits, self.init_kwargs["lam"])
+
 
 class ThermalRelaxationChannel(KrausChannel):
     """Single-qubit thermal relaxation error channel.
@@ -637,6 +661,11 @@ class ThermalRelaxationChannel(KrausChannel):
         self.name = "ThermalRelaxationChannel"
         self.draw_label = "TR"
 
+    def on_qubits(self, qubit_map: dict):
+        qubit = qubit_map.get(self.init_args[0])
+        parameters = self.init_args[1:] + [self.init_kwargs["excited_population"]]
+        return self.__class__(qubit, parameters)
+
     def apply_density_matrix(self, backend, state, nqubits):
         qubit = self.target_qubits[0]
 
@@ -646,12 +675,14 @@ class ThermalRelaxationChannel(KrausChannel):
                 self.init_kwargs["p_1"],
                 self.init_kwargs["e_t2"],
             )
-            matrix = [
-                [1 - preset1, 0, 0, preset0],
-                [0, e_t2, 0, 0],
-                [0, 0, e_t2, 0],
-                [preset1, 0, 0, 1 - preset0],
-            ]
+            matrix = backend.cast(
+                [
+                    [1 - preset1, 0, 0, preset0],
+                    [0, e_t2, 0, 0],
+                    [0, 0, e_t2, 0],
+                    [preset1, 0, 0, 1 - preset0],
+                ]
+            )
 
             qubits = (qubit, qubit + nqubits)
             gate = Unitary(matrix, *qubits)
@@ -700,6 +731,11 @@ class AmplitudeDampingChannel(KrausChannel):
         super().__init__([(qubit,)] * len(operators), operators)
         self.name = "AmplitudeDampingChannel"
         self.draw_label = "AD"
+        self.init_kwargs["gamma"] = gamma
+
+    def on_qubits(self, qubit_map: dict):
+        qubit = qubit_map.get(self.qubits[0])
+        return self.__class__(qubit, self.init_kwargs["gamma"])
 
 
 class PhaseDampingChannel(KrausChannel):
@@ -735,6 +771,11 @@ class PhaseDampingChannel(KrausChannel):
         super().__init__([(qubit,)] * len(operators), operators)
         self.name = "PhaseDampingChannel"
         self.draw_label = "PD"
+        self.init_kwargs["gamma"] = gamma
+
+    def on_qubits(self, qubit_map):
+        qubit = qubit_map.get(self.qubits[0])
+        return self.__class__(qubit, self.init_kwargs["gamma"])
 
 
 class ReadoutErrorChannel(KrausChannel):
@@ -742,7 +783,7 @@ class ReadoutErrorChannel(KrausChannel):
 
     Args:
         qubits (int or list or tuple): Qubit ids that the channel acts on.
-        probabilities (array): row-stochastic matrix :math:`P` with all
+        probabilities (ndarray): row-stochastic matrix :math:`P` with all
             readout transition probabilities.
 
     Example:
@@ -774,10 +815,15 @@ class ReadoutErrorChannel(KrausChannel):
                 operator = np.zeros((dim, dim))
                 operator[j, k] = sqrt(probabilities[k, j])
                 operators.append(operator)
-
         super().__init__([qubits] * len(operators), operators)
         self.name = "ReadoutErrorChannel"
         self.draw_label = "RE"
+        self.probabilities = probabilities
+        self._qubits = qubits
+
+    def on_qubits(self, qubit_map: dict):
+        _qubits = _get_new_qubit_map(self, self._qubits, qubit_map)
+        return self.__class__(_qubits, self.probabilities)
 
 
 class ResetChannel(KrausChannel):
@@ -830,3 +876,17 @@ class ResetChannel(KrausChannel):
 
     def apply_density_matrix(self, backend, state, nqubits):
         return backend.reset_error_density_matrix(self, state, nqubits)
+
+    def on_qubits(self, qubit_map: dict):
+        qubit = qubit_map.get(self.qubits[0])
+        return self.__class__(qubit, [self.init_kwargs["p_0"], self.init_kwargs["p_1"]])
+
+
+def _get_new_qubit_map(gate, qubits, qubit_map: dict):
+    """Wrapper for qubit_map used in ``on_qubits``."""
+    if isinstance(gate, (PauliNoiseChannel, ReadoutErrorChannel)) or (
+        isinstance(gate, ResetChannel) and len(gate.qubits) == 1
+    ):
+        return tuple(qubit_map.get(qubit) for qubit in qubits)
+
+    return [tuple(qubit_map.get(qubit) for qubit in row) for row in qubits]
