@@ -1,8 +1,9 @@
 import math
-from typing import List
+from typing import List, Tuple, Union
 
 import numpy as np
 
+from qibo.backends import _check_backend
 from qibo.config import PRECISION_TOL, raise_error
 from qibo.gates.abstract import Gate, ParametrizedGate
 from qibo.parameter import Parameter
@@ -82,7 +83,7 @@ class X(Gate):
             gate = super().controlled_by(*q)
         return gate
 
-    def decompose(self, *free, use_toffolis=True):
+    def _base_decompose(self, *free, use_toffolis=True):
         """Decomposes multi-control ``X`` gate to one-qubit, ``CNOT`` and ``TOFFOLI`` gates.
 
         Args:
@@ -92,8 +93,8 @@ class X(Gate):
                 See :class:`qibo.gates.TOFFOLI` for more details on this representation.
 
         Returns:
-            List with one-qubit, ``CNOT`` and ``TOFFOLI`` gates that have the
-            same effect as applying the original multi-control gate.
+            list: Set of one-qubit, :class:`qibo.gates.CNOT`, and :class:`qibo.gates.TOFFOLI`
+            gates that have the same effect as applying the original multi-control gate.
         """
         if set(free) & set(self.qubits):
             raise_error(
@@ -104,23 +105,25 @@ class X(Gate):
 
         controls = self.control_qubits
         target = self.target_qubits[0]
-        m = len(controls)
-        if m < 3:
+        ncontrols = len(controls)
+        if ncontrols < 3:
             return [self.__class__(target).controlled_by(*controls)]
 
         decomp_gates = []
-        n = m + 1 + len(free)
-        if (n >= 2 * m - 1) and (m >= 3):
+        nqubits = ncontrols + 1 + len(free)
+        if (nqubits >= 2 * ncontrols - 1) and (ncontrols >= 3):
             gates1 = [
                 TOFFOLI(
-                    controls[m - 2 - i], free[m - 4 - i], free[m - 3 - i]
-                ).congruent(use_toffolis=use_toffolis)
-                for i in range(m - 3)
+                    controls[ncontrols - 2 - k],
+                    free[ncontrols - 4 - k],
+                    free[ncontrols - 3 - k],
+                ).congruent()
+                for k in range(ncontrols - 3)
             ]
-            gates2 = TOFFOLI(controls[0], controls[1], free[0]).congruent(
-                use_toffolis=use_toffolis
+            gates2 = TOFFOLI(controls[0], controls[1], free[0]).congruent()
+            first_toffoli = TOFFOLI(
+                controls[ncontrols - 1], free[ncontrols - 3], target
             )
-            first_toffoli = TOFFOLI(controls[m - 1], free[m - 3], target)
 
             decomp_gates.append(first_toffoli)
             for gates in gates1:
@@ -130,15 +133,15 @@ class X(Gate):
                 decomp_gates.extend(gates)
 
         elif len(free) >= 1:
-            m1 = n // 2
+            m1 = nqubits // 2
             free1 = controls[m1:] + (target,) + tuple(free[1:])
             x1 = self.__class__(free[0]).controlled_by(*controls[:m1])
-            part1 = x1.decompose(*free1, use_toffolis=use_toffolis)
+            part1 = x1._base_decompose(*free1, use_toffolis=use_toffolis)
 
             free2 = controls[:m1] + tuple(free[1:])
             controls2 = controls[m1:] + (free[0],)
             x2 = self.__class__(target).controlled_by(*controls2)
-            part2 = x2.decompose(*free2, use_toffolis=use_toffolis)
+            part2 = x2._base_decompose(*free2, use_toffolis=use_toffolis)
 
             decomp_gates = [*part1, *part2]
 
@@ -146,12 +149,15 @@ class X(Gate):
             # impractical case
             raise_error(
                 NotImplementedError,
-                "X decomposition not implemented for zero free qubits.",
+                "``X`` decomposition not implemented for zero free qubits.",
             )
 
         decomp_gates.extend(decomp_gates)
         return decomp_gates
 
+    def decompose(self, *free, use_toffolis: bool = True) -> List["Gate"]:
+        return self._base_decompose(*free, use_toffolis=use_toffolis)
+
     def basis_rotation(self):
         return H(self.target_qubits[0])
 
@@ -200,7 +206,9 @@ class Y(Gate):
         from qibo import matrices  # pylint: disable=C0415
 
         matrix = (matrices.Y + matrices.Z) / math.sqrt(2)
-        return Unitary(matrix, self.target_qubits[0], trainable=False)
+        gate = Unitary(matrix, self.target_qubits[0], trainable=False)
+        gate.clifford = True
+        return gate
 
 
 class Z(Gate):
@@ -230,6 +238,10 @@ class Z(Gate):
     def clifford(self):
         return True
 
+    @property
+    def hamming_weight(self):
+        return True
+
     @property
     def qasm_label(self):
         return "z"
@@ -239,6 +251,8 @@ class Z(Gate):
         """Fall back to CZ if there is only one control."""
         if len(q) == 1:
             gate = CZ(q[0], self.target_qubits[0])
+        elif len(q) == 2:
+            gate = CCZ(q[0], q[1], self.target_qubits[0])
         else:
             gate = super().controlled_by(*q)
         return gate
@@ -278,7 +292,7 @@ class SX(Gate):
     def qasm_label(self):
         return "sx"
 
-    def decompose(self):
+    def _base_decompose(self, *free, use_toffolis=True):
         """Decomposition of :math:`\\sqrt{X}` up to global phase.
 
         A global phase difference exists between the definitions of
@@ -328,7 +342,7 @@ class SXDG(Gate):
     def qasm_label(self):
         return "sxdg"
 
-    def decompose(self):
+    def _base_decompose(self, *free, use_toffolis=True):
         """Decomposition of :math:`(\\sqrt{X})^{\\dagger}` up to global phase.
 
         A global phase difference exists between the definitions of
@@ -374,6 +388,10 @@ class S(Gate):
     def clifford(self):
         return True
 
+    @property
+    def hamming_weight(self):
+        return True
+
     @property
     def qasm_label(self):
         return "s"
@@ -409,6 +427,10 @@ class SDG(Gate):
     def clifford(self):
         return True
 
+    @property
+    def hamming_weight(self):
+        return True
+
     @property
     def qasm_label(self):
         return "sdg"
@@ -440,6 +462,10 @@ class T(Gate):
         self.init_args = [q]
         self.unitary = True
 
+    @property
+    def hamming_weight(self):
+        return True
+
     @property
     def qasm_label(self):
         return "t"
@@ -471,6 +497,10 @@ class TDG(Gate):
         self.init_args = [q]
         self.unitary = True
 
+    @property
+    def hamming_weight(self):
+        return True
+
     @property
     def qasm_label(self):
         return "tdg"
@@ -498,6 +528,10 @@ class I(Gate):
     def clifford(self):
         return True
 
+    @property
+    def hamming_weight(self):
+        return True
+
     @property
     def qasm_label(self):
         return "id"
@@ -506,13 +540,16 @@ class I(Gate):
 class Align(ParametrizedGate):
     """Aligns proceeding qubit operations and (optionally) waits ``delay`` amount of time.
 
+    .. note::
+        For this gate, the ``trainable`` parameter is by default set to ``False``.
+
     Args:
         q (int): The qubit ID.
         delay (int, optional): The time (in ns) for which to delay circuit execution on the specified qubits.
             Defaults to ``0`` (zero).
     """
 
-    def __init__(self, q, delay=0, trainable=True):
+    def __init__(self, q, delay=0, trainable=False):
         if not isinstance(delay, int):
             raise_error(
                 TypeError, f"delay must be type int, but it is type {type(delay)}."
@@ -535,6 +572,11 @@ def _is_clifford_given_angle(angle):
     return isinstance(angle, (float, int)) and (angle % (np.pi / 2)).is_integer()
 
 
+def _is_hamming_weight_given_angle(angle, target=2 * np.pi):
+    """Helper function to update Hamming weight boolean condition according to the given angles ``angle`` and ``target``."""
+    return isinstance(angle, (float, int)) and (angle % target).is_integer()
+
+
 class _Rn_(ParametrizedGate):
     """Abstract class for defining the RX, RY and RZ rotations.
 
@@ -582,6 +624,20 @@ class _Rn_(ParametrizedGate):
             gate = super().controlled_by(*q)
         return gate
 
+    def generator_eigenvalue(self):
+        return 0.5
+
+    def gradient(self, backend=None) -> "Gate":
+        """Gradient of the Gate."""
+        backend = _check_backend(backend=backend)
+        return Unitary(
+            -1j
+            * self.generator_eigenvalue()
+            * self.generator(backend)
+            @ backend.matrix_parametrized(self),
+            *self.target_qubits,
+        )
+
 
 class RX(_Rn_):
     """Rotation around the X-axis of the Bloch sphere.
@@ -610,12 +666,17 @@ class RX(_Rn_):
         self.draw_label = "RX"
         self._controlled_gate = CRX
 
+    @property
+    def hamming_weight(self):
+        return _is_hamming_weight_given_angle(self.parameters[0])
+
     @property
     def qasm_label(self):
         return "rx"
 
-    def generator_eigenvalue(self):
-        return 0.5
+    def generator(self, backend=None):
+        backend = _check_backend(backend)
+        return backend.matrices.X
 
 
 class RY(_Rn_):
@@ -645,12 +706,17 @@ class RY(_Rn_):
         self.draw_label = "RY"
         self._controlled_gate = CRY
 
+    @property
+    def hamming_weight(self):
+        return _is_hamming_weight_given_angle(self.parameters[0])
+
     @property
     def qasm_label(self):
         return "ry"
 
-    def generator_eigenvalue(self):
-        return 0.5
+    def generator(self, backend=None):
+        backend = _check_backend(backend)
+        return backend.matrices.Y
 
 
 class RZ(_Rn_):
@@ -678,12 +744,17 @@ class RZ(_Rn_):
         self.draw_label = "RZ"
         self._controlled_gate = CRZ
 
+    @property
+    def hamming_weight(self):
+        return True
+
     @property
     def qasm_label(self):
         return "rz"
 
-    def generator_eigenvalue(self):
-        return 0.5
+    def generator(self, backend=None):
+        backend = _check_backend(backend)
+        return backend.matrices.Z
 
 
 class PRX(ParametrizedGate):
@@ -691,11 +762,11 @@ class PRX(ParametrizedGate):
 
     Corresponds to the following unitary matrix
 
-        .. math::
-            \\begin{pmatrix}
-                \\cos{(\\theta / 2)} & -i e^{-i \\phi} \\sin{(\\theta / 2)} \\
-                -i e^{i \\phi} \\sin{(\\theta / 2)} & \\cos{(\\theta / 2)}
-            \\end{pmatrix}
+    .. math::
+        \\begin{pmatrix}
+            \\cos{(\\theta / 2)} & -i e^{-i \\phi} \\sin{(\\theta / 2)} \\\\
+            -i e^{i \\phi} \\sin{(\\theta / 2)} & \\cos{(\\theta / 2)}
+        \\end{pmatrix}
 
     Args:
         q (int): the qubit id number.
@@ -726,6 +797,10 @@ class PRX(ParametrizedGate):
             "trainable": trainable,
         }
 
+    @property
+    def hamming_weight(self):
+        return _is_hamming_weight_given_angle(self.parameters[0])
+
     @property
     def qasm_label(self):
         return "prx"
@@ -737,7 +812,7 @@ class PRX(ParametrizedGate):
             self.target_qubits[0], theta, phi
         )  # pylint: disable=E1130
 
-    def decompose(self):
+    def _base_decompose(self, *free, use_toffolis=True):
         """Decomposition of Phase-:math:`RX` gate."""
         from qibo.transpiler.decompositions import (  # pylint: disable=C0415
             standard_decompositions,
@@ -821,6 +896,10 @@ class GPI2(ParametrizedGate):
     def qasm_label(self):
         return "gpi2"
 
+    @property
+    def clifford(self):
+        return _is_clifford_given_angle(self.parameters[0])
+
     def _dagger(self) -> "Gate":
         """"""
         return self.__class__(self.target_qubits[0], self.parameters[0] + math.pi)
@@ -888,6 +967,10 @@ class U1(_Un_):
         self.parameters = theta
         self.init_kwargs = {"theta": theta, "trainable": trainable}
 
+    @property
+    def hamming_weight(self):
+        return True
+
     @property
     def qasm_label(self):
         return "u1"
@@ -979,6 +1062,10 @@ class U3(_Un_):
         self.parameter_names = ["theta", "phi", "lam"]
         self.parameters = theta, phi, lam
 
+    @property
+    def hamming_weight(self):
+        return _is_hamming_weight_given_angle(self.parameters[0])
+
     @property
     def qasm_label(self):
         return "u3"
@@ -988,7 +1075,7 @@ class U3(_Un_):
         theta, lam, phi = tuple(-x for x in self.parameters)  # pylint: disable=E1130
         return self.__class__(self.target_qubits[0], theta, phi, lam)
 
-    def decompose(self) -> List[Gate]:
+    def _base_decompose(self, *free, use_toffolis=True) -> List[Gate]:
         """Decomposition of :math:`U_{3}` up to global phase.
 
         A global phase difference exists between the definitions of
@@ -1045,6 +1132,10 @@ class U1q(_Un_):
         self.parameter_names = ["theta", "phi"]
         self.parameters = theta, phi
 
+    @property
+    def hamming_weight(self):
+        return _is_hamming_weight_given_angle(self.parameters[0])
+
     def _dagger(self) -> "Gate":
         """"""
         theta, phi = self.init_kwargs["theta"], self.init_kwargs["phi"]
@@ -1086,7 +1177,7 @@ class CNOT(Gate):
     def qasm_label(self):
         return "cx"
 
-    def decompose(self, *free, use_toffolis: bool = True) -> List[Gate]:
+    def _base_decompose(self, *free, use_toffolis: bool = True) -> List[Gate]:
         q0, q1 = self.control_qubits[0], self.target_qubits[0]
         return [self.__class__(q0, q1)]
 
@@ -1126,7 +1217,7 @@ class CY(Gate):
     def qasm_label(self):
         return "cy"
 
-    def decompose(self) -> List[Gate]:
+    def _base_decompose(self, *free, use_toffolis=True) -> List[Gate]:
         """Decomposition of :math:`\\text{CY}` gate.
 
         Decompose :math:`\\text{CY}` gate into :class:`qibo.gates.SDG` in
@@ -1171,11 +1262,15 @@ class CZ(Gate):
     def clifford(self):
         return True
 
+    @property
+    def hamming_weight(self):
+        return True
+
     @property
     def qasm_label(self):
         return "cz"
 
-    def decompose(self) -> List[Gate]:
+    def _base_decompose(self, *free, use_toffolis=True) -> List[Gate]:
         """Decomposition of :math:`\\text{CZ}` gate.
 
         Decompose :math:`\\text{CZ}` gate into :class:`qibo.gates.H` in
@@ -1220,7 +1315,7 @@ class CSX(Gate):
     def qasm_label(self):
         return "csx"
 
-    def decompose(self, *free, use_toffolis: bool = True) -> List[Gate]:
+    def _base_decompose(self, *free, use_toffolis: bool = True) -> List[Gate]:
         """"""
         from qibo.transpiler.decompositions import (  # pylint: disable=C0415
             standard_decompositions,
@@ -1264,7 +1359,7 @@ class CSXDG(Gate):
     def qasm_label(self):
         return "csxdg"
 
-    def decompose(self, *free, use_toffolis: bool = True) -> List[Gate]:
+    def _base_decompose(self, *free, use_toffolis: bool = True) -> List[Gate]:
         """"""
         from qibo.transpiler.decompositions import (  # pylint: disable=C0415
             standard_decompositions,
@@ -1339,6 +1434,10 @@ class CRX(_CRn_):
         self.name = "crx"
         self.draw_label = "RX"
 
+    @property
+    def hamming_weight(self):
+        return _is_hamming_weight_given_angle(self.parameters[0])
+
     @property
     def qasm_label(self):
         return "crx"
@@ -1373,10 +1472,22 @@ class CRY(_CRn_):
         self.name = "cry"
         self.draw_label = "RY"
 
+    @property
+    def hamming_weight(self):
+        return _is_hamming_weight_given_angle(self.parameters[0])
+
     @property
     def qasm_label(self):
         return "cry"
 
+    def _base_decompose(self, *free, use_toffolis=True) -> List[Gate]:
+        """Decomposition of :math:`\\text{CRY}` gate."""
+        from qibo.transpiler.decompositions import (  # pylint: disable=C0415
+            standard_decompositions,
+        )
+
+        return standard_decompositions(self)
+
 
 class CRZ(_CRn_):
     """Controlled rotation around the Z-axis for the Bloch sphere.
@@ -1405,6 +1516,10 @@ class CRZ(_CRn_):
         self.name = "crz"
         self.draw_label = "RZ"
 
+    @property
+    def hamming_weight(self):
+        return True
+
     @property
     def qasm_label(self):
         return "crz"
@@ -1464,6 +1579,10 @@ class CU1(_CUn_):
         self.parameters = theta
         self.init_kwargs = {"theta": theta, "trainable": trainable}
 
+    @property
+    def hamming_weight(self):
+        return True
+
     @property
     def qasm_label(self):
         return "cu1"
@@ -1560,6 +1679,10 @@ class CU3(_CUn_):
         self.parameter_names = ["theta", "phi", "lam"]
         self.parameters = theta, phi, lam
 
+    @property
+    def hamming_weight(self):
+        return _is_hamming_weight_given_angle(self.parameters[0])
+
     @property
     def qasm_label(self):
         return "cu3"
@@ -1602,6 +1725,10 @@ class SWAP(Gate):
     def clifford(self):
         return True
 
+    @property
+    def hamming_weight(self):
+        return True
+
     @property
     def qasm_label(self):
         return "swap"
@@ -1637,6 +1764,10 @@ class iSWAP(Gate):
     def clifford(self):
         return True
 
+    @property
+    def hamming_weight(self):
+        return True
+
     @property
     def qasm_label(self):
         return "iswap"
@@ -1668,6 +1799,10 @@ class SiSWAP(Gate):
         self.init_args = [q0, q1]
         self.unitary = True
 
+    @property
+    def hamming_weight(self):
+        return True
+
     def _dagger(self) -> "Gate":
         return SiSWAPDG(*self.qubits)
 
@@ -1698,6 +1833,10 @@ class SiSWAPDG(Gate):
         self.init_args = [q0, q1]
         self.unitary = True
 
+    @property
+    def hamming_weight(self):
+        return True
+
     def _dagger(self) -> "Gate":
         return SiSWAP(*self.qubits)
 
@@ -1732,14 +1871,18 @@ class FSWAP(Gate):
     def clifford(self):
         return True
 
+    @property
+    def hamming_weight(self):
+        return True
+
     @property
     def qasm_label(self):
         return "fswap"
 
-    def decompose(self, *free, use_toffolis: bool = True) -> List[Gate]:
+    def _base_decompose(self, *free, use_toffolis: bool = True) -> List[Gate]:
         """"""
         q0, q1 = self.target_qubits
-        return [X(q1)] + GIVENS(q0, q1, np.pi / 2).decompose() + [X(q0)]
+        return [X(q1)] + GIVENS(q0, q1, np.pi / 2)._base_decompose() + [X(q0)]
 
 
 class fSim(ParametrizedGate):
@@ -1782,6 +1925,10 @@ class fSim(ParametrizedGate):
         self.init_args = [q0, q1]
         self.init_kwargs = {"theta": theta, "phi": phi, "trainable": trainable}
 
+    @property
+    def hamming_weight(self):
+        return True
+
     def _dagger(self) -> "Gate":
         """"""
         q0, q1 = self.target_qubits
@@ -1819,6 +1966,10 @@ class SYC(Gate):
         self.init_args = [q0, q1]
         self.unitary = True
 
+    @property
+    def hamming_weight(self):
+        return True
+
     def _dagger(self) -> "Gate":
         """"""
         return fSim(*self.target_qubits, -np.pi / 2, -np.pi / 6)
@@ -1861,6 +2012,10 @@ class GeneralizedfSim(ParametrizedGate):
         self.init_args = [q0, q1]
         self.init_kwargs = {"unitary": unitary, "phi": phi, "trainable": trainable}
 
+    @property
+    def hamming_weight(self):
+        return True
+
     def _dagger(self):
         q0, q1 = self.target_qubits
         u, phi = self.parameters
@@ -1936,6 +2091,10 @@ class RXX(_Rnn_):
         self.name = "rxx"
         self.draw_label = "RXX"
 
+    @property
+    def hamming_weight(self):
+        return _is_hamming_weight_given_angle(self.parameters[0])
+
     @property
     def qasm_label(self):
         return "rxx"
@@ -1967,6 +2126,10 @@ class RYY(_Rnn_):
         self.name = "ryy"
         self.draw_label = "RYY"
 
+    @property
+    def hamming_weight(self):
+        return _is_hamming_weight_given_angle(self.parameters[0])
+
     @property
     def qasm_label(self):
         return "ryy"
@@ -1999,6 +2162,10 @@ class RZZ(_Rnn_):
         self.name = "rzz"
         self.draw_label = "RZZ"
 
+    @property
+    def hamming_weight(self):
+        return True
+
     @property
     def qasm_label(self):
         return "rzz"
@@ -2037,7 +2204,11 @@ class RZX(_Rnn_):
         self.name = "rzx"
         self.draw_label = "RZX"
 
-    def decompose(self, *free, use_toffolis: bool = True) -> List[Gate]:
+    @property
+    def hamming_weight(self):
+        return _is_hamming_weight_given_angle(self.parameters[0])
+
+    def _base_decompose(self, *free, use_toffolis: bool = True) -> List[Gate]:
         """"""
         from qibo.transpiler.decompositions import (  # pylint: disable=C0415
             standard_decompositions,
@@ -2075,7 +2246,11 @@ class RXXYY(_Rnn_):
         self.name = "rxxyy"
         self.draw_label = "RXXYY"
 
-    def decompose(self, *free, use_toffolis: bool = True) -> List[Gate]:
+    @property
+    def hamming_weight(self):
+        return True
+
+    def _base_decompose(self, *free, use_toffolis: bool = True) -> List[Gate]:
         """Decomposition of :math:`\\text{R_{XX-YY}}` up to global phase.
 
         This decomposition has a global phase difference with respect to
@@ -2143,6 +2318,10 @@ class MS(ParametrizedGate):
             "trainable": trainable,
         }
 
+    @property
+    def hamming_weight(self):
+        return _is_hamming_weight_given_angle(self.parameters[2])
+
     @property
     def qasm_label(self):
         return "ms"
@@ -2190,11 +2369,15 @@ class GIVENS(ParametrizedGate):
         self.init_args = [q0, q1]
         self.init_kwargs = {"theta": theta, "trainable": trainable}
 
+    @property
+    def hamming_weight(self):
+        return True
+
     def _dagger(self) -> "Gate":
         """"""
         return self.__class__(*self.target_qubits, -self.parameters[0])
 
-    def decompose(self, *free, use_toffolis: bool = True) -> List[Gate]:
+    def _base_decompose(self, *free, use_toffolis: bool = True) -> List[Gate]:
         """Decomposition of RBS gate according to `ArXiv:2109.09685
         <https://arxiv.org/abs/2109.09685>`_."""
         from qibo.transpiler.decompositions import (  # pylint: disable=C0415
@@ -2244,11 +2427,15 @@ class RBS(ParametrizedGate):
         self.init_args = [q0, q1]
         self.init_kwargs = {"theta": theta, "trainable": trainable}
 
+    @property
+    def hamming_weight(self):
+        return True
+
     def _dagger(self) -> "Gate":
         """"""
         return self.__class__(*self.target_qubits, -self.parameters[0])
 
-    def decompose(self, *free, use_toffolis: bool = True) -> List[Gate]:
+    def _base_decompose(self, *free, use_toffolis: bool = True) -> List[Gate]:
         """Decomposition of RBS gate according to `ArXiv:2109.09685
         <https://arxiv.org/abs/2109.09685>`_."""
         from qibo.transpiler.decompositions import (  # pylint: disable=C0415
@@ -2289,7 +2476,7 @@ class ECR(Gate):
     def clifford(self):
         return True
 
-    def decompose(self, *free, use_toffolis: bool = True) -> List[Gate]:
+    def _base_decompose(self, *free, use_toffolis: bool = True) -> List[Gate]:
         """Decomposition of :math:`\\textup{ECR}` gate up to global phase.
 
         A global phase difference exists between the definitions of
@@ -2342,10 +2529,18 @@ class TOFFOLI(Gate):
     def qasm_label(self):
         return "ccx"
 
-    def decompose(self, *free, use_toffolis: bool = True) -> List[Gate]:
-        c0, c1 = self.control_qubits
-        t = self.target_qubits[0]
-        return [self.__class__(c0, c1, t)]
+    def _base_decompose(self, *free, use_toffolis: bool = True) -> List[Gate]:
+        """Decomposition of :math:`\\text{TOFFOLI}` gate.
+
+        Decompose :math:`\\text{TOFFOLI}` gate into :class:`qibo.gates.CNOT` gates,
+        :class:`qibo.gates.T` gates, :class:`qibo.gates.TDG` gates,
+        and :class:`qibo.gates.H` gates.
+        """
+        from qibo.transpiler.decompositions import (  # pylint: disable=C0415
+            standard_decompositions,
+        )
+
+        return standard_decompositions(self)
 
     def congruent(self, use_toffolis: bool = True) -> List[Gate]:
         """Congruent representation of ``TOFFOLI`` gate.
@@ -2364,8 +2559,6 @@ class TOFFOLI(Gate):
             List with ``RY`` and ``CNOT`` gates that have the same effect as
             applying the original ``TOFFOLI`` gate.
         """
-        if use_toffolis:
-            return self.decompose()
 
         control0, control1 = self.control_qubits
         target = self.target_qubits[0]
@@ -2412,11 +2605,15 @@ class CCZ(Gate):
         self.init_args = [q0, q1, q2]
         self.unitary = True
 
+    @property
+    def hamming_weight(self):
+        return True
+
     @property
     def qasm_label(self):
         return "ccz"
 
-    def decompose(self) -> List[Gate]:
+    def _base_decompose(self, *free, use_toffolis=True) -> List[Gate]:
         """Decomposition of :math:`\\text{CCZ}` gate.
 
         Decompose :math:`\\text{CCZ}` gate into :class:`qibo.gates.H` in
@@ -2468,6 +2665,129 @@ class DEUTSCH(ParametrizedGate):
         self.init_args = [q0, q1, q2]
         self.init_kwargs = {"theta": theta, "trainable": trainable}
 
+    @property
+    def hamming_weight(self):
+        return _is_hamming_weight_given_angle(self.parameters[0], np.pi)
+
+
+class FanOut(Gate):
+    """The :math:`\\textrm{FanOut}` gate.
+
+    Given a register containing :math:`k \\geq 2` qubits, this gate is equivalent
+    to :math:`k - 1` :class:`qibo.gates.CNOT` gates in which the control qubit
+    for all CNOTs is the first qubit in the register while the each CNOT has
+    each of the remaining qubits in the register as its target qubit.
+
+    Args:
+        *q (int): Ids of the qubits which the gate acts on.
+            The first id is considered to be the control qubit.
+    """
+
+    def __init__(self, *q):
+        if len(q) < 2:
+            raise_error(
+                ValueError, "``FanOut`` gate must be applied to at least two qubits."
+            )
+
+        super().__init__()
+        self.name = "fanout"
+        self.draw_label = "X"
+        self.control_qubits = (q[0],)
+        self.target_qubits = q[1:]
+        self.init_args = [*q]
+        self.unitary = True
+
+    @property
+    def clifford(self):
+        return True
+
+    def _base_decompose(self, *free, use_toffolis=True) -> List[Gate]:
+        """Decomposition of :math:`\\text{FanOut}` gate.
+
+        Decompose :math:`\\text{FanOut}` gate :class:`qibo.gates.CNOT`s.
+        """
+        from qibo.transpiler.decompositions import (  # pylint: disable=C0415
+            standard_decompositions,
+        )
+
+        return standard_decompositions(self)
+
+
+class GeneralizedRBS(ParametrizedGate):
+    """The generalized (complex) Reconfigurable Beam Splitter gate (:math:`\\text{gRBS}`).
+
+    Given a register called ``qubits_in`` containing :math:`m` qubits and a
+    register named ``qubits_out`` containing :math:`m'` qubits, the :math:`\\text{gRBS}`
+    is a :math:`(m + m')`-qubit gate that has the following matrix representation:
+
+    .. math::
+
+        \\begin{pmatrix}
+            I &           &  &             &  \\\\
+              & e^{-i\\phi}\\cos\\theta &      & e^{-i\\phi}\\sin\\theta   &  \\\\
+              &  & I'       &    &  \\\\
+              & -e^{i\\phi}\\sin\\theta &      & e^{i\\phi}\\cos\\theta   &  \\\\
+              &           &  &             &  I\\\\
+        \\end{pmatrix} \\,\\, ,
+
+    where :math:`I` and :math:`I'` are, respectively, identity matrices of size
+    :math:`2^{m} - 1` and :math:`2^{m}(2^{m'} - 2)`.
+
+    This unitary matrix is also known as a
+    `Givens rotation <https://en.wikipedia.org/wiki/Givens_rotation>`_.
+
+    References:
+        1. R. M. S. Farias, T. O. Maciel, G. Camilo, R. Lin, S. Ramos-Calderer, and L. Aolita,
+        *Quantum encoder for fixed-Hamming-weight subspaces*,
+        `Phys. Rev. Applied 23, 044014 (2025) <https://doi.org/10.1103/PhysRevApplied.23.044014>`_
+
+    Args:
+        qubits_in (tuple or list): ids of "input" qubits.
+        qubits_out (tuple or list): ids of "output" qubits.
+        theta (float): the rotation angle.
+        phi (float): the phase angle. Defaults to :math:`0.0`.
+        trainable (bool): whether gate parameter can be updated using
+            :meth:`qibo.models.circuit.AbstractCircuit.set_parameters`.
+            Defaults to ``True``.
+    """
+
+    def __init__(
+        self,
+        qubits_in: Union[Tuple[int], List[int]],
+        qubits_out: Union[Tuple[int], List[int]],
+        theta: float,
+        phi: float = 0.0,
+        trainable: bool = True,
+    ):
+        super().__init__(trainable)
+        self.name = "grbs"
+        self.draw_label = "gRBS"
+        self.target_qubits = tuple(qubits_in) + tuple(qubits_out)
+        self.unitary = True
+
+        self.parameter_names = ["theta", "phi"]
+        self.parameters = theta, phi
+        self.nparams = 2
+
+        self.init_args = [qubits_in, qubits_out]
+        self.init_kwargs = {"theta": theta, "phi": phi, "trainable": trainable}
+
+    @property
+    def hamming_weight(self):
+        return len(self.init_args[0]) == len(self.init_args[1])
+
+    def _base_decompose(self, *free, use_toffolis=True) -> List[Gate]:
+        """Decomposition of :math:`\\text{gRBS}` gate.
+
+        Decompose :math:`\\text{gRBS}` gate into :class:`qibo.gates.X`, :class:`qibo.gates.CNOT`,
+        :class:`qibo.gates.RY`, and :class:`qibo.gates.RZ`.
+        """
+        from qibo.transpiler.decompositions import (  # pylint: disable=C0415
+            standard_decompositions,
+        )
+
+        return standard_decompositions(self)
+
 
 class Unitary(ParametrizedGate):
     """Arbitrary unitary gate.
@@ -2487,12 +2807,18 @@ class Unitary(ParametrizedGate):
     """
 
     def __init__(
-        self, unitary, *q, trainable=True, name: str = None, check_unitary: bool = True
+        self,
+        unitary,
+        *q,
+        trainable=True,
+        name: str = None,
+        check_unitary: bool = True,
     ):
         super().__init__(trainable)
         self.name = "Unitary" if name is None else name
         self.draw_label = "U"
         self.target_qubits = tuple(q)
+        self._hamming_weight = False
         self._clifford = False
 
         # TODO: Check that given ``unitary`` has proper shape?
@@ -2508,24 +2834,12 @@ class Unitary(ParametrizedGate):
         }
 
         if check_unitary:
-            if unitary.__class__.__name__ == "Tensor":
-                import torch  # pylint: disable=C0145
-
-                diag_function = torch.diag
-                all_function = torch.all
-                conj_function = torch.conj
-                transpose_function = torch.transpose
-            else:
-                diag_function = np.diag
-                all_function = np.all
-                conj_function = np.conj
-                transpose_function = np.transpose
-
-            product = transpose_function(conj_function(unitary), (1, 0)) @ unitary
-            diagonals = all(np.abs(1 - diag_function(product)) < PRECISION_TOL)
+            engine = _check_engine(unitary)
+            product = engine.transpose(engine.conj(unitary), (1, 0)) @ unitary
+            diagonals = all(engine.abs(1 - engine.diag(product)) < PRECISION_TOL)
             off_diagonals = bool(
-                all_function(
-                    np.abs(product - diag_function(diag_function(product)))
+                engine.all(
+                    engine.abs(product - engine.diag(engine.diag(product)))
                     < PRECISION_TOL
                 )
             )
@@ -2536,7 +2850,11 @@ class Unitary(ParametrizedGate):
     @Gate.parameters.setter
     def parameters(self, x):
         shape = self.parameters[0].shape
-        self._parameters = (np.reshape(x, shape),)
+        engine = _check_engine(x)
+        # Reshape doesn't accept a tuple if engine is pytorch.
+        if isinstance(x, tuple):
+            x = x[0]
+        self._parameters = (engine.reshape(x, shape),)
         for gate in self.device_gates:  # pragma: no cover
             gate.parameters = x
 
@@ -2548,7 +2866,15 @@ class Unitary(ParametrizedGate):
     def clifford(self, value):
         self._clifford = value
 
-    def on_qubits(self, qubit_map):
+    @property
+    def hamming_weight(self):
+        return self._hamming_weight
+
+    @hamming_weight.setter
+    def hamming_weight(self, value):
+        self._hamming_weight = value
+
+    def on_qubits(self, qubit_map: dict):
         args = [self.init_args[0]]
         args.extend(qubit_map.get(i) for i in self.target_qubits)
         gate = self.__class__(*args, **self.init_kwargs)
@@ -2559,5 +2885,18 @@ class Unitary(ParametrizedGate):
         return gate
 
     def _dagger(self):
-        ud = np.conj(np.transpose(self.parameters[0]))
+        engine = _check_engine(self.parameters[0])
+        ud = engine.conj(self.parameters[0].T)
         return self.__class__(ud, *self.target_qubits, **self.init_kwargs)
+
+
+def _check_engine(array):
+    """Check if the array is a numpy or torch tensor and return the corresponding library."""
+    if (array.__class__.__name__ == "Tensor") or (
+        isinstance(array, tuple) and array[0].__class__.__name__ == "Tensor"
+    ):
+        import torch  # pylint: disable=C0415
+
+        return torch
+
+    return np
