@@ -12,7 +12,7 @@ class CallbackGate(SpecialGate):
         callback (:class:`qibo.callbacks.Callback`): Callback object to calculate.
     """
 
-    def __init__(self, callback: "Callback"):
+    def __init__(self, callback: "Callback"):  # type: ignore
         super().__init__()
         self.name = callback.__class__.__name__
         self.draw_label = "".join([c for c in self.name if c.isupper()])
@@ -99,7 +99,7 @@ class FusedGate(SpecialGate):
         """Returns matrix representation of special gate.
 
         Args:
-            backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used in the execution. If ``None``, it uses :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used in the execution. If ``None``, it uses the current backend. Defaults to ``None``.
 
         Returns:
             ndarray: Matrix representation of special gate.
