@@ -11,11 +11,11 @@ def convert_bit_to_energy(hamiltonian, bitstring):
     make sure the bitstring is of the right length
     """
     n = len(bitstring)
-    c = Circuit(n)
+    circuit = Circuit(n)
     active_bit = [i for i in range(n) if bitstring[i] == "1"]
     for i in active_bit:
-        c.add(gates.X(i))
-    result = c()  # this is an execution result, a quantum state
+        circuit.add(gates.X(i))
+    result = circuit()  # this is an execution result, a quantum state
     return hamiltonian.expectation(result.state())
 
 
