@@ -58,7 +58,7 @@ def optimize(
         .. testcode::
 
             import numpy as np
-            from qibo import gates, models
+            from qibo import Circuit, gates
             from qibo.optimizers import optimize
 
             # create custom loss function
@@ -68,7 +68,7 @@ def optimize(
                 return np.square(np.sum(circuit().state())) # returns numpy array
 
             # create circuit ansatz for two qubits
-            circuit = models.Circuit(2)
+            circuit = Circuit(2)
             circuit.add(gates.RY(0, theta=0))
 
             # optimize using random initial variational parameters
@@ -78,24 +78,21 @@ def optimize(
             # set parameters to circuit
             circuit.set_parameters(params)
     """
+    from qibo.backends import _check_backend
+
+    backend = _check_backend(backend)
     if method == "cma":
         if bounds is not None:  # pragma: no cover
             raise_error(
                 RuntimeError,
                 "The keyword 'bounds' cannot be used with the cma optimizer. Please use 'options' instead as defined by the cma documentation: ex. options['bounds'] = [0.0, 1.0].",
             )
-        return cmaes(loss, initial_parameters, args, callback, options)
+        return cmaes(
+            loss, backend.to_numpy(initial_parameters), args, callback, options
+        )
     elif method == "sgd":
-        from qibo.backends import _check_backend
-
-        backend = _check_backend(backend)
-
         return sgd(loss, initial_parameters, args, callback, options, compile, backend)
     else:
-        from qibo.backends import _check_backend
-
-        backend = _check_backend(backend)
-
         return newtonian(
             loss,
             initial_parameters,
@@ -208,13 +205,13 @@ def newtonian(
             callback=callback,
             options=options,
         )
-        m = o.run(initial_parameters)
+        m = o.run(backend.to_numpy(initial_parameters))
     else:
         from scipy.optimize import minimize
 
         m = minimize(
             loss,
-            initial_parameters,
+            backend.to_numpy(initial_parameters),
             args=args,
             method=method,
             jac=jac,
@@ -241,7 +238,9 @@ def sgd(
     """Stochastic Gradient Descent (SGD) optimizer using Tensorflow backpropagation.
 
     See `tf.keras.Optimizers <https://www.tensorflow.org/api_docs/python/tf/keras/optimizers>`_
-    for a list of the available optimizers.
+    for a list of the available optimizers for Tensorflow.
+    See `torch.optim <https://pytorch.org/docs/stable/optim.html>`_ for a list of the available
+    optimizers for PyTorch.
 
     Args:
         loss (callable): Loss as a function of variational parameters to be
@@ -260,9 +259,6 @@ def sgd(
               a message of the loss function.
     """
 
-    if not backend.name == "tensorflow":
-        raise_error(RuntimeError, "SGD optimizer requires Tensorflow backend.")
-
     sgd_options = {
         "nepochs": 1000000,
         "nmessage": 1000,
@@ -272,7 +268,53 @@ def sgd(
     if options is not None:
         sgd_options.update(options)
 
-    # proceed with the training
+    if backend.platform == "tensorflow":
+        return _sgd_tf(
+            loss,
+            initial_parameters,
+            args,
+            sgd_options,
+            compile,
+            backend,
+            callback=callback,
+        )
+
+    if backend.platform == "pytorch":
+        if compile:
+            log.warning(
+                "PyTorch does not support compilation of the optimization graph."
+            )
+        return _sgd_torch(
+            loss, initial_parameters, args, sgd_options, backend, callback=callback
+        )
+
+    raise_error(RuntimeError, "SGD optimizer requires Tensorflow or PyTorch backend.")
+
+
+def _sgd_torch(loss, initial_parameters, args, sgd_options, backend, callback=None):
+
+    vparams = initial_parameters
+    optimizer = getattr(backend.np.optim, sgd_options["optimizer"])(
+        params=[vparams], lr=sgd_options["learning_rate"]
+    )
+
+    for e in range(sgd_options["nepochs"]):
+        optimizer.zero_grad()
+        l = loss(vparams, *args)
+        l.backward()
+        optimizer.step()
+        if callback is not None:
+            callback(backend.to_numpy(vparams))
+        if e % sgd_options["nmessage"] == 1:
+            log.info("ite %d : loss %f", e, l.item())
+
+    return loss(vparams, *args).item(), vparams.detach().numpy(), sgd_options
+
+
+def _sgd_tf(
+    loss, initial_parameters, args, sgd_options, compile, backend, callback=None
+):
+
     vparams = backend.tf.Variable(initial_parameters)
     optimizer = getattr(backend.tf.optimizers, sgd_options["optimizer"])(
         learning_rate=sgd_options["learning_rate"]
