@@ -34,7 +34,7 @@ class Clifford:
             :class:`qibo.backends.CliffordBackend`. It accepts ``"numpy"``, ``"numba"``,
             ``"cupy"``, and ``"stim"`` (see `stim <https://github.com/quantumlib/Stim>`_).
             If ``None``, defaults to the corresponding engine
-            from :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            from the current backend. Defaults to ``None``.
     """
 
     symplectic_matrix: np.ndarray = field(init=False)
@@ -61,7 +61,10 @@ class Clifford:
             self.symplectic_matrix = self.data
             if self.symplectic_matrix.shape[0] % 2 == 0:
                 self.symplectic_matrix = np.vstack(
-                    (self.symplectic_matrix, np.zeros(self.symplectic_matrix.shape[1]))
+                    (
+                        self.symplectic_matrix,
+                        np.zeros(self.symplectic_matrix.shape[1], dtype=np.uint8),
+                    )
                 )
             self.nqubits = int((self.symplectic_matrix.shape[1] - 1) / 2)
         if self._backend is None:
@@ -89,7 +92,7 @@ class Clifford:
                 :class:`qibo.backends.CliffordBackend`. It accepts ``"numpy"``, ``"numba"``,
                 ``"cupy"``, and ``"stim"`` (see `stim <https://github.com/quantumlib/Stim>`_).
                 If ``None``, defaults to the corresponding engine
-                from :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+                from the current backend. Defaults to ``None``.
 
         Returns:
             (:class:`qibo.quantum_info.clifford.Clifford`): Object storing the result of the circuit execution.
@@ -258,7 +261,8 @@ class Clifford:
                     [
                         [p0.get(q) for q in measured_qubits],
                         [p1.get(q) for q in measured_qubits],
-                    ]
+                    ],
+                    dtype=self._backend.np.float64,
                 )
                 samples = self._backend.cast(samples, dtype="int32")
                 samples = self._backend.apply_bitflips(samples, bitflip_probabilities)
@@ -269,7 +273,7 @@ class Clifford:
             self._samples = self._backend.cast(samples, dtype="int32")
             for gate in self.measurements:
                 rqubits = tuple(qubit_map.get(q) for q in gate.target_qubits)
-                gate.result.register_samples(self._samples[:, rqubits], self._backend)
+                gate.result.register_samples(self._samples[:, rqubits])
 
         if registers:
             return {
@@ -297,16 +301,16 @@ class Clifford:
             of times each measured value/bitstring appears.
 
             If ``binary`` is ``True``
-                the keys of the `Counter` are in binary form, as strings of
-                :math:`0`s and :math`1`s.
+                the keys of the :class:`collections.Counter` are in binary form,
+                as strings of :math:`0` and :math`1`.
             If ``binary`` is ``False``
-                the keys of the ``Counter`` are integers.
+                the keys of the :class:`collections.Counter` are integers.
             If ``registers`` is ``True``
-                a `dict` of `Counter` s is returned where keys are the name of
-                each register.
+                a `dict` of :class:`collections.Counter` is returned where keys are
+                the name of each register.
             If ``registers`` is ``False``
-                a single ``Counter`` is returned which contains samples from all
-                the measured qubits, independently of their registers.
+                a single :class:`collections.Counter` is returned which contains samples
+                from all the measured qubits, independently of their registers.
         """
         measured_qubits = self.measurement_gate.target_qubits
         freq = self._backend.calculate_frequencies(self.samples(False))
