@@ -3,11 +3,11 @@
 from typing import Union
 
 import numpy as np
-from scipy.linalg import fractional_matrix_power
 
 from qibo.backends import _check_backend
 from qibo.config import PRECISION_TOL, raise_error
-from qibo.quantum_info.metrics import _check_hermitian_or_not_gpu, purity
+from qibo.quantum_info.linalg_operations import matrix_power, partial_trace
+from qibo.quantum_info.metrics import purity
 
 
 def shannon_entropy(prob_dist, base: float = 2, backend=None):
@@ -24,11 +24,11 @@ def shannon_entropy(prob_dist, base: float = 2, backend=None):
         prob_dist (ndarray or list): a probability array :math:`\\mathbf{p}`.
         base (float): the base of the log. Defaults to  :math:`2`.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
-            in the execution. If ``None``, it uses :class:`qibo.backends.GlobalBackend`.
+            in the execution. If ``None``, it uses the current backend.
             Defaults to ``None``.
 
     Returns:
-        (float): Shannon entropy :math:`H(\\mathcal{p})`.
+        float: Shannon entropy :math:`H(\\mathcal{p})`.
     """
     backend = _check_backend(backend)
 
@@ -56,17 +56,19 @@ def shannon_entropy(prob_dist, base: float = 2, backend=None):
 
     total_sum = backend.np.sum(prob_dist)
 
-    if np.abs(total_sum - 1.0) > PRECISION_TOL:
+    if np.abs(float(total_sum) - 1.0) > PRECISION_TOL:
         raise_error(ValueError, "Probability array must sum to 1.")
 
-    log_prob = np.where(prob_dist != 0, np.log2(prob_dist) / np.log2(base), 0.0)
+    log_prob = backend.np.where(
+        prob_dist != 0, backend.np.log2(prob_dist) / np.log2(base), 0.0
+    )
 
     shan_entropy = -backend.np.sum(prob_dist * log_prob)
 
     # absolute value if entropy == 0.0 to avoid returning -0.0
-    shan_entropy = np.abs(shan_entropy) if shan_entropy == 0.0 else shan_entropy
+    shan_entropy = backend.np.abs(shan_entropy) if shan_entropy == 0.0 else shan_entropy
 
-    return complex(shan_entropy).real
+    return np.real(float(shan_entropy))
 
 
 def classical_relative_entropy(prob_dist_p, prob_dist_q, base: float = 2, backend=None):
@@ -74,7 +76,7 @@ def classical_relative_entropy(prob_dist_p, prob_dist_q, base: float = 2, backen
 
     For probabilities :math:`\\mathbf{p}` and :math:`\\mathbf{q}`, it is defined as
 
-    ..math::
+    .. math::
         D(\\mathbf{p} \\, \\| \\, \\mathbf{q}) = \\sum_{x} \\, \\mathbf{p}(x) \\,
             \\log\\left( \\frac{\\mathbf{p}(x)}{\\mathbf{q}(x)} \\right) \\, .
 
@@ -87,19 +89,14 @@ def classical_relative_entropy(prob_dist_p, prob_dist_q, base: float = 2, backen
         base (float): the base of the log. Defaults to  :math:`2`.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be
             used in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
         float: Classical relative entropy between :math:`\\mathbf{p}` and :math:`\\mathbf{q}`.
     """
     backend = _check_backend(backend)
-
-    if isinstance(prob_dist_p, list):
-        # np.float64 is necessary instead of native float because of tensorflow
-        prob_dist_p = backend.cast(prob_dist_p, dtype=np.float64)
-    if isinstance(prob_dist_q, list):
-        # np.float64 is necessary instead of native float because of tensorflow
-        prob_dist_q = backend.cast(prob_dist_q, dtype=np.float64)
+    prob_dist_p = backend.cast(prob_dist_p, dtype=np.float64)
+    prob_dist_q = backend.cast(prob_dist_q, dtype=np.float64)
 
     if (len(prob_dist_p.shape) != 1) or (len(prob_dist_q.shape) != 1):
         raise_error(
@@ -125,25 +122,59 @@ def classical_relative_entropy(prob_dist_p, prob_dist_q, base: float = 2, backen
 
     total_sum_q = backend.np.sum(prob_dist_q)
 
-    if np.abs(total_sum_p - 1.0) > PRECISION_TOL:
+    if np.abs(float(total_sum_p) - 1.0) > PRECISION_TOL:
         raise_error(ValueError, "First probability array must sum to 1.")
 
-    if np.abs(total_sum_q - 1.0) > PRECISION_TOL:
+    if np.abs(float(total_sum_q) - 1.0) > PRECISION_TOL:
         raise_error(ValueError, "Second probability array must sum to 1.")
 
     entropy_p = -1 * shannon_entropy(prob_dist_p, base=base, backend=backend)
 
-    log_prob_q = np.where(
-        prob_dist_q != 0.0, np.log2(prob_dist_q) / np.log2(base), -np.inf
+    log_prob_q = backend.np.where(
+        prob_dist_q != 0.0, backend.np.log2(prob_dist_q) / np.log2(base), -np.inf
     )
 
-    log_prob = np.where(prob_dist_p != 0.0, log_prob_q, 0.0)
+    log_prob = backend.np.where(prob_dist_p != 0.0, log_prob_q, 0.0)
 
     relative = backend.np.sum(prob_dist_p * log_prob)
 
     return entropy_p - relative
 
 
+def classical_mutual_information(
+    prob_dist_joint, prob_dist_p, prob_dist_q, base: float = 2, backend=None
+):
+    """Calculates the classical mutual information of two random variables.
+
+    Given two random variables :math:`(X, \\, Y)`, their mutual information is given by
+
+    .. math::
+        I(X, \\, Y) \\equiv H(p(x)) + H(q(y)) - H(p(x, \\, y)) \\, ,
+
+    where :math:`p(x, \\, y)` is the joint probability distribution of :math:`(X, Y)`,
+    :math:`p(x)` is the marginal probability distribution of :math:`X`,
+    :math:`q(y)` is the marginal probability distribution of :math:`Y`,
+    and :math:`H(\\cdot)` is the :func:`qibo.quantum_info.entropies.shannon_entropy`.
+
+    Args:
+        prob_dist_joint (ndarray): joint probability distribution :math:`p(x, \\, y)`.
+        prob_dist_p (ndarray): marginal probability distribution :math:`p(x)`.
+        prob_dist_q (ndarray): marginal probability distribution :math:`q(y)`.
+        base (float): the base of the log. Defaults to  :math:`2`.
+        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
+            in the execution. If ``None``, it uses the current backend.
+            Defaults to ``None``.
+
+    Returns:
+        float: Mutual information :math:`I(X, \\, Y)`.
+    """
+    return (
+        shannon_entropy(prob_dist_p, base, backend)
+        + shannon_entropy(prob_dist_q, base, backend)
+        - shannon_entropy(prob_dist_joint, base, backend)
+    )
+
+
 def classical_renyi_entropy(
     prob_dist, alpha: Union[float, int], base: float = 2, backend=None
 ):
@@ -175,16 +206,13 @@ def classical_renyi_entropy(
         base (float): the base of the log. Defaults to  :math:`2`.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be
             used in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
         float: Classical Rényi entropy :math:`H_{\\alpha}`.
     """
     backend = _check_backend(backend)
-
-    if isinstance(prob_dist, list):
-        # np.float64 is necessary instead of native float because of tensorflow
-        prob_dist = backend.cast(prob_dist, dtype=np.float64)
+    prob_dist = backend.cast(prob_dist, dtype=np.float64)
 
     if not isinstance(alpha, (float, int)):
         raise_error(
@@ -214,7 +242,7 @@ def classical_renyi_entropy(
 
     total_sum = backend.np.sum(prob_dist)
 
-    if np.abs(total_sum - 1.0) > PRECISION_TOL:
+    if np.abs(float(total_sum) - 1.0) > PRECISION_TOL:
         raise_error(ValueError, "Probability array must sum to 1.")
 
     if alpha == 0.0:
@@ -224,11 +252,11 @@ def classical_renyi_entropy(
         return shannon_entropy(prob_dist, base=base, backend=backend)
 
     if alpha == np.inf:
-        return -1 * np.log2(max(prob_dist)) / np.log2(base)
+        return -1 * backend.np.log2(max(prob_dist)) / np.log2(base)
 
     total_sum = backend.np.sum(prob_dist**alpha)
 
-    renyi_ent = (1 / (1 - alpha)) * np.log2(total_sum) / np.log2(base)
+    renyi_ent = (1 / (1 - alpha)) * backend.np.log2(total_sum) / np.log2(base)
 
     return renyi_ent
 
@@ -267,19 +295,14 @@ def classical_relative_renyi_entropy(
         base (float): the base of the log. Defaults to  :math:`2`.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be
             used in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
         float: Classical relative Rényi entropy :math:`H_{\\alpha}(\\mathbf{p} \\, \\| \\, \\mathbf{q})`.
     """
     backend = _check_backend(backend)
-
-    if isinstance(prob_dist_p, list):
-        # np.float64 is necessary instead of native float because of tensorflow
-        prob_dist_p = backend.cast(prob_dist_p, dtype=np.float64)
-    if isinstance(prob_dist_q, list):
-        # np.float64 is necessary instead of native float because of tensorflow
-        prob_dist_q = backend.cast(prob_dist_q, dtype=np.float64)
+    prob_dist_p = backend.cast(prob_dist_p, dtype=np.float64)
+    prob_dist_q = backend.cast(prob_dist_q, dtype=np.float64)
 
     if (len(prob_dist_p.shape) != 1) or (len(prob_dist_q.shape) != 1):
         raise_error(
@@ -313,17 +336,17 @@ def classical_relative_renyi_entropy(
     total_sum_p = backend.np.sum(prob_dist_p)
     total_sum_q = backend.np.sum(prob_dist_q)
 
-    if np.abs(total_sum_p - 1.0) > PRECISION_TOL:
+    if np.abs(float(total_sum_p) - 1.0) > PRECISION_TOL:
         raise_error(ValueError, "First probability array must sum to 1.")
 
-    if np.abs(total_sum_q - 1.0) > PRECISION_TOL:
+    if np.abs(float(total_sum_q) - 1.0) > PRECISION_TOL:
         raise_error(ValueError, "Second probability array must sum to 1.")
 
     if alpha == 0.5:
-        total_sum = np.sqrt(prob_dist_p * prob_dist_q)
+        total_sum = backend.np.sqrt(prob_dist_p * prob_dist_q)
         total_sum = backend.np.sum(total_sum)
 
-        return -2 * np.log2(total_sum) / np.log2(base)
+        return -2 * backend.np.log2(total_sum) / np.log2(base)
 
     if alpha == 1.0:
         return classical_relative_entropy(
@@ -331,14 +354,14 @@ def classical_relative_renyi_entropy(
         )
 
     if alpha == np.inf:
-        return np.log2(max(prob_dist_p / prob_dist_q)) / np.log2(base)
+        return backend.np.log2(max(prob_dist_p / prob_dist_q)) / np.log2(base)
 
     prob_p = prob_dist_p**alpha
     prob_q = prob_dist_q ** (1 - alpha)
 
     total_sum = backend.np.sum(prob_p * prob_q)
 
-    return (1 / (alpha - 1)) * np.log2(total_sum) / np.log2(base)
+    return (1 / (alpha - 1)) * backend.np.log2(total_sum) / np.log2(base)
 
 
 def classical_tsallis_entropy(prob_dist, alpha: float, base: float = 2, backend=None):
@@ -357,7 +380,7 @@ def classical_tsallis_entropy(prob_dist, alpha: float, base: float = 2, backend=
             Defaults to  :math:`2`.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be
             used in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
         float: Classical Tsallis entropy :math:`S_{\\alpha}(\\mathbf{p})`.
@@ -396,7 +419,7 @@ def classical_tsallis_entropy(prob_dist, alpha: float, base: float = 2, backend=
 
     total_sum = backend.np.sum(prob_dist)
 
-    if np.abs(total_sum - 1.0) > PRECISION_TOL:
+    if np.abs(float(total_sum) - 1.0) > PRECISION_TOL:
         raise_error(ValueError, "Probability array must sum to 1.")
 
     if alpha == 1.0:
@@ -405,35 +428,81 @@ def classical_tsallis_entropy(prob_dist, alpha: float, base: float = 2, backend=
     total_sum = prob_dist**alpha
     total_sum = backend.np.sum(total_sum)
 
-    return (1 / (1 - alpha)) * (total_sum - 1)
+    return (1 / (alpha - 1)) * (1 - total_sum)
+
+
+def classical_relative_tsallis_entropy(
+    prob_dist_p, prob_dist_q, alpha: float, base: float = 2, backend=None
+):
+    """Calculate the classical relative Tsallis entropy between two discrete probability distributions.
+
+    Given a discrete random variable :math:`\\chi` that has values :math:`x` in the set
+    :math:`\\mathcal{X}` with probability :math:`\\mathrm{p}(x)` and a discrete random variable
+    :math:`\\upsilon` that has the values :math:`x` in the same set :math:`\\mathcal{X}` with
+    probability :math:`\\mathrm{q}(x)`, their relative Tsallis entropy is given by
+
+    .. math::
+        D_{\\alpha}^{\\text{ts}}(\\chi \\, \\| \\, \\upsilon) = \\sum_{x \\in \\mathcal{X}} \\,
+            \\mathrm{p}^{\\alpha}(x) \\, \\ln_{\\alpha}
+            \\left( \\frac{\\mathrm{p}(x)}{\\mathrm{q}(x)} \\right) \\, ,
+
+    where :math:`\\ln_{\\alpha}(x) \\equiv \\frac{x^{1 - \\alpha} - 1}{1 - \\alpha}`
+    is the so-called :math:`\\alpha`-logarithm. When :math:`\\alpha = 1`, it reduces to
+    :class:`qibo.quantum_info.entropies.classical_relative_entropy`.
+
+    Args:
+        prob_dist_p (ndarray or list): discrete probability distribution :math:`p`.
+        prob_dist_q (ndarray or list): discrete probability distribution :math:`q`.
+        alpha (float): entropic index.
+        base (float): the base of the log used when :math:`\\alpha = 1`. Defaults to :math:`2`.
+        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be
+            used in the execution. If ``None``, it uses
+            the current backend. Defaults to ``None``.
+
+    Returns:
+        float: Tsallis relative entropy :math:`D_{\\alpha}^{\\text{ts}}`.
+    """
+    if alpha == 1.0:
+        return classical_relative_entropy(prob_dist_p, prob_dist_q, base, backend)
+
+    backend = _check_backend(backend)
+
+    if isinstance(prob_dist_p, list):
+        # np.float64 is necessary instead of native float because of tensorflow
+        prob_dist_p = backend.cast(prob_dist_p, dtype=np.float64)
+
+    if isinstance(prob_dist_q, list):
+        # np.float64 is necessary instead of native float because of tensorflow
+        prob_dist_q = backend.cast(prob_dist_q, dtype=np.float64)
+
+    element_wise = prob_dist_p**alpha
+    element_wise = element_wise * _q_logarithm(prob_dist_p / prob_dist_q, alpha)
+
+    return backend.np.sum(element_wise)
 
 
 def von_neumann_entropy(
     state,
     base: float = 2,
-    check_hermitian: bool = False,
     return_spectrum: bool = False,
     backend=None,
 ):
-    """Calculates the von-Neumann entropy :math:`S(\\rho)` of a quantum ``state`` :math:`\\rho`.
+    """Calculate the von-Neumann entropy :math:`S(\\rho)` of a quantum ``state`` :math:`\\rho`.
 
     It is given by
 
     .. math::
-        S(\\rho) = - \\text{tr}\\left[\\rho \\, \\log(\\rho)\\right]
+        S(\\rho) = - \\text{tr}\\left(\\rho \\, \\log(\\rho)\\right)
 
     Args:
         state (ndarray): statevector or density matrix.
         base (float, optional): the base of the log. Defaults to :math:`2`.
-        check_hermitian (bool, optional): if ``True``, checks if ``state`` is Hermitian.
-            If ``False``, it assumes ``state`` is Hermitian .
-            Defaults to ``False``.
         return_spectrum: if ``True``, returns ``entropy`` and
             :math:`-\\log_{\\textup{b}}(\\textup{eigenvalues})`, where :math:`b` is ``base``.
             If ``False``, returns only ``entropy``. Default is ``False``.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
             in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
         float: The von-Neumann entropy :math:`S` of ``state`` :math:`\\rho`.
@@ -453,64 +522,61 @@ def von_neumann_entropy(
     if base <= 0.0:
         raise_error(ValueError, "log base must be non-negative.")
 
-    if not isinstance(check_hermitian, bool):
-        raise_error(
-            TypeError,
-            f"check_hermitian must be type bool, but it is type {type(check_hermitian)}.",
-        )
-
-    if purity(state) == 1.0:
+    if purity(state, backend=backend) == 1.0:
         if return_spectrum:
             return 0.0, backend.cast([0.0], dtype=float)
 
         return 0.0
 
-    if not check_hermitian or _check_hermitian_or_not_gpu(state, backend=backend):
-        eigenvalues = np.linalg.eigvalsh(state)
-    else:
-        eigenvalues = np.linalg.eigvals(state)
-
-    log_prob = np.where(eigenvalues > 0, np.log2(eigenvalues) / np.log2(base), 0.0)
-
-    ent = -np.sum(eigenvalues * log_prob)
-    # absolute value if entropy == 0.0 to avoid returning -0.0
-    ent = np.abs(ent) if ent == 0.0 else ent
-
-    ent = float(ent)
+    ent = backend.calculate_matrix_log(state, base=base)
+    ent = -backend.np.trace(state @ ent)
+    ent = backend.np.real(ent)
 
     if return_spectrum:
+        eigenvalues = backend.calculate_eigenvalues(state)
+
+        log_prob = backend.np.where(
+            backend.np.real(eigenvalues) > 0.0,
+            backend.np.log2(eigenvalues) / np.log2(base),
+            0.0,
+        )
         log_prob = backend.cast(log_prob, dtype=log_prob.dtype)
+
         return ent, -log_prob
 
     return ent
 
 
 def relative_von_neumann_entropy(
-    state, target, base: float = 2, check_hermitian: bool = False, backend=None
+    state,
+    target,
+    base: float = 2,
+    backend=None,
 ):
-    """Calculates the relative entropy :math:`S(\\rho \\, \\| \\, \\sigma)` between ``state`` :math:`\\rho` and ``target`` :math:`\\sigma`.
+    """Calculates the relative von Neumann entropy  between two quantum states.
 
-    It is given by
+    Also known as *quantum relative entropy*, :math:`S(\\rho \\, \\| \\, \\sigma)` is given by
 
     .. math::
-        S(\\rho \\, \\| \\, \\sigma) = \\text{tr}\\left[\\rho \\, \\log(\\rho)\\right]
-            - \\text{tr}\\left[\\rho \\, \\log(\\sigma)\\right]
+        S(\\rho \\, \\| \\, \\sigma) = \\text{tr}\\left(\\rho \\, \\log(\\rho)\\right)
+            - \\text{tr}\\left(\\rho \\, \\log(\\sigma)\\right)
+
+    where ``state`` :math:`\\rho` and ``target`` :math:`\\sigma` are two quantum states.
 
     Args:
         state (ndarray): statevector or density matrix :math:`\\rho`.
         target (ndarray): statevector or density matrix :math:`\\sigma`.
         base (float, optional): the base of the log. Defaults to :math:`2`.
-        check_hermitian (bool, optional): If ``True``, checks if ``state`` is Hermitian.
-            If ``False``, it assumes ``state`` is Hermitian .
-            Defaults to ``False``.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
             in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
-        float: Relative (von-Neumann) entropy :math:`S(\\rho \\, \\| \\, \\sigma)`.
+        float: Relative von-Neumann entropy :math:`S(\\rho \\, \\| \\, \\sigma)`.
     """
     backend = _check_backend(backend)
+    state = backend.cast(state)
+    target = backend.cast(target)
 
     if (
         (len(state.shape) >= 3)
@@ -535,45 +601,62 @@ def relative_von_neumann_entropy(
     if base <= 0.0:
         raise_error(ValueError, "log base must be non-negative.")
 
-    if not isinstance(check_hermitian, bool):
-        raise_error(
-            TypeError,
-            f"check_hermitian must be type bool, but it is type {type(check_hermitian)}.",
-        )
-
-    if purity(state) == 1.0 and purity(target) == 1.0:
+    if purity(state, backend=backend) == 1.0 and purity(target, backend=backend) == 1.0:
         return 0.0
 
     if len(state.shape) == 1:
-        state = np.outer(state, np.conj(state))
+        state = backend.np.outer(state, backend.np.conj(state))
 
     if len(target.shape) == 1:
-        target = np.outer(target, np.conj(target))
+        target = backend.np.outer(target, backend.np.conj(target))
 
-    if not check_hermitian or _check_hermitian_or_not_gpu(state, backend=backend):
-        eigenvalues_state = np.linalg.eigvalsh(state)
-    else:
-        eigenvalues_state = np.linalg.eigvals(state)
-
-    if not check_hermitian or _check_hermitian_or_not_gpu(target, backend=backend):
-        eigenvalues_target = np.linalg.eigvalsh(target)
-    else:
-        eigenvalues_target = np.linalg.eigvals(target)
-
-    log_state = np.where(
-        eigenvalues_state > 0, np.log2(eigenvalues_state) / np.log2(base), 0.0
-    )
-    log_target = np.where(
-        eigenvalues_target > 0, np.log2(eigenvalues_target) / np.log2(base), -np.inf
+    relative = backend.calculate_matrix_log(target, base=base)
+    relative = -backend.np.trace(state @ relative)
+    relative -= von_neumann_entropy(
+        state, base=base, return_spectrum=False, backend=backend
     )
 
-    log_target = np.where(eigenvalues_state != 0.0, log_target, 0.0)
+    return backend.np.real(relative)
 
-    entropy_state = np.sum(eigenvalues_state * log_state)
 
-    relative = np.sum(eigenvalues_state * log_target)
+def mutual_information(state, partition, base: float = 2, backend=None):
+    """Calculates the mutual information of a bipartite state.
 
-    return float(entropy_state - relative)
+    Given a qubit ``partition`` :math:`A`, the mutual information
+    of state :math:`\\rho` is given by
+
+    .. math::
+        I(\\rho) \\equiv S(\\rho_{A}) + S(\\rho_{B}) - S(\\rho) \\, ,
+
+    where :math:`B` is the remaining qubits that are not in partition :math:`A`,
+    and :math:`S(\\cdot)` is the :func:`qibo.quantum_info.von_neumann_entropy`.
+
+    Args:
+        state (ndarray): statevector or density matrix.
+        partition (Union[List[int], Tuple[int]]): indices of qubits in partition :math:`A`.
+        base (float, optional): the base of the log. Defaults to :math:`2`.
+        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
+            in the execution. If ``None``, it uses
+            the current backend. Defaults to ``None``.
+
+    Returns:
+        float: Mutual information :math:`I(\\rho)` of ``state`` :math:`\\rho`.
+    """
+    nqubits = np.log2(len(state))
+
+    if not nqubits.is_integer():
+        raise_error(ValueError, f"dimensions of ``state`` must be a power of 2.")
+
+    partition_b = set(list(range(int(nqubits)))) ^ set(list(partition))
+
+    state_a = partial_trace(state, partition_b, backend)
+    state_b = partial_trace(state, partition, backend)
+
+    return backend.np.real(
+        von_neumann_entropy(state_a, base, False, backend)
+        + von_neumann_entropy(state_b, base, False, backend)
+        - von_neumann_entropy(state, base, False, backend)
+    )
 
 
 def renyi_entropy(state, alpha: Union[float, int], base: float = 2, backend=None):
@@ -604,7 +687,7 @@ def renyi_entropy(state, alpha: Union[float, int], base: float = 2, backend=None
         base (float): the base of the log. Defaults to  :math:`2`.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be
             used in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
         float: Rényi entropy :math:`H_{\\alpha}`.
@@ -632,7 +715,7 @@ def renyi_entropy(state, alpha: Union[float, int], base: float = 2, backend=None
     if base <= 0.0:
         raise_error(ValueError, "log base must be non-negative.")
 
-    if abs(purity(state) - 1.0) < PRECISION_TOL:
+    if abs(purity(state, backend=backend) - 1.0) < PRECISION_TOL:
         return 0.0
 
     if alpha == 0.0:
@@ -644,11 +727,11 @@ def renyi_entropy(state, alpha: Union[float, int], base: float = 2, backend=None
     if alpha == np.inf:
         return (
             -1
-            * np.log2(backend.calculate_norm_density_matrix(state, order=2))
+            * backend.np.log2(backend.calculate_matrix_norm(state, order=2))
             / np.log2(base)
         )
 
-    log = np.log2(np.trace(_matrix_power(state, alpha, backend)))
+    log = backend.np.log2(backend.np.trace(matrix_power(state, alpha, backend=backend)))
 
     return (1 / (1 - alpha)) * log / np.log2(base)
 
@@ -667,7 +750,7 @@ def relative_renyi_entropy(
             \\sigma^{1 - \\alpha} \\right) \\right) \\, .
 
     A special case is the limit :math:`\\alpha \\to 1`, in which the Rényi entropy
-    coincides with the :func:`qibo.quantum_info.entropies.relative_entropy`.
+    coincides with the :func:`qibo.quantum_info.entropies.relative_von_neumann_entropy`.
 
     In the limit :math:`\\alpha \\to \\infty`, the function reduces to
     :math:`-2 \\, \\log(\\|\\sqrt{\\rho} \\, \\sqrt{\\sigma}\\|_{1})`,
@@ -676,7 +759,7 @@ def relative_renyi_entropy(
     This is known as the `min-relative entropy <https://arxiv.org/abs/1310.7178>`_.
 
     .. note::
-        Function raises ``NotImplementedError`` when ``target`` :math:`sigma`
+        Function raises ``NotImplementedError`` when ``target`` :math:`\\sigma`
         is a pure state and :math:`\\alpha > 1`. This is due to the fact that
         it is not possible to calculate :math:`\\sigma^{1 - \\alpha}` when
         :math:`\\alpha > 1` and :math:`\\sigma` is a projector, i.e. a singular matrix.
@@ -688,13 +771,14 @@ def relative_renyi_entropy(
         base (float): the base of the log. Defaults to  :math:`2`.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be
             used in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
         float: Relative Rényi entropy :math:`H_{\\alpha}(\\rho \\, \\| \\, \\sigma)`.
     """
     backend = _check_backend(backend)
-
+    state = backend.cast(state)
+    target = backend.cast(target)
     if (
         (len(state.shape) >= 3)
         or (len(state) == 0)
@@ -726,9 +810,9 @@ def relative_renyi_entropy(
     if base <= 0.0:
         raise_error(ValueError, "log base must be non-negative.")
 
-    purity_target = purity(target)
+    purity_target = purity(target, backend=backend)
     if (
-        abs(purity(state) - 1.0) < PRECISION_TOL
+        abs(purity(state, backend=backend) - 1.0) < PRECISION_TOL
         and abs(purity_target - 1) < PRECISION_TOL
     ):
         return 0.0
@@ -740,24 +824,24 @@ def relative_renyi_entropy(
         )
 
     if len(state.shape) == 1:
-        state = np.outer(state, np.conj(state))
+        state = backend.np.outer(state, backend.np.conj(state))
 
     if alpha == 1.0:
         return relative_von_neumann_entropy(state, target, base, backend=backend)
 
     if alpha == np.inf:
-        new_state = _matrix_power(state, 0.5, backend)
-        new_target = _matrix_power(target, 0.5, backend)
+        new_state = matrix_power(state, 0.5, backend=backend)
+        new_target = matrix_power(target, 0.5, backend=backend)
 
-        log = np.log2(
-            backend.calculate_norm_density_matrix(new_state @ new_target, order=1)
+        log = backend.np.log2(
+            backend.calculate_matrix_norm(new_state @ new_target, order=1)
         )
 
         return -2 * log / np.log2(base)
 
-    log = _matrix_power(state, alpha, backend)
-    log = log @ _matrix_power(target, 1 - alpha, backend)
-    log = np.log2(np.trace(log))
+    log = matrix_power(state, alpha, backend=backend)
+    log = log @ matrix_power(target, 1 - alpha, backend=backend)
+    log = backend.np.log2(backend.np.trace(log))
 
     return (1 / (alpha - 1)) * log / np.log2(base)
 
@@ -779,7 +863,7 @@ def tsallis_entropy(state, alpha: float, base: float = 2, backend=None):
             Defaults to :math:`2`.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
             in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
         float: Tsallis entropy :math:`S_{\\alpha}(\\rho)`.
@@ -807,20 +891,95 @@ def tsallis_entropy(state, alpha: float, base: float = 2, backend=None):
     if base <= 0.0:
         raise_error(ValueError, "log base must be non-negative.")
 
-    if abs(purity(state) - 1.0) < PRECISION_TOL:
+    if abs(purity(state, backend=backend) - 1.0) < PRECISION_TOL:
         return 0.0
 
     if alpha == 1.0:
         return von_neumann_entropy(state, base=base, backend=backend)
 
-    return (1 / (1 - alpha)) * (np.trace(_matrix_power(state, alpha, backend)) - 1)
+    return (1 / (1 - alpha)) * (
+        backend.np.trace(matrix_power(state, alpha, backend=backend)) - 1
+    )
+
+
+def relative_tsallis_entropy(
+    state,
+    target,
+    alpha: Union[float, int],
+    base: float = 2,
+    backend=None,
+):
+    """Calculate the relative Tsallis entropy between two quantum states.
+
+    For :math:`\\alpha \\in [0, \\, 2]` and quantum states :math:`\\rho` and
+    :math:`\\sigma`, the relative Tsallis entropy is defined as
+
+    .. math::
+        \\Delta_{\\alpha}^{\\text{ts}}(\\rho, \\, \\sigma) = \\frac{1 -
+            \\text{tr}\\left(\\rho^{\\alpha} \\, \\sigma^{1 - \\alpha}\\right)}{1 - \\alpha} \\, .
+
+    A special case is the limit :math:`\\alpha \\to 1`, in which the Tsallis entropy
+    coincides with the :func:`qibo.quantum_info.entropies.relative_von_neumann_entropy`.
+
+    Args:
+        state (ndarray): statevector or density matrix :math:`\\rho`.
+        target (ndarray): statevector or density matrix :math:`\\sigma`.
+        alpha (float or int): entropic index :math:`\\alpha \\in [0, \\, 2]`.
+        base (float, optional): the base of the log used when :math:`\\alpha = 1`.
+            Defaults to :math:`2`.
+        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
+            in the execution. If ``None``, it uses
+            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+
+    Returns:
+        float: Relative Tsallis entropy :math:`\\Delta_{\\alpha}^{\\text{ts}}`.
+
+    References:
+        1. S. Abe, *Nonadditive generalization of the quantum Kullback-Leibler
+        divergence for measuring the degree of purification*,
+        `Phys. Rev. A 68, 032302 <https://doi.org/10.1103/PhysRevA.68.032302>`_.
+
+        2. S. Furuichi, K. Yanagi, and K. Kuriyama,
+        *Fundamental properties of Tsallis relative entropy*,
+        `J. Math. Phys., Vol. 45, Issue 12, pp. 4868-4877 (2004)
+        <https://doi.org/10.1063/1.1805729>`_ .
+    """
+    if alpha == 1.0:
+        return relative_von_neumann_entropy(state, target, base=base, backend=backend)
+
+    if not isinstance(alpha, (float, int)):
+        raise_error(
+            TypeError,
+            f"``alpha`` must be type float or int, but it is type {type(alpha)}.",
+        )
+
+    if alpha < 0.0 or alpha > 2.0:
+        raise_error(
+            ValueError, f"``alpha`` must be in the interval [0, 2], but it is {alpha}."
+        )
+
+    if alpha < 1.0:
+        alpha = 2 - alpha
+
+    factor = 1 - alpha
+
+    if len(state.shape) == 1:
+        state = backend.np.outer(state, backend.np.conj(state.T))
+
+    if len(target.shape) == 1:
+        target = backend.np.outer(target, backend.np.conj(target.T))
+
+    trace = matrix_power(state, alpha, backend=backend)
+    trace = trace @ matrix_power(target, factor, backend=backend)
+    trace = backend.np.trace(trace)
+
+    return (1 - trace) / factor
 
 
 def entanglement_entropy(
     state,
     bipartition,
     base: float = 2,
-    check_hermitian: bool = False,
     return_spectrum: bool = False,
     backend=None,
 ):
@@ -837,13 +996,11 @@ def entanglement_entropy(
         state (ndarray): statevector or density matrix.
         bipartition (list or tuple or ndarray): qubits in the subsystem to be traced out.
         base (float, optional): the base of the log. Defaults to :math: `2`.
-        check_hermitian (bool, optional): if ``True``, checks if :math:`\\rho_{A}` is Hermitian.
-            If ``False``, it assumes ``state`` is Hermitian . Default: ``False``.
         return_spectrum: if ``True``, returns ``entropy`` and eigenvalues of ``state``.
             If ``False``, returns only ``entropy``. Default is ``False``.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
             in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
         float: Entanglement entropy :math:`S` of ``state`` :math:`\\rho`.
@@ -863,18 +1020,11 @@ def entanglement_entropy(
             f"state must have dims either (k,) or (k,k), but have dims {state.shape}.",
         )
 
-    nqubits = int(np.log2(state.shape[0]))
-
-    reduced_density_matrix = (
-        backend.partial_trace(state, bipartition, nqubits)
-        if len(state.shape) == 1
-        else backend.partial_trace_density_matrix(state, bipartition, nqubits)
-    )
+    reduced_density_matrix = partial_trace(state, bipartition, backend=backend)
 
     entropy_entanglement = von_neumann_entropy(
         reduced_density_matrix,
         base=base,
-        check_hermitian=check_hermitian,
         return_spectrum=return_spectrum,
         backend=backend,
     )
@@ -882,19 +1032,7 @@ def entanglement_entropy(
     return entropy_entanglement
 
 
-def _matrix_power(matrix, alpha, backend):
-    """Calculates ``matrix ** alpha`` according to backend."""
-    if backend.__class__.__name__ in [
-        "CupyBackend",
-        "CuQuantumBackend",
-    ]:  # pragma: no cover
-        new_matrix = backend.to_numpy(matrix)
-    else:
-        new_matrix = np.copy(matrix)
-
-    if len(new_matrix.shape) == 1:
-        new_matrix = np.outer(new_matrix, np.conj(new_matrix))
-
-    new_matrix = fractional_matrix_power(new_matrix, alpha)
-
-    return backend.cast(new_matrix, dtype=new_matrix.dtype)
+def _q_logarithm(x, q: float):
+    """Generalization of logarithm function necessary for classical (relative) Tsallis entropy."""
+    factor = 1 - q
+    return (x**factor - 1) / factor
