@@ -9,7 +9,7 @@ from qibo.backends import _check_backend
 from qibo.config import PRECISION_TOL, raise_error
 
 
-def purity(state):
+def purity(state, backend=None):
     """Purity of a quantum state :math:`\\rho`.
 
     This is given by
@@ -22,7 +22,7 @@ def purity(state):
     Returns:
         float: Purity of quantum ``state`` :math:`\\rho`.
     """
-
+    backend = _check_backend(backend)
     if (
         (len(state.shape) >= 3)
         or (len(state) == 0)
@@ -34,18 +34,13 @@ def purity(state):
         )
 
     if len(state.shape) == 1:
-        pur = np.abs(np.dot(np.conj(state), state)) ** 2
+        pur = backend.np.real(backend.calculate_vector_norm(state)) ** 2
     else:
-        pur = np.real(np.trace(np.dot(state, state)))
-
-    # this is necessary to remove the float from inside
-    # a 0-dim ndarray
-    pur = float(pur)
-
-    return pur
+        pur = backend.np.real(backend.np.trace(backend.np.matmul(state, state)))
+    return float(pur)
 
 
-def impurity(state):
+def impurity(state, backend=None):
     """Impurity of quantum state :math:`\\rho`.
 
     This is given by :math:`1 - \\text{purity}(\\rho)`, where :math:`\\text{purity}`
@@ -57,10 +52,10 @@ def impurity(state):
     Returns:
         float: impurity of ``state`` :math:`\\rho`.
     """
-    return 1 - purity(state)
+    return 1 - purity(state, backend=backend)
 
 
-def trace_distance(state, target, check_hermitian: bool = False, backend=None):
+def trace_distance(state, target, backend=None):
     """Trace distance between two quantum states, :math:`\\rho` and
     :math:`\\sigma`:
 
@@ -74,12 +69,8 @@ def trace_distance(state, target, check_hermitian: bool = False, backend=None):
     Args:
         state (ndarray): statevector or density matrix.
         target (ndarray): statevector or density matrix.
-        check_hermitian (bool, optional): if ``True``, checks if
-            :math:`\\rho - \\sigma` is Hermitian. If ``False``,
-            it assumes the difference is Hermitian.
-            Defaults to ``False``.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
-            in the execution. If ``None``, it uses :class:`qibo.backends.GlobalBackend`.
+            in the execution. If ``None``, it uses the current backend.
             Defaults to ``None``.
 
     Returns:
@@ -100,63 +91,70 @@ def trace_distance(state, target, check_hermitian: bool = False, backend=None):
             + f"but have dims {state.shape} and {target.shape}",
         )
 
-    if isinstance(check_hermitian, bool) is False:
-        raise_error(
-            TypeError,
-            f"check_hermitian must be type bool, but it is type {type(check_hermitian)}.",
-        )
-
     if len(state.shape) == 1:
-        state = np.outer(np.conj(state), state)
-        target = np.outer(np.conj(target), target)
+        state = backend.np.outer(backend.np.conj(state), state)
+        target = backend.np.outer(backend.np.conj(target), target)
 
-    difference = state - target
-    if check_hermitian is True:
-        hermitian = bool(
-            float(
-                backend.calculate_norm_density_matrix(
-                    np.transpose(np.conj(difference)) - difference, order=2
-                )
-            )
-            <= PRECISION_TOL
-        )
-        if (
-            not hermitian and backend.__class__.__name__ == "CupyBackend"
-        ):  # pragma: no cover
-            raise_error(
-                NotImplementedError,
-                "CupyBackend does not support `np.linalg.eigvals`"
-                + "for non-Hermitian `state - target`.",
-            )
-        eigenvalues = (
-            np.linalg.eigvalsh(difference)
-            if hermitian
-            else np.linalg.eigvals(difference)
-        )
-    else:
-        eigenvalues = np.linalg.eigvalsh(difference)
+    distance = state - target
+    distance = backend.np.conj(distance.T) @ distance
+    distance = backend.calculate_matrix_sqrt(distance)
 
-    distance = np.sum(np.absolute(eigenvalues)) / 2
-    distance = float(distance)
-
-    return distance
+    return backend.np.trace(distance) / 2
 
 
-def hilbert_schmidt_distance(state, target):
-    """Hilbert-Schmidt distance between two quantum states:
+def hilbert_schmidt_inner_product(operator_A, operator_B, backend=None):
+    """Calculate the Hilbert-Schmidt inner product between two operators.
+
+    Given two operators :math:`A, \\, B \\in \\mathcal{H}`, the Hilbert-Schmidt
+    inner product between the two is given by
 
     .. math::
-        \\langle \\rho \\, , \\, \\sigma \\rangle_{\\text{HS}} =
-            \\text{tr}\\left((\\rho - \\sigma)^{2}\\right)
+        \\braket{A, \\, B}_{\\text{HS}} = \\text{tr}\\left(A^{\\dagger} \\, B\\right) \\, .
+
+    Args:
+        operator_A (ndarray): operator :math:`A`.
+        operator_B (ndarray): operator :math:`B`.
+        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
+            in the execution. If ``None``, it uses the current backend.
+            Defaults to ``None``.
+
+    Returns:
+        float: Hilbert-Schmidt inner product :math:`\\braket{A, \\, B}_{\\text{HS}}`.
+    """
+    backend = _check_backend(backend)
+
+    inner_product = backend.np.trace(backend.np.conj(operator_A.T) @ operator_B)
+
+    return backend.np.real(inner_product)
+
+
+def hilbert_schmidt_distance(state, target, backend=None):
+    """Calculate the Hilbert-Schmidt distance between two quantum states:
+
+    .. math::
+        \\braket{\\rho - \\sigma, \\, \\rho - \\sigma}_{\\text{HS}} =
+            \\text{tr}\\left((\\rho - \\sigma)^{2}\\right) \\, ,
+
+    where :math:`\\braket{\\cdot, \\, \\cdot}_{\\text{HS}}` is the
+    :func:`qibo.quantum_info.hilbert_schmidt_inner_product`.
 
     Args:
         state (ndarray): statevector or density matrix.
         target (ndarray): statevector or density matrix.
+        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
+            in the execution. If ``None``, it uses the current backend.
+            Defaults to ``None``.
 
     Returns:
         float: Hilbert-Schmidt distance between ``state`` :math:`\\rho`
         and ``target`` :math:`\\sigma`.
+
+    References:
+        1. P. J. Coles, M. Cerezo, and L. Cincio, *Strong bound between trace distance
+        and Hilbert-Schmidt distance for low-rank states*, `Phys. Rev. A 100, 022103
+        <https://doi.org/10.1103/PhysRevA.100.022103>`_ (2019).
     """
+    backend = _check_backend(backend)
 
     if state.shape != target.shape:
         raise_error(
@@ -172,16 +170,15 @@ def hilbert_schmidt_distance(state, target):
         )
 
     if len(state.shape) == 1:
-        state = np.outer(np.conj(state), state)
-        target = np.outer(np.conj(target), target)
+        state = backend.np.outer(backend.np.conj(state), state)
+        target = backend.np.outer(backend.np.conj(target), target)
 
-    distance = np.real(np.trace((state - target) ** 2))
-    distance = float(distance)
+    difference = state - target
 
-    return distance
+    return hilbert_schmidt_inner_product(difference, difference, backend=backend)
 
 
-def fidelity(state, target, check_hermitian: bool = False, backend=None):
+def fidelity(state, target, backend=None):
     """Fidelity :math:`F(\\rho, \\sigma)` between ``state`` :math:`\\rho` and
     ``target`` state :math:`\\sigma`. In general,
 
@@ -197,17 +194,16 @@ def fidelity(state, target, check_hermitian: bool = False, backend=None):
     Args:
         state (ndarray): statevector or density matrix.
         target (ndarray): statevector or density matrix.
-        check_hermitian (bool, optional): if ``True``, checks if ``state`` is Hermitian.
-            Defaults to ``False``.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
-            in the execution. If ``None``, it uses :class:`qibo.backends.GlobalBackend`.
+            in the execution. If ``None``, it uses the current backend.
             Defaults to ``None``.
 
     Returns:
         float: Fidelity between ``state`` :math:`\\rho` and ``target`` :math:`\\sigma`.
     """
     backend = _check_backend(backend)
-
+    state = backend.cast(state, dtype=state.dtype)
+    target = backend.cast(target, dtype=target.dtype)
     if state.shape != target.shape:
         raise_error(
             TypeError,
@@ -221,72 +217,34 @@ def fidelity(state, target, check_hermitian: bool = False, backend=None):
             + f"but have dims {state.shape} and {target.shape}",
         )
 
-    if isinstance(check_hermitian, bool) is False:
-        raise_error(
-            TypeError,
-            f"check_hermitian must be type bool, but it is type {type(check_hermitian)}.",
-        )
-
     # check purity if both states are density matrices
     if len(state.shape) == 2 and len(target.shape) == 2:
-        purity_state = purity(state)
-        purity_target = purity(target)
+        purity_state = purity(state, backend=backend)
+        purity_target = purity(target, backend=backend)
 
         # if both states are mixed, default to full fidelity calculation
         if (
             abs(purity_state - 1) > PRECISION_TOL
             and abs(purity_target - 1) > PRECISION_TOL
         ):
-            # using eigh since rho is supposed to be Hermitian
-            if check_hermitian is False or _check_hermitian_or_not_gpu(
-                state, backend=backend
-            ):
-                eigenvalues, eigenvectors = np.linalg.eigh(state)
-            else:
-                eigenvalues, eigenvectors = np.linalg.eig(state)
-            state = np.zeros(state.shape, dtype=complex)
-            state = backend.cast(state, dtype=state.dtype)
-            for eig, eigvec in zip(eigenvalues, np.transpose(eigenvectors)):
-                matrix = np.sqrt(eig) * np.outer(eigvec, np.conj(eigvec))
-                matrix = backend.cast(matrix, dtype=matrix.dtype)
-                state += matrix
-                del matrix
+            fid = backend.calculate_matrix_sqrt(state)
+            fid = fid @ backend.np.conj(target.T) @ fid
+            fid = backend.calculate_matrix_sqrt(fid)
+            fid = backend.np.real(backend.np.trace(fid))
 
-            fid = state @ target @ state
-
-            # since sqrt(rho) is Hermitian, we can use eigh again
-            if check_hermitian is False or _check_hermitian_or_not_gpu(
-                fid, backend=backend
-            ):
-                eigenvalues, eigenvectors = np.linalg.eigh(fid)
-            else:
-                eigenvalues, eigenvectors = np.linalg.eig(fid)
-            fid = np.zeros(state.shape, dtype=complex)
-            fid = backend.cast(fid, dtype=fid.dtype)
-            for eig, eigvec in zip(eigenvalues, np.transpose(eigenvectors)):
-                if eig > PRECISION_TOL:
-                    matrix = np.sqrt(eig) * np.outer(eigvec, np.conj(eigvec))
-                    matrix = backend.cast(matrix, dtype=matrix.dtype)
-                    fid += matrix
-                    del matrix
-
-            fid = np.real(np.trace(fid)) ** 2
-
-            return fid
+            return fid**2
 
     # if any of the states is pure, perform lighter calculation
     fid = (
-        np.abs(np.dot(np.conj(state), target)) ** 2
+        backend.np.abs(backend.np.matmul(backend.np.conj(state), target)) ** 2
         if len(state.shape) == 1
-        else np.real(np.trace(np.dot(state, target)))
+        else backend.np.real(backend.np.trace(backend.np.matmul(state, target)))
     )
 
-    fid = float(fid)
-
     return fid
 
 
-def infidelity(state, target, check_hermitian: bool = False, backend=None):
+def infidelity(state, target, backend=None):
     """Infidelity between ``state`` :math:`\\rho` and ``target`` state
     :math:`\\sigma`, which is given by
 
@@ -299,19 +257,17 @@ def infidelity(state, target, check_hermitian: bool = False, backend=None):
     Args:
         state (ndarray): statevector or density matrix.
         target (ndarray): statevector or density matrix.
-        check_hermitian (bool, optional): if ``True``, checks if ``state`` is Hermitian.
-            Defaults to ``False``.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
-            in the execution. If ``None``, it uses :class:`qibo.backends.GlobalBackend`.
+            in the execution. If ``None``, it uses the current backend.
             Defaults to ``None``.
 
     Returns:
         float: Infidelity between ``state`` :math:`\\rho` and ``target`` :math:`\\sigma`.
     """
-    return 1 - fidelity(state, target, check_hermitian=check_hermitian, backend=backend)
+    return 1 - fidelity(state, target, backend=backend)
 
 
-def bures_angle(state, target, check_hermitian: bool = False, backend=None):
+def bures_angle(state, target, backend=None):
     """Calculates the Bures angle :math:`D_{A}` between a ``state``
     :math:`\\rho` and a ``target`` state :math:`\\sigma`. This is given by
 
@@ -324,23 +280,21 @@ def bures_angle(state, target, check_hermitian: bool = False, backend=None):
     Args:
         state (ndarray): statevector or density matrix.
         target (ndarray): statevector or density matrix.
-        check_hermitian (bool, optional): if ``True``, checks if ``state`` is Hermitian.
-            Defaults to ``False``.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
-            in the execution. If ``None``, it uses :class:`qibo.backends.GlobalBackend`.
+            in the execution. If ``None``, it uses the current backend.
             Defaults to ``None``.
 
     Returns:
         float: Bures angle between ``state`` and ``target``.
     """
-    angle = np.arccos(
-        np.sqrt(fidelity(state, target, check_hermitian, backend=backend))
-    )
+    backend = _check_backend(backend)
+
+    angle = backend.np.arccos(backend.np.sqrt(fidelity(state, target, backend=backend)))
 
     return angle
 
 
-def bures_distance(state, target, check_hermitian: bool = False, backend=None):
+def bures_distance(state, target, backend=None):
     """Calculates the Bures distance :math:`D_{B}` between a ``state``
     :math:`\\rho` and a ``target`` state :math:`\\sigma`. This is given by
 
@@ -353,18 +307,17 @@ def bures_distance(state, target, check_hermitian: bool = False, backend=None):
     Args:
         state (ndarray): statevector or density matrix.
         target (ndarray): statevector or density matrix.
-        check_hermitian (bool, optional): if ``True``, checks if ``state`` is Hermitian.
-            Defaults to ``False``.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
-            in the execution. If ``None``, it uses :class:`qibo.backends.GlobalBackend`.
+            in the execution. If ``None``, it uses the current backend.
             Defaults to ``None``.
 
     Returns:
         float: Bures distance between ``state`` and ``target``.
     """
-    distance = np.sqrt(
-        2 * (1 - np.sqrt(fidelity(state, target, check_hermitian, backend=backend)))
-    )
+    backend = _check_backend(backend)
+
+    sqrt_fid = backend.np.sqrt(fidelity(state, target, backend=backend))
+    distance = backend.np.sqrt(2 * (1 - sqrt_fid))
 
     return distance
 
@@ -384,7 +337,7 @@ def process_fidelity(channel, target=None, check_unitary: bool = False, backend=
         check_unitary (bool, optional): if ``True``, checks if one of the
             input channels is unitary. Default: ``False``.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
-            in the execution. If ``None``, it uses :class:`qibo.backends.GlobalBackend`.
+            in the execution. If ``None``, it uses the current backend.
             Defaults to ``None``.
 
     Returns:
@@ -403,16 +356,22 @@ def process_fidelity(channel, target=None, check_unitary: bool = False, backend=
 
     if check_unitary is True:
         norm_channel = float(
-            backend.calculate_norm_density_matrix(
-                np.dot(np.conj(np.transpose(channel)), channel) - np.eye(dim**2)
+            backend.calculate_matrix_norm(
+                backend.np.matmul(
+                    backend.np.conj(backend.np.transpose(channel, (1, 0))), channel
+                )
+                - backend.np.eye(dim**2)
             )
         )
         if target is None and norm_channel > PRECISION_TOL:
             raise_error(TypeError, "Channel is not unitary and Target is None.")
         if target is not None:
             norm_target = float(
-                backend.calculate_norm(
-                    np.dot(np.conj(np.transpose(target)), target) - np.eye(dim**2)
+                backend.calculate_vector_norm(
+                    backend.np.matmul(
+                        backend.np.conj(backend.np.transpose(target, (1, 0))), target
+                    )
+                    - backend.np.eye(dim**2)
                 )
             )
             if (norm_channel > PRECISION_TOL) and (norm_target > PRECISION_TOL):
@@ -420,14 +379,15 @@ def process_fidelity(channel, target=None, check_unitary: bool = False, backend=
 
     if target is None:
         # With no target, return process fidelity with Identity channel
-        process_fid = np.real(np.trace(channel)) / dim**2
+        process_fid = backend.np.real(backend.np.trace(channel)) / dim**2
         process_fid = float(process_fid)
 
         return process_fid
 
-    process_fid = np.dot(np.transpose(np.conj(channel)), target)
-    process_fid = np.real(np.trace(process_fid)) / dim**2
-    process_fid = float(process_fid)
+    process_fid = backend.np.matmul(
+        backend.np.transpose(backend.np.conj(channel), (1, 0)), target
+    )
+    process_fid = backend.np.real(backend.np.trace(process_fid)) / dim**2
 
     return process_fid
 
@@ -448,7 +408,7 @@ def process_infidelity(channel, target=None, check_unitary: bool = False, backen
         check_unitary (bool, optional): if ``True``, checks if one of the
             input channels is unitary. Defaults to ``False``.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
-            in the execution. If ``None``, it uses :class:`qibo.backends.GlobalBackend`.
+            in the execution. If ``None``, it uses the current backend.
             Defaults to ``None``.
 
 
@@ -485,7 +445,7 @@ def average_gate_fidelity(
         check_unitary (bool, optional): if ``True``, checks if one of the
             input channels is unitary. Default: ``False``.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
-            in the execution. If ``None``, it uses :class:`qibo.backends.GlobalBackend`.
+            in the execution. If ``None``, it uses the current backend.
             Defaults to ``None``.
 
     Returns:
@@ -520,7 +480,7 @@ def gate_error(channel, target=None, check_unitary: bool = False, backend=None):
         check_unitary (bool, optional): if ``True``, checks if one of the
             input channels is unitary. Default: ``False``.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
-            in the execution. If ``None``, it uses :class:`qibo.backends.GlobalBackend`.
+            in the execution. If ``None``, it uses the current backend.
             Defaults to ``None``.
 
     Returns:
@@ -567,6 +527,9 @@ def diamond_norm(channel, target=None, backend=None, **kwargs):  # pragma: no co
         channel (ndarray): row-vectorized Choi representation of a quantum channel.
         target (ndarray, optional): row-vectorized Choi representation of a target
             quantum channel. Defaults to ``None``.
+        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
+            in the execution. If ``None``, it uses the current backend.
+            Defaults to ``None``.
         kwargs: optional arguments to pass to CVXPY solver. For more information,
             please visit `CVXPY's API documentation
             <https://www.cvxpy.org/api_reference/cvxpy.problems.html#problem>`_.
@@ -579,6 +542,8 @@ def diamond_norm(channel, target=None, backend=None, **kwargs):  # pragma: no co
     """
     import cvxpy
 
+    backend = _check_backend(backend)
+
     if target is not None:
         if channel.shape != target.shape:
             raise_error(
@@ -595,9 +560,9 @@ def diamond_norm(channel, target=None, backend=None, **kwargs):  # pragma: no co
 
     channel = backend.to_numpy(channel)
 
-    channel = np.transpose(channel)
-    channel_real = np.real(channel)
-    channel_imag = np.imag(channel)
+    channel = backend.np.transpose(channel, (1, 0))
+    channel_real = backend.np.real(channel)
+    channel_imag = backend.np.imag(channel)
 
     dim = int(np.sqrt(channel.shape[0]))
 
@@ -685,7 +650,7 @@ def expressibility(
             For specifications, see :meth:`qibo.backends.abstract.calculate_norm`.
             Defaults to :math:`2`.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
-            in the execution. If ``None``, it uses :class:`qibo.backends.GlobalBackend`.
+            in the execution. If ``None``, it uses the current backend.
             Defaults to ``None``.
 
     Returns:
@@ -713,7 +678,7 @@ def expressibility(
         circuit.nqubits, power_t, samples=None, backend=backend
     ) - pqc_integral(circuit, power_t, samples, backend=backend)
 
-    fid = float(backend.calculate_norm(deviation, order=order))
+    fid = float(backend.calculate_vector_norm(deviation, order=order))
 
     return fid
 
@@ -749,7 +714,7 @@ def frame_potential(
         power_t (int): power that defines the :math:`t`-design.
         samples (int): number of samples to estimate the integral.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
-            in the execution. If ``None``, it uses :class:`qibo.backends.GlobalBackend`.
+            in the execution. If ``None``, it uses the current backend.
             Defaults to ``None``.
 
     Returns:
@@ -787,47 +752,80 @@ def frame_potential(
             unitary_2.set_parameters(params_2)
             unitary_2 = unitary_2.unitary(backend) / np.sqrt(dim)
 
-            potential += np.abs(
-                np.trace(np.transpose(np.conj(unitary_1)) @ unitary_2)
+            potential += backend.np.abs(
+                backend.np.trace(
+                    backend.np.transpose(backend.np.conj(unitary_1), (1, 0)) @ unitary_2
+                )
             ) ** (2 * power_t)
 
     return potential / samples**2
 
 
-def _check_hermitian_or_not_gpu(matrix, backend=None):
-    """Checks if a given matrix is Hermitian and whether the backend is neither
-    :class:`qibojit.backends.CupyBackend` nor
-    :class:`qibojit.backends.CuQuantumBackend`.
+def quantum_fisher_information_matrix(
+    circuit,
+    parameters=None,
+    initial_state=None,
+    return_complex: bool = True,
+    backend=None,
+):
+    """Calculate the Quantum Fisher Information Matrix (QFIM) of a parametrized ``circuit``.
+
+    Given a set of ``parameters`` :math:`\\theta = \\{\\theta_{k}\\}_{k\\in[M]}` and a
+    parameterized unitary ``circuit`` :math:`U(\\theta)` acting on an ``initial_state``
+    :math:`\\ket{\\phi}`, the QFIM is such that its elements can be calculated as
+
+    .. math::
+        \\mathbf{F}_{jk} = 4 \\, \\text{Re}\\left\\{ \\braket{\\partial_{j} \\psi | \\partial_{k}
+            \\psi} - \\braket{\\partial_{j} \\psi | \\psi}\\!\\braket{\\psi | \\partial_{k} \\psi}
+            \\right\\} \\, ,
+
+    where we have used the short notations :math:`\\ket{\\psi} \\equiv \\ket{\\psi(\\theta)}
+    = U(\\theta) \\ket{\\phi}`, and :math:`\\ket{\\partial_{k} \\psi} \\equiv \\frac{\\partial}
+    {\\partial\\theta_{k}} \\ket{\\psi(\\theta)}`.
+    If the ``initial_state`` :math:`\\ket{\\phi}` is not specified, it defaults to
+    :math:`\\ket{0}^{\\otimes n}`.
 
     Args:
-        matrix: input array.
+        circuit (:class:`qibo.models.circuit.Circuit`): parametrized circuit :math:`U(\\theta)`.
+        parameters (ndarray, optional): parameters whose QFIM to calculate.
+            If ``None``, QFIM is calculated with the paremeters from ``circuit``, i.e.
+            ``parameters = circuit.get_parameters()``. Defaults to ``None``.
+        initial_state (ndarray, optional): Initial configuration. It can be specified
+            by the setting the state vector using an array or a circuit. If ``None``,
+            the initial state is :math:`\\ket{0}^{\\otimes n}`. Defaults to ``None``.
+        return_complex (bool, optional): If ``True``, calculates the Jacobian matrix
+            of real and imaginary parts of :math:`\\ket{\\psi(\\theta)}`. If ``False``,
+            calculates only the Jacobian matrix of the real part. Defaults to ``True``.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
             in the execution. If ``None``, it uses :class:`qibo.backends.GlobalBackend`.
             Defaults to ``None``.
 
     Returns:
-        bool: whether the matrix is Hermitian.
-
-    Raises:
-        NotImplementedError: If `matrix` is not Hermitian and
-        `backend` is not :class:`qibojit.backends.CupyBackend`
+        ndarray: Quantum Fisher Information :math:`\\mathbf{F}`.
     """
     backend = _check_backend(backend)
 
-    norm = backend.calculate_norm_density_matrix(
-        np.transpose(np.conj(matrix)) - matrix, order=2
+    if parameters is None:
+        parameters = circuit.get_parameters()
+        parameters = backend.cast(parameters, dtype=float).flatten()
+
+    jacobian = backend.calculate_jacobian_matrix(
+        circuit, parameters, initial_state, return_complex
     )
 
-    hermitian = bool(float(norm) <= PRECISION_TOL)
+    if return_complex:
+        jacobian = jacobian[0] + 1j * jacobian[1]
 
-    if hermitian is False and backend.__class__.__name__ in [
-        "CupyBackend",
-        "CuQuantumBackend",
-    ]:  # pragma: no cover
-        raise_error(
-            NotImplementedError,
-            "GPU backends do not support `np.linalg.eig` "
-            + "or `np.linalg.eigvals` for non-Hermitian matrices.",
-        )
+    jacobian = backend.cast(jacobian, dtype=np.complex128)
 
-    return hermitian
+    copied = circuit.copy(deep=True)
+    copied.set_parameters(parameters)
+
+    state = backend.execute_circuit(copied, initial_state=initial_state).state()
+
+    overlaps = jacobian.T @ state
+
+    qfim = jacobian.T @ jacobian
+    qfim = qfim - backend.np.outer(overlaps, backend.np.conj(overlaps.T))
+
+    return 4 * backend.np.real(qfim)
