@@ -54,7 +54,7 @@ def uniform_sampling_U3(ngates: int, seed=None, backend=None):
             numbers or a fixed seed to initialize a generator. If ``None``, initializes
             a generator with a random seed. Default: ``None``.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
-            in the execution. If ``None``, it uses :class:`qibo.backends.GlobalBackend`.
+            in the execution. If ``None``, it uses the current backend.
             Defaults to ``None``.
 
     Returns:
@@ -110,7 +110,7 @@ def random_gaussian_matrix(
             numbers or a fixed seed to initialize a generator. If ``None``, initializes
             a generator with a random seed. Default: ``None``.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
-            in the execution. If ``None``, it uses :class:`qibo.backends.GlobalBackend`.
+            in the execution. If ``None``, it uses the current backend.
             Defaults to ``None``.
 
     Returns:
@@ -166,7 +166,7 @@ def random_hermitian(
             random numbers or a fixed seed to initialize a generator. If ``None``,
             initializes a generator with a random seed. Defaults to ``None``.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
-            in the execution. If ``None``, it uses :class:`qibo.backends.GlobalBackend`.
+            in the execution. If ``None``, it uses the current backend.
             Defaults to ``None``.
 
     Returns:
@@ -184,12 +184,12 @@ def random_hermitian(
     matrix = random_gaussian_matrix(dims, dims, seed=local_state, backend=backend)
 
     if semidefinite:
-        matrix = np.dot(np.transpose(np.conj(matrix)), matrix)
+        matrix = backend.np.matmul(backend.np.conj(matrix).T, matrix)
     else:
-        matrix = (matrix + np.transpose(np.conj(matrix))) / 2
+        matrix = (matrix + backend.np.conj(matrix).T) / 2
 
     if normalize:
-        matrix = matrix / np.linalg.norm(matrix)
+        matrix = matrix / np.linalg.norm(backend.to_numpy(matrix))
 
     return matrix
 
@@ -208,7 +208,7 @@ def random_unitary(dims: int, measure: Optional[str] = None, seed=None, backend=
             random numbers or a fixed seed to initialize a generator. If ``None``,
             initializes a generator with a random seed. Defaults to ``None``.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
-            in the execution. If ``None``, it uses :class:`qibo.backends.GlobalBackend`.
+            in the execution. If ``None``, it uses the current backend.
             Defaults to ``None``.
 
     Returns:
@@ -230,18 +230,18 @@ def random_unitary(dims: int, measure: Optional[str] = None, seed=None, backend=
 
     if measure == "haar":
         unitary = random_gaussian_matrix(dims, dims, seed=local_state, backend=backend)
-        Q, R = np.linalg.qr(unitary)
-        D = np.diag(R)
-        D = D / np.abs(D)
-        R = np.diag(D)
-        unitary = np.dot(Q, R)
+        # Tensorflow experi
+        Q, R = backend.np.linalg.qr(unitary)
+        D = backend.np.diag(R)
+        D = D / backend.np.abs(D)
+        R = backend.np.diag(D)
+        unitary = backend.np.matmul(Q, R)
     elif measure is None:
         from scipy.linalg import expm
 
         H = random_hermitian(dims, seed=seed, backend=NumpyBackend())
         unitary = expm(-1.0j * H / 2)
-
-    unitary = backend.cast(unitary, dtype=unitary.dtype)
+        unitary = backend.cast(unitary, dtype=unitary.dtype)
 
     return unitary
 
@@ -310,7 +310,7 @@ def random_quantum_channel(
             random numbers or a fixed seed to initialize a generator. If ``None``,
             initializes a generator with a random seed. Defaults to ``None``.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
-            in the execution. If ``None``, it uses :class:`qibo.backends.GlobalBackend`.
+            in the execution. If ``None``, it uses the current backend.
             Defaults to ``None``.
 
     Returns:
@@ -351,7 +351,7 @@ def random_quantum_channel(
     else:
         super_op = random_unitary(dims, measure, local_state, backend)
         super_op = vectorization(super_op, order=order, backend=backend)
-        super_op = np.outer(super_op, np.conj(super_op))
+        super_op = backend.np.outer(super_op, backend.np.conj(super_op))
 
     if "chi" in representation:
         pauli_order = "IXYZ"
@@ -399,7 +399,7 @@ def random_quantum_channel(
     return super_op
 
 
-def random_statevector(dims: int, seed=None, backend=None):
+def random_statevector(dims: int, dtype=None, seed=None, backend=None):
     """Creates a random statevector :math:`\\ket{\\psi}`.
 
     .. math::
@@ -409,13 +409,23 @@ def random_statevector(dims: int, seed=None, backend=None):
     where :math:`d` is ``dims``, and :math:`p_{k}` and :math:`\\phi_{k}` are, respectively,
     the probability and phase corresponding to the computational basis state :math:`\\ket{k}`.
 
+    .. note::
+        If ``dtype`` if ``"float64"``, ``"float32"``, or an equivalent type,
+        then :math:`\\ket{\\psi} \\in \\mathbb{R}^{\\text{dims}}`. Otherwise,
+        :math:`\\ket{\\psi} \\in \\mathbb{C}^{\\text{dims}}`.
+
     Args:
         dims (int): dimension of the matrix.
+        dtype (str or type, optional): data type for the random statevector.
+            Options are ``"complex128``, ``"complex64"``, ``"float64"``, or ``"float32"``.
+            It also accepts the equivalent data types of ``backend``'s engines, *e.g.*
+            ``numpy.complex128`` or ``cupy.float64``. If ``None``, defaults to
+            ``backend.dtype``. Defaults to ``None``.
         seed (int or :class:`numpy.random.Generator`, optional): Either a generator of
             random numbers or a fixed seed to initialize a generator. If ``None``,
             initializes a generator with a random seed. Defaults to ``None``.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
-            in the execution. If ``None``, it uses :class:`qibo.backends.GlobalBackend`.
+            in the execution. If ``None``, it uses the current backend.
             Defaults to ``None``.
 
     Returns:
@@ -436,10 +446,17 @@ def random_statevector(dims: int, seed=None, backend=None):
 
     backend, local_state = _check_backend_and_local_state(seed, backend)
 
-    state = local_state.standard_normal(dims).astype(complex)
-    state += 1.0j * local_state.standard_normal(dims)
-    state /= np.linalg.norm(state)
-    state = backend.cast(state, dtype=state.dtype)
+    if dtype is None:
+        dtype = backend.dtype
+
+    state = local_state.standard_normal(dims)
+    state = backend.cast(state, dtype=dtype)
+    if "complex" in str(dtype):
+        state += 0j
+        state = state + 1.0j * backend.cast(
+            local_state.standard_normal(dims), dtype=dtype
+        )
+    state = state / backend.calculate_vector_norm(state)
 
     return state
 
@@ -493,7 +510,7 @@ def random_density_matrix(
             random numbers or a fixed seed to initialize a generator. If ``None``,
             initializes a generator with a random seed. Defaults to ``None``.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
-            in the execution. If ``None``, it uses :class:`qibo.backends.GlobalBackend`.
+            in the execution. If ``None``, it uses the current backend.
             Defaults to ``None``.
 
     Returns:
@@ -546,24 +563,28 @@ def random_density_matrix(
 
     if pure:
         state = random_statevector(dims, seed=local_state, backend=backend)
-        state = np.outer(state, np.transpose(np.conj(state)))
+        state = backend.np.outer(state, backend.np.conj(state).T)
     else:
         if metric in ["hilbert-schmidt", "ginibre"]:
             state = random_gaussian_matrix(
                 dims, rank, mean=0, stddev=1, seed=local_state, backend=backend
             )
-            state = np.dot(state, np.transpose(np.conj(state)))
-            state = state / np.trace(state)
+            state = backend.np.matmul(
+                state, backend.np.transpose(backend.np.conj(state), (1, 0))
+            )
+            state = state / backend.np.trace(state)
         else:
             nqubits = int(np.log2(dims))
             state = backend.identity_density_matrix(nqubits, normalize=False)
             state += random_unitary(dims, seed=local_state, backend=backend)
-            state = np.dot(
+            state = backend.np.matmul(
                 state,
                 random_gaussian_matrix(dims, rank, seed=local_state, backend=backend),
             )
-            state = np.dot(state, np.transpose(np.conj(state)))
-            state /= np.trace(state)
+            state = backend.np.matmul(
+                state, backend.np.transpose(backend.np.conj(state), (1, 0))
+            )
+            state /= backend.np.trace(state)
 
     state = backend.cast(state, dtype=state.dtype)
 
@@ -601,7 +622,7 @@ def random_clifford(
             random numbers or a fixed seed to initialize a generator. If ``None``,
             initializes a generator with a random seed. Defaults to ``None``.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
-            in the execution. If ``None``, it uses :class:`qibo.backends.GlobalBackend`.
+            in the execution. If ``None``, it uses the current backend.
             Defaults to ``None``.
 
     Returns:
@@ -753,7 +774,7 @@ def random_pauli(
             random numbers or a fixed seed to initialize a generator. If ``None``,
             initializes a generator with a random seed. Defaults to ``None``.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
-            in the execution. If ``None``, it uses :class:`qibo.backends.GlobalBackend`.
+            in the execution. If ``None``, it uses the current backend.
             Defaults to ``None``.
 
     Returns:
@@ -874,7 +895,7 @@ def random_pauli_hamiltonian(
             random numbers or a fixed seed to initialize a generator. If ``None``,
             initializes a generator with a random seed. Defaults to ``None``.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
-            in the execution. If ``None``, it uses :class:`qibo.backends.GlobalBackend`.
+            in the execution. If ``None``, it uses the current backend.
             Defaults to ``None``.
 
     Returns:
@@ -918,12 +939,15 @@ def random_pauli_hamiltonian(
 
     hamiltonian = random_hermitian(d, normalize=True, seed=local_state, backend=backend)
 
-    eigenvalues, eigenvectors = np.linalg.eigh(hamiltonian)
+    eigenvalues, eigenvectors = backend.calculate_eigenvectors(hamiltonian)
+    if backend.platform == "tensorflow":
+        eigenvalues = backend.to_numpy(eigenvalues)
+        eigenvectors = backend.to_numpy(eigenvectors)
 
     if normalize is True:
-        eigenvalues -= eigenvalues[0]
+        eigenvalues = eigenvalues - eigenvalues[0]
 
-        eigenvalues /= eigenvalues[1]
+        eigenvalues = eigenvalues / eigenvalues[1]
 
         shift = 2
         eigenvectors[:, shift:] = (
@@ -935,15 +959,17 @@ def random_pauli_hamiltonian(
         hamiltonian = backend.cast(hamiltonian, dtype=hamiltonian.dtype)
         # excluding the first eigenvector because first eigenvalue is zero
         for eigenvalue, eigenvector in zip(
-            eigenvalues[1:], np.transpose(eigenvectors)[1:]
+            eigenvalues[1:], backend.np.transpose(eigenvectors, (1, 0))[1:]
         ):
-            hamiltonian += eigenvalue * np.outer(eigenvector, np.conj(eigenvector))
+            hamiltonian = hamiltonian + eigenvalue * backend.np.outer(
+                eigenvector, backend.np.conj(eigenvector)
+            )
 
     U = comp_basis_to_pauli(
         nqubits, normalize=True, pauli_order=pauli_order, backend=backend
     )
 
-    hamiltonian = np.real(U @ vectorization(hamiltonian, backend=backend))
+    hamiltonian = backend.np.real(U @ vectorization(hamiltonian, backend=backend))
 
     return hamiltonian, eigenvalues
 
@@ -976,7 +1002,7 @@ def random_stochastic_matrix(
             random numbers or a fixed seed to initialize a generator. If ``None``,
             initializes a statevectorgenerator with a random seed. Defaults to ``None``.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
-            in the execution. If ``None``, it uses :class:`qibo.backends.GlobalBackend`.
+            in the execution. If ``None``, it uses the current backend.
             Defaults to ``None``.
 
     Returns:
@@ -1185,7 +1211,7 @@ def _super_op_from_bcsz_measure(dims: int, rank: int, order: str, seed, backend)
             random numbers or a fixed seed to initialize a generator. If ``None``,
             initializes a generator with a random seed. Defaults to ``None``.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
-            in the execution. If ``None``, it uses :class:`qibo.backends.GlobalBackend`.
+            in the execution. If ``None``, it uses the current backend.
             Defaults to ``None``.
     """
     nqubits = int(np.log2(dims))
@@ -1193,10 +1219,12 @@ def _super_op_from_bcsz_measure(dims: int, rank: int, order: str, seed, backend)
     super_op = random_gaussian_matrix(
         dims**2, rank=rank, mean=0, stddev=1, seed=seed, backend=backend
     )
-    super_op = super_op @ np.transpose(np.conj(super_op))
+    super_op = super_op @ backend.np.conj(super_op).T
 
     # partial trace implemented with einsum
-    super_op_reduced = np.einsum("ijik->jk", np.reshape(super_op, (dims,) * 4))
+    super_op_reduced = np.einsum(
+        "ijik->jk", np.reshape(backend.to_numpy(super_op), (dims,) * 4)
+    )
 
     eigenvalues, eigenvectors = np.linalg.eigh(super_op_reduced)
 
@@ -1204,8 +1232,12 @@ def _super_op_from_bcsz_measure(dims: int, rank: int, order: str, seed, backend)
 
     operator = np.zeros((dims, dims), dtype=complex)
     operator = backend.cast(operator, dtype=operator.dtype)
-    for eigenvalue, eigenvector in zip(eigenvalues, np.transpose(eigenvectors)):
-        operator += eigenvalue * np.outer(eigenvector, np.conj(eigenvector))
+    for eigenvalue, eigenvector in zip(
+        backend.cast(eigenvalues), backend.cast(eigenvectors).T
+    ):
+        operator = operator + eigenvalue * backend.np.outer(
+            eigenvector, backend.np.conj(eigenvector)
+        )
 
     if order == "row":
         operator = backend.np.kron(
