@@ -1,7 +1,7 @@
 """Module with the most commom superoperator transformations."""
 
 import warnings
-from typing import Optional
+from typing import List, Optional, Tuple, Union
 
 import numpy as np
 from scipy.optimize import minimize
@@ -11,43 +11,46 @@ from qibo.config import PRECISION_TOL, raise_error
 from qibo.gates.abstract import Gate
 from qibo.gates.gates import Unitary
 from qibo.gates.special import FusedGate
+from qibo.quantum_info.linalg_operations import singular_value_decomposition
 
 
 def vectorization(state, order: str = "row", backend=None):
-    """Returns state :math:`\\rho` in its Liouville
-    representation :math:`|\\rho\\rangle\\rangle`.
+    """Returns state :math:`\\rho` in its Liouville representation :math:`|\\rho)`.
 
     If ``order="row"``, then:
 
     .. math::
-        |\\rho\\rangle\\rangle = \\sum_{k, l} \\, \\rho_{kl} \\, \\ket{k} \\otimes \\ket{l}
+        |\\rho) = \\sum_{k, l} \\, \\rho_{kl} \\, \\ket{k} \\otimes \\ket{l} \\, .
 
     If ``order="column"``, then:
 
     .. math::
-        |\\rho\\rangle\\rangle = \\sum_{k, l} \\, \\rho_{kl} \\, \\ket{l} \\otimes \\ket{k}
+        |\\rho) = \\sum_{k, l} \\, \\rho_{kl} \\, \\ket{l} \\otimes \\ket{k} \\, .
+
+    If ``state`` is a 3-dimensional tensor, it is interpreted as a batch of states.
 
     Args:
-        state: state vector or density matrix.
+        state (ndarray): statevector, density matrix, an array of statevectors,
+            or an array of density matrices.
         order (str, optional): If ``"row"``, vectorization is performed
             row-wise. If ``"column"``, vectorization is performed
             column-wise. If ``"system"``, a block-vectorization is
             performed. Defaults to ``"row"``.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend
             to be used in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
         ndarray: Liouville representation of ``state``.
     """
     if (
-        (len(state.shape) >= 3)
+        (len(state.shape) > 3)
         or (len(state) == 0)
         or (len(state.shape) == 2 and state.shape[0] != state.shape[1])
     ):
         raise_error(
             TypeError,
-            f"Object must have dims either (k,) or (k,k), but have dims {state.shape}.",
+            f"Object must have dims either (k,), (k, k), (N, 1, k) or (N, k, k), but have dims {state.shape}.",
         )
 
     if not isinstance(order, str):
@@ -63,38 +66,48 @@ def vectorization(state, order: str = "row", backend=None):
 
     backend = _check_backend(backend)
 
+    dims = state.shape[-1]
+
     if len(state.shape) == 1:
-        state = np.outer(state, np.conj(state))
+        state = backend.np.outer(state, backend.np.conj(state))
+    elif len(state.shape) == 3 and state.shape[1] == 1:
+        state = backend.np.einsum(
+            "aij,akl->aijkl", state, backend.np.conj(state)
+        ).reshape(state.shape[0], dims, dims)
 
     if order == "row":
-        state = np.reshape(state, (1, -1), order="C")[0]
+        state = backend.np.reshape(state, (-1, dims**2))
     elif order == "column":
-        state = np.reshape(state, (1, -1), order="F")[0]
+        indices = list(range(len(state.shape)))
+        indices[-2:] = reversed(indices[-2:])
+        state = backend.np.transpose(state, indices)
+        state = backend.np.reshape(state, (-1, dims**2))
     else:
-        dim = len(state)
-        nqubits = int(np.log2(dim))
+        nqubits = int(np.log2(state.shape[-1]))
 
-        new_axis = []
+        new_axis = [0]
         for qubit in range(nqubits):
-            new_axis += [qubit + nqubits, qubit]
+            new_axis.extend([qubit + nqubits + 1, qubit + 1])
 
-        state = np.reshape(state, [2] * 2 * nqubits)
-        state = np.transpose(state, axes=new_axis)
-        state = np.reshape(state, -1)
+        state = backend.np.reshape(state, [-1] + [2] * 2 * nqubits)
+        state = backend.np.transpose(state, new_axis)
+        state = backend.np.reshape(state, (-1, 2 ** (2 * nqubits)))
 
-    state = backend.cast(state, dtype=state.dtype)
+    state = backend.np.squeeze(
+        state, axis=tuple(i for i, ax in enumerate(state.shape) if ax == 1)
+    )
 
     return state
 
 
 def unvectorization(state, order: str = "row", backend=None):
     """Returns state :math:`\\rho` from its Liouville
-    representation :math:`|\\rho\\rangle\\rangle`. This operation is
+    representation :math:`|\\rho)`. This operation is
     the inverse function of :func:`vectorization`, i.e.
 
     .. math::
         \\begin{align}
-            \\rho &= \\text{unvectorization}(|\\rho\\rangle\\rangle) \\nonumber \\\\
+            \\rho &= \\text{unvectorization}(|\\rho)) \\nonumber \\\\
             &= \\text{unvectorization}(\\text{vectorization}(\\rho)) \\nonumber
         \\end{align}
 
@@ -106,7 +119,7 @@ def unvectorization(state, order: str = "row", backend=None):
             performed. Defaults to ``"row"``.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend
             to be used in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
         ndarray: Density matrix of ``state``.
@@ -130,20 +143,21 @@ def unvectorization(state, order: str = "row", backend=None):
             )
 
     backend = _check_backend(backend)
+    state = backend.cast(state)
 
     dim = int(np.sqrt(len(state)))
 
     if order in ["row", "column"]:
         order = "C" if order == "row" else "F"
-        state = np.reshape(state, (dim, dim), order=order)
+        state = backend.cast(
+            np.reshape(backend.to_numpy(state), (dim, dim), order=order)
+        )
     else:
         nqubits = int(np.log2(dim))
         axes_old = list(np.arange(0, 2 * nqubits))
-        state = np.reshape(state, [2] * 2 * nqubits)
-        state = np.transpose(state, axes=axes_old[1::2] + axes_old[0::2])
-        state = np.reshape(state, [2**nqubits] * 2)
-
-    state = backend.cast(state, dtype=state.dtype)
+        state = backend.np.reshape(state, [2] * 2 * nqubits)
+        state = backend.np.transpose(state, axes_old[1::2] + axes_old[0::2])
+        state = backend.np.reshape(state, [2**nqubits] * 2)
 
     return state
 
@@ -152,9 +166,9 @@ def to_choi(channel, order: str = "row", backend=None):
     """Converts quantum ``channel`` :math:`U` to its Choi representation :math:`\\Lambda`.
 
     .. math::
-        \\Lambda = | U \\rangle\\rangle \\langle\\langle U | \\, ,
+        \\Lambda = | U ) ( U | \\, ,
 
-    where :math:`| \\cdot \\rangle\\rangle` is the :func:`qibo.quantum_info.vectorization`
+    where :math:`| \\cdot )` is the :func:`qibo.quantum_info.vectorization`
     operation.
 
     Args:
@@ -165,13 +179,15 @@ def to_choi(channel, order: str = "row", backend=None):
             performed. Default is ``"row"``.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend
             to be used in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
         ndarray: quantum channel in its Choi representation.
     """
+    backend = _check_backend(backend)
+
     channel = vectorization(channel, order=order, backend=backend)
-    channel = np.outer(channel, np.conj(channel))
+    channel = backend.np.outer(channel, backend.np.conj(channel))
 
     return channel
 
@@ -189,11 +205,13 @@ def to_liouville(channel, order: str = "row", backend=None):
             performed. Default is ``"row"``.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend
             to be used in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
         ndarray: quantum channel in its Liouville representation.
     """
+    backend = _check_backend(backend)
+
     channel = to_choi(channel, order=order, backend=backend)
     channel = _reshuffling(channel, order=order, backend=backend)
 
@@ -224,12 +242,14 @@ def to_pauli_liouville(
             Pauli elements. Default is "IXYZ".
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend
             to be used in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
         ndarray: quantum channel in its Pauli-Liouville representation.
     """
-    from qibo.quantum_info.basis import comp_basis_to_pauli
+    from qibo.quantum_info.basis import comp_basis_to_pauli  # pylint: disable=C0415
+
+    backend = _check_backend(backend)
 
     nqubits = int(np.log2(channel.shape[0]))
 
@@ -239,7 +259,7 @@ def to_pauli_liouville(
         nqubits, normalize, pauli_order=pauli_order, backend=backend
     )
 
-    channel = unitary @ channel @ np.transpose(np.conj(unitary))
+    channel = unitary @ channel @ backend.np.conj(unitary).T
 
     return channel
 
@@ -266,7 +286,7 @@ def to_chi(
             Pauli elements. Default is "IXYZ".
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend
             to be used in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
         ndarray: quantum channel in its :math:`\\chi`-representation.
@@ -283,6 +303,50 @@ def to_chi(
     return channel
 
 
+def to_stinespring(
+    channel,
+    partition: Optional[Union[List[int], Tuple[int, ...]]] = None,
+    nqubits: Optional[int] = None,
+    initial_state_env=None,
+    backend=None,
+):
+    """Convert quantum ``channel`` :math:`U` to its Stinespring representation :math:`U_{0}`.
+
+    It uses the Kraus representation as an intermediate step.
+
+    Args:
+        channel (ndarray): quantum channel.
+        nqubits (int, optional): total number of qubits in the system that is
+            interacting with the environment. Must be equal or greater than
+            the number of qubits ``channel`` acts on. If ``None``,
+            defaults to the number of qubits in ``channel``.
+            Defauts to ``None``.
+        initial_state_env (ndarray, optional): statevector representing the
+            initial state of the enviroment. If ``None``, it assumes the
+            environment in its ground state. Defaults to ``None``.
+        backend (:class:`qibo.backends.abstract.Backend`, optional): backend
+            to be used in the execution. If ``None``, it uses
+            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+
+    Returns:
+        ndarray: Quantum channel in its Stinespring representation :math:`U_{0}`.
+    """
+    backend = _check_backend(backend)
+
+    if partition is None:
+        nqubits_channel = int(np.log2(channel.shape[-1]))
+        partition = tuple(range(nqubits_channel))
+
+    channel = kraus_to_stinespring(
+        [(partition, channel)],
+        nqubits=nqubits,
+        initial_state_env=initial_state_env,
+        backend=backend,
+    )
+
+    return channel
+
+
 def choi_to_liouville(choi_super_op, order: str = "row", backend=None):
     """Converts Choi representation :math:`\\Lambda` of quantum channel
     to its Liouville representation :math:`\\mathcal{E}`.
@@ -300,7 +364,6 @@ def choi_to_liouville(choi_super_op, order: str = "row", backend=None):
         \\Lambda_{\\alpha\\beta, \\, \\gamma\\delta} \\mapsto
             \\Lambda_{\\delta\\beta, \\, \\gamma\\alpha} \\equiv \\mathcal{E}
 
-
     Args:
         choi_super_op: Choi representation of quantum channel.
         order (str, optional): If ``"row"``, reshuffling is performed
@@ -312,7 +375,7 @@ def choi_to_liouville(choi_super_op, order: str = "row", backend=None):
             respect to system-wise vectorization. Defaults to ``"row"``.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend
             to be used in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
         ndarray: Liouville representation of quantum channel.
@@ -343,7 +406,7 @@ def choi_to_pauli(
             Pauli elements. Defaults to "IXYZ".
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend
             to be used in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
         ndarray: superoperator in the Pauli-Liouville representation.
@@ -370,7 +433,9 @@ def choi_to_kraus(
 
     .. math::
         \\Lambda = \\sum_{\\alpha} \\, \\lambda_{\\alpha}^{2} \\,
-            |\\tilde{K}_{\\alpha}\\rangle\\rangle \\langle\\langle \\tilde{K}_{\\alpha}| \\, .
+            |\\tilde{K}_{\\alpha})(\\tilde{K}_{\\alpha}| \\, .
+
+    where :math:`|\\cdot)` is the :func:`qibo.quantum_info.vectorization` operation.
 
     This is the spectral decomposition of :math:`\\Lambda`, Hence, the set
     :math:`\\{\\lambda_{\\alpha}, \\, \\tilde{K}_{\\alpha}\\}_{\\alpha}`
@@ -379,7 +444,7 @@ def choi_to_kraus(
 
     .. math::
         K_{\\alpha} = \\lambda_{\\alpha} \\,
-            \\text{unvectorization}(|\\tilde{K}_{\\alpha}\\rangle\\rangle) \\, .
+            \\text{unvectorization}(|\\tilde{K}_{\\alpha})) \\, .
 
     If :math:`\\mathcal{E}` is not CP, then spectral composition is replaced by
     a singular value decomposition (SVD), i.e.
@@ -414,7 +479,7 @@ def choi_to_kraus(
             Defaults to ``True``.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend
             to be used in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
         tuple(ndarray, ndarray): The set
@@ -448,11 +513,12 @@ def choi_to_kraus(
         )
 
     backend = _check_backend(backend)
+    choi_super_op = backend.cast(choi_super_op)
 
     if validate_cp:
         norm = float(
-            backend.calculate_norm_density_matrix(
-                choi_super_op - np.transpose(np.conj(choi_super_op)), order=2
+            backend.calculate_matrix_norm(
+                choi_super_op - backend.np.conj(choi_super_op).T, order=2
             )
         )
         if norm > PRECISION_TOL:
@@ -460,25 +526,27 @@ def choi_to_kraus(
         else:
             # using eigh because, in this case, choi_super_op is
             # *already confirmed* to be Hermitian
-            eigenvalues, eigenvectors = np.linalg.eigh(choi_super_op)
-            eigenvectors = np.transpose(eigenvectors)
+            eigenvalues, eigenvectors = backend.calculate_eigenvectors(choi_super_op)
+            eigenvectors = eigenvectors.T
 
-            non_cp = bool(any(eigenvalues < -PRECISION_TOL))
+            non_cp = bool(any(backend.np.real(eigenvalues) < -PRECISION_TOL))
     else:
         non_cp = False
         # using eigh because, in this case, choi_super_op is
         # *assumed* to be Hermitian
-        eigenvalues, eigenvectors = np.linalg.eigh(choi_super_op)
-        eigenvectors = np.transpose(eigenvectors)
+        eigenvalues, eigenvectors = backend.calculate_eigenvectors(choi_super_op)
+        eigenvectors = eigenvectors.T
 
     if non_cp:
         warnings.warn("Input choi_super_op is a non-completely positive map.")
 
         # using singular value decomposition because choi_super_op is non-CP
-        U, coefficients, V = np.linalg.svd(choi_super_op)
-        U = np.transpose(U)
-        coefficients = np.sqrt(coefficients)
-        V = np.conj(V)
+        U, coefficients, V = singular_value_decomposition(
+            choi_super_op, backend=backend
+        )
+        U = U.T
+        coefficients = backend.np.sqrt(coefficients)
+        V = backend.np.conj(V)
 
         kraus_left, kraus_right = [], []
         for coeff, eigenvector_left, eigenvector_right in zip(coefficients, U, V):
@@ -495,8 +563,8 @@ def choi_to_kraus(
         # when choi_super_op is CP
         kraus_ops, coefficients = [], []
         for eig, kraus in zip(eigenvalues, eigenvectors):
-            if np.abs(eig) > precision_tol:
-                eig = np.sqrt(eig)
+            if backend.np.abs(eig) > precision_tol:
+                eig = backend.np.sqrt(eig)
                 kraus_ops.append(
                     eig * unvectorization(kraus, order=order, backend=backend)
                 )
@@ -537,7 +605,7 @@ def choi_to_chi(
             single-qubit Pauli elements. Defaults to "IXYZ".
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend
             to be used in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
     Returns:
         ndarray: Chi-matrix representation of the quantum channel.
     """
@@ -593,7 +661,7 @@ def choi_to_stinespring(
             environment in its ground state. Defaults to ``None``.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend
             to be used in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
         ndarray: Choi representation of quantum channel.
@@ -615,7 +683,7 @@ def choi_to_stinespring(
     if nqubits is None:
         nqubits = int(np.log2(kraus_ops[0].shape[0]))
 
-    nqubits_list = [tuple(range(nqubits)) for _ in range(len(kraus_ops))]
+    nqubits_list = [tuple(range(nqubits))] * len(kraus_ops)
 
     kraus_ops = list(zip(nqubits_list, kraus_ops))
 
@@ -631,7 +699,11 @@ def kraus_to_choi(kraus_ops, order: str = "row", backend=None):
     of quantum channel to its Choi representation :math:`\\Lambda`.
 
     .. math::
-        \\Lambda = \\sum_{\\alpha} \\, |K_{\\alpha}\\rangle\\rangle \\langle\\langle K_{\\alpha}|
+        \\Lambda = \\sum_{\\alpha} \\, |K_{\\alpha})( K_{\\alpha}|
+
+    where :math:`|K_{\\alpha})` is the vectorization of the Kraus operator
+    :math:`K_{\\alpha}`.
+    For a definition of vectorization, see :func:`qibo.quantum_info.vectorization`.
 
     Args:
         kraus_ops (list): List of Kraus operators as pairs ``(qubits, Ak)``
@@ -646,7 +718,7 @@ def kraus_to_choi(kraus_ops, order: str = "row", backend=None):
             respect to system-wise vectorization. Defaults to ``"row"``.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend
             to be used in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
         ndarray: Choi representation of the Kraus channel.
@@ -664,7 +736,7 @@ def kraus_to_choi(kraus_ops, order: str = "row", backend=None):
         kraus_op.append(gate)
         kraus_op = kraus_op.matrix(backend)
         kraus_op = vectorization(kraus_op, order=order, backend=backend)
-        super_op += np.outer(kraus_op, np.conj(kraus_op))
+        super_op = super_op + backend.np.outer(kraus_op, backend.np.conj(kraus_op))
         del kraus_op
 
     return super_op
@@ -692,7 +764,7 @@ def kraus_to_liouville(kraus_ops, order: str = "row", backend=None):
             respect to system-wise vectorization. Defaults to ``"row"``.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend
             to be used in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
         ndarray: Liouville representation of quantum channel.
@@ -728,7 +800,7 @@ def kraus_to_pauli(
             single-qubit Pauli elements. Defaults to "IXYZ".
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend
             to be used in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
         ndarray: superoperator in the Pauli-Liouville representation.
@@ -750,10 +822,11 @@ def kraus_to_chi(
     of quantum channel to  its :math:`\\chi`-matrix representation.
 
     .. math::
-        \\chi = \\sum_{\\alpha} \\, |c_{\\alpha}\\rangle\\rangle \\langle\\langle c_{\\alpha}|,
+        \\chi = \\sum_{\\alpha} \\, |c_{\\alpha})( c_{\\alpha}|,
 
-    where :math:`|c_{\\alpha}\\rangle\\rangle \\cong |K_{\\alpha}\\rangle\\rangle`
-    in Pauli-Liouville basis.
+    where :math:`|c_{\\alpha}) \\cong |K_{\\alpha})` in Pauli-Liouville basis,
+    and :math:`| \\cdot )` is the :func:`qibo.quantum_info.vectorization`
+    operation.
 
     Args:
         kraus_ops (list): List of Kraus operators as pairs ``(qubits, Ak)``
@@ -773,12 +846,12 @@ def kraus_to_chi(
             Pauli elements in the Pauli basis. Defaults to "IXYZ".
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend
             to be used in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
         ndarray: Chi-matrix representation of the Kraus channel.
     """
-    from qibo.quantum_info.basis import comp_basis_to_pauli
+    from qibo.quantum_info.basis import comp_basis_to_pauli  # pylint: disable=C0415
 
     backend = _check_backend(backend)
 
@@ -802,7 +875,7 @@ def kraus_to_chi(
         kraus_op = kraus_op.matrix(backend)
         kraus_op = vectorization(kraus_op, order=order, backend=backend)
         kraus_op = comp_to_pauli @ kraus_op
-        super_op += np.outer(kraus_op, np.conj(kraus_op))
+        super_op = super_op + backend.np.outer(kraus_op, backend.np.conj(kraus_op))
         del kraus_op
 
     return super_op
@@ -837,7 +910,7 @@ def kraus_to_stinespring(
             environment in its ground state. Defaults to ``None``.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend
             to be used in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
         ndarray: Stinespring representation (restricted unitary) of the Kraus channel.
@@ -872,7 +945,7 @@ def kraus_to_stinespring(
 
     # only utility is for outer product,
     # so np.conj here to only do it once
-    initial_state_env = np.conj(initial_state_env)
+    initial_state_env = backend.np.conj(initial_state_env)
 
     stinespring = np.zeros((dim_stinespring, dim_stinespring), dtype=complex)
     stinespring = backend.cast(stinespring, dtype=stinespring.dtype)
@@ -884,9 +957,9 @@ def kraus_to_stinespring(
         kraus_op.append(gate)
         kraus_op = kraus_op.matrix(backend)
         kraus_op = backend.cast(kraus_op, dtype=kraus_op.dtype)
-        stinespring += np.kron(
+        stinespring = stinespring + backend.np.kron(
             kraus_op,
-            np.outer(vector_alpha, initial_state_env),
+            backend.np.outer(vector_alpha, initial_state_env),
         )
         del kraus_op, vector_alpha
 
@@ -921,7 +994,7 @@ def liouville_to_choi(super_op, order: str = "row", backend=None):
             respect to system-wise vectorization. Defaults to ``"row"``.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend
             to be used in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
         ndarray: Choi representation of quantum channel.
@@ -952,7 +1025,7 @@ def liouville_to_pauli(
             Pauli elements in the basis. Defaults to "IXYZ".
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend
             to be used in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
         ndarray: superoperator in the Pauli-Liouville representation.
@@ -979,7 +1052,7 @@ def liouville_to_pauli(
         backend=backend,
     )
 
-    return comp_to_pauli @ super_op @ np.conj(np.transpose(comp_to_pauli))
+    return comp_to_pauli @ super_op @ backend.np.conj(comp_to_pauli.T)
 
 
 def liouville_to_kraus(
@@ -1009,7 +1082,7 @@ def liouville_to_kraus(
             respect to system-wise vectorization. Defaults to ``"row"``.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend
             to be used in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
         (ndarray, ndarray): Kraus operators of quantum channel and their respective coefficients.
@@ -1052,7 +1125,7 @@ def liouville_to_chi(
             Pauli elements in the basis. Defaults to "IXYZ".
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend
             to be used in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
         ndarray: Chi-matrix representation of quantum channel.
@@ -1111,7 +1184,7 @@ def liouville_to_stinespring(
             environment in its ground state. Defaults to ``None``.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend
             to be used in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
         ndarray: Stinespring representation of quantum channel.
@@ -1153,12 +1226,12 @@ def pauli_to_liouville(
             Pauli elements. Defaults to "IXYZ".
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend
             to be used in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
         ndarray: superoperator in the Liouville representation.
     """
-    from qibo.quantum_info.basis import pauli_to_comp_basis
+    from qibo.quantum_info.basis import pauli_to_comp_basis  # pylint: disable=C0415
 
     backend = _check_backend(backend)
 
@@ -1180,7 +1253,7 @@ def pauli_to_liouville(
         backend=backend,
     )
 
-    return pauli_to_comp @ pauli_op @ np.conj(np.transpose(pauli_to_comp))
+    return pauli_to_comp @ pauli_op @ backend.np.conj(pauli_to_comp).T
 
 
 def pauli_to_choi(
@@ -1205,7 +1278,7 @@ def pauli_to_choi(
             Pauli elements. Defaults to "IXYZ".
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend
             to be used in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
         ndarray: Choi representation of the superoperator.
@@ -1247,7 +1320,7 @@ def pauli_to_kraus(
             ``qibo.config.PRECISION_TOL=1e-8``. Defaults to ``None``.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend
             to be used in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
         (ndarray, ndarray): Kraus operators and their coefficients.
@@ -1282,7 +1355,7 @@ def pauli_to_chi(
             Pauli elements. Defaults to "IXYZ".
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend
             to be used in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
         ndarray: Chi-matrix representation of the quantum channel.
@@ -1342,7 +1415,7 @@ def pauli_to_stinespring(
             environment in its ground state. Defaults to ``None``.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend
             to be used in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
         ndarray: Stinestring representation of quantum channel.
@@ -1397,7 +1470,7 @@ def chi_to_choi(
             Pauli elements. Defaults to "IXYZ".
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend
             to be used in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
         ndarray: Choi representation of quantum channel.
@@ -1442,7 +1515,7 @@ def chi_to_liouville(
             Pauli elements. Defaults to "IXYZ".
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend
             to be used in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
         ndarray: Liouville representation of quantum channel.
@@ -1488,7 +1561,7 @@ def chi_to_pauli(
             Pauli elements. Defaults to "IXYZ".
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend
             to be used in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
         ndarray: superoperator in the Pauli-Liouville representation.
@@ -1551,7 +1624,7 @@ def chi_to_kraus(
             Defaults to ``True``.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend
             to be used in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
         (ndarray, ndarray): Kraus operators and their coefficients.
@@ -1629,7 +1702,7 @@ def chi_to_stinespring(
             environment in its ground state. Defaults to ``None``.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend
             to be used in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
         ndarray: Stinespring representation of quantum channel.
@@ -1685,7 +1758,7 @@ def stinespring_to_choi(
             respect to system-wise vectorization. Defaults to ``"row"``.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend
             to be used in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
         ndarray: Choi representation of quantum channel.
@@ -1741,7 +1814,7 @@ def stinespring_to_liouville(
             respect to system-wise vectorization. Defaults to ``"row"``.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend
             to be used in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
         ndarray: Liouville representation of quantum channel.
@@ -1801,7 +1874,7 @@ def stinespring_to_pauli(
             single-qubit Pauli elements. Defaults to "IXYZ".
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend
             to be used in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
         ndarray: Pauli-Liouville representation of quantum channel.
@@ -1862,7 +1935,7 @@ def stinespring_to_kraus(
         nqubits (int, optional): number of qubits in the system. Defaults to ``None``.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend
             to be used in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
         ndarray: Kraus operators.
@@ -1908,15 +1981,15 @@ def stinespring_to_kraus(
             initial_state_env, dtype=initial_state_env.dtype
         )
 
-    stinespring = np.reshape(stinespring, (dim, dim_env, dim, dim_env))
-    stinespring = np.swapaxes(stinespring, 1, 2)
+    stinespring = backend.np.reshape(stinespring, (dim, dim_env, dim, dim_env))
+    stinespring = backend.np.swapaxes(stinespring, 1, 2)
 
     kraus_ops = []
     for alpha in range(dim_env):
         vector_alpha = np.zeros(dim_env, dtype=complex)
         vector_alpha[alpha] = 1.0
         vector_alpha = backend.cast(vector_alpha, dtype=vector_alpha.dtype)
-        kraus = np.conj(vector_alpha) @ stinespring @ initial_state_env
+        kraus = backend.np.conj(vector_alpha) @ stinespring @ initial_state_env
         kraus_ops.append(kraus)
 
     return kraus_ops
@@ -1959,7 +2032,7 @@ def stinespring_to_chi(
             Pauli elements in the Pauli basis. Defaults to "IXYZ".
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend
             to be used in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
         ndarray: :math:`\\chi`-representation of quantum channel.
@@ -2014,7 +2087,7 @@ def kraus_to_unitaries(
             defaults to ``1e-7``. Defaults to ``None``.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend
             to be used in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
         (ndarray, ndarray): Unitary operators and their associated probabilities.
@@ -2057,9 +2130,9 @@ def kraus_to_unitaries(
         operator = (1 - np.sum(x0)) * np.eye(dim**2, dtype=complex)
         operator = backend.cast(operator, dtype=operator.dtype)
         for prob, oper in zip(x0, operators):
-            operator += prob * oper
+            operator = operator + prob * oper
 
-        return float(backend.calculate_norm_density_matrix(target - operator, order=2))
+        return float(backend.calculate_matrix_norm(target - operator, order=2))
 
     # initial parameters as flat distribution
     x0 = [1.0 / (len(kraus_ops) + 1)] * len(kraus_ops)
@@ -2118,12 +2191,11 @@ def _reshuffling(super_op, order: str = "row", backend=None):
             respect to system-wise vectorization. Defaults to ``"row"``.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend
             to be used in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
         ndarray: Choi (Liouville) representation of the quantum channel.
     """
-
     if not isinstance(order, str):
         raise_error(TypeError, f"order must be type str, but it is type {type(order)}.")
 
@@ -2142,6 +2214,8 @@ def _reshuffling(super_op, order: str = "row", backend=None):
 
     backend = _check_backend(backend)
 
+    super_op = backend.cast(super_op, dtype=super_op.dtype)
+
     dim = np.sqrt(super_op.shape[0])
 
     if (
@@ -2152,13 +2226,12 @@ def _reshuffling(super_op, order: str = "row", backend=None):
         raise_error(ValueError, "super_op must be of shape (4^n, 4^n)")
 
     dim = int(dim)
-    super_op = np.reshape(super_op, [dim] * 4)
+    super_op = backend.np.reshape(super_op, [dim] * 4)
 
     axes = [1, 2] if order == "row" else [0, 3]
-    super_op = np.swapaxes(super_op, *axes)
+    super_op = backend.np.swapaxes(super_op, *axes)
 
-    super_op = np.reshape(super_op, [dim**2, dim**2])
-    super_op = backend.cast(super_op, dtype=super_op.dtype)
+    super_op = backend.np.reshape(super_op, [dim**2, dim**2])
 
     return super_op
 
@@ -2217,7 +2290,7 @@ def _individual_kraus_to_liouville(
         kraus_op.append(gate)
         kraus_op = kraus_op.matrix(backend)
         kraus_op = vectorization(kraus_op, order=order, backend=backend)
-        kraus_op = np.outer(kraus_op, np.conj(kraus_op))
+        kraus_op = backend.np.outer(kraus_op, backend.np.conj(kraus_op))
         super_ops.append(choi_to_liouville(kraus_op, order=order, backend=backend))
 
     return super_ops
