@@ -4,7 +4,7 @@ from functools import reduce
 from itertools import permutations
 from math import factorial
 from re import finditer
-from typing import Optional, Union
+from typing import List, Optional, Tuple, Union
 
 import numpy as np
 
@@ -27,7 +27,7 @@ def hamming_weight(
             non-zero elements. Defaults to ``False``.
 
     Returns:
-        (int or list): Hamming weight of bitstring or list of indexes of non-zero elements.
+        int or list: Hamming weight of bitstring or list of indexes of non-zero elements.
     """
     if not isinstance(return_indexes, bool):
         raise_error(
@@ -35,17 +35,11 @@ def hamming_weight(
             f"return_indexes must be type bool, but it is type {type(return_indexes)}",
         )
 
-    if not isinstance(bitstring, (int, str, list, tuple, np.ndarray)):
-        raise_error(
-            TypeError,
-            "bitstring must be either type int, list, tuple, or numpy.ndarray. "
-            f"However, it is type {type(bitstring)}.",
-        )
-
-    if isinstance(bitstring, int):
-        bitstring = f"{bitstring:b}"
-    elif isinstance(bitstring, (list, tuple, np.ndarray)):
-        bitstring = "".join([str(bit) for bit in bitstring])
+    bitstring = (
+        "".join([str(bit) for bit in list(bitstring)])
+        if not isinstance(bitstring, int)
+        else f"{bitstring:b}"
+    )
 
     indexes = [item.start() for item in finditer("1", bitstring)]
 
@@ -80,25 +74,17 @@ def hamming_distance(
             f"return_indexes must be type bool, but it is type {type(return_indexes)}",
         )
 
-    if not isinstance(bitstring_1, (int, str, list, tuple)):
-        raise_error(
-            TypeError,
-            "bitstring_1 must be either type int, list, tuple, or numpy.ndarray. "
-            f"However, it is type {type(bitstring_1)}.",
-        )
+    if isinstance(bitstring_1, int):
+        bitstring_1 = f"{bitstring_1:b}"
 
-    if not isinstance(bitstring_2, (int, str, list, tuple)):
-        raise_error(
-            TypeError,
-            "bitstring_2 must be either type int, list, tuple, or numpy.ndarray. "
-            f"However, it is type {type(bitstring_2)}.",
-        )
+    if isinstance(bitstring_2, int):
+        bitstring_2 = f"{bitstring_2:b}"
 
-    if isinstance(bitstring_1, (list, tuple)):
-        bitstring_1 = "".join(bitstring_1)
+    if not isinstance(bitstring_1, str):
+        bitstring_1 = "".join([str(bit) for bit in bitstring_1])
 
-    if isinstance(bitstring_2, (list, tuple)):
-        bitstring_2 = "".join(bitstring_2)
+    if not isinstance(bitstring_2, str):
+        bitstring_2 = "".join([str(bit) for bit in bitstring_2])
 
     nbits = max(len(bitstring_1), len(bitstring_2))
 
@@ -136,7 +122,7 @@ def hadamard_transform(array, implementation: str = "fast", backend=None):
             for matrices. If ``"fast"``, computational complexity is
             :math:`\\mathcal{O}(n \\, 2^{n})` in both cases.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
-            in the execution. If ``None``, it uses :class:`qibo.backends.GlobalBackend`.
+            in the execution. If ``None``, it uses the current backend.
             Defaults to ``None``.
 
     Returns:
@@ -185,11 +171,10 @@ def hadamard_transform(array, implementation: str = "fast", backend=None):
 
         return array
 
-    array = _hadamard_transform_1d(array)
+    array = _hadamard_transform_1d(array, backend=backend)
 
     if len(array.shape) == 2:
-        array = _hadamard_transform_1d(np.transpose(array))
-        array = np.transpose(array)
+        array = _hadamard_transform_1d(array.T, backend=backend).T
 
     # needed for the tensorflow backend
     array = backend.cast(array, dtype=array.dtype)
@@ -215,7 +200,7 @@ def hellinger_distance(prob_dist_p, prob_dist_q, validate: bool = False, backend
             probability distributions. Defaults to ``False``.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be
             used in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
         (float): Hellinger distance :math:`H(p, q)`.
@@ -252,7 +237,10 @@ def hellinger_distance(prob_dist_p, prob_dist_q, validate: bool = False, backend
             raise_error(ValueError, "Second probability array must sum to 1.")
 
     distance = float(
-        backend.calculate_norm(np.sqrt(prob_dist_p) - np.sqrt(prob_dist_q)) / np.sqrt(2)
+        backend.calculate_vector_norm(
+            backend.np.sqrt(prob_dist_p) - backend.np.sqrt(prob_dist_q)
+        )
+        / np.sqrt(2)
     )
 
     return distance
@@ -275,10 +263,10 @@ def hellinger_fidelity(prob_dist_p, prob_dist_q, validate: bool = False, backend
             probability distributions. Defaults to ``False``.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be
             used in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
-        (float): Hellinger fidelity.
+        float: Hellinger fidelity.
 
     """
     backend = _check_backend(backend)
@@ -312,10 +300,10 @@ def hellinger_shot_error(
             probability distributions. Defaults to ``False``.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be
             used in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
-        (float): Hellinger fidelity error.
+        float: Hellinger fidelity error.
 
     """
     backend = _check_backend(backend)
@@ -337,6 +325,58 @@ def hellinger_shot_error(
     return hellinger_error
 
 
+def total_variation_distance(
+    prob_dist_p, prob_dist_q, validate: bool = False, backend=None
+):
+    """Calculate the total variation distance between two discrete probability distributions.
+
+    For probabilities :math:`p` and :math:`q`, the total variation distance is defined as
+
+    .. math::
+        \\operatorname{TVD}(p, \\, q) = \\frac{1}{2} \\, \\|p - q\\|_{1}
+            = \\frac{1}{2} \\, \\sum_{x} \\, \\left|p(x) - q(x)\\right| \\, ,
+
+    where :math:`\\|\\cdot\\|_{1}` detones the :math:`\\ell_{1}`-norm.
+
+    Args:
+        prob_dist_p (ndarray or list): discrete probability distribution :math:`p`.
+        prob_dist_q (ndarray or list): discrete probability distribution :math:`q`.
+        validate (bool, optional): if ``True``, checks if :math:`p` and :math:`q` are proper
+            probability distributions. Defaults to ``False``.
+        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be
+            used in the execution. If ``None``, it uses
+            the current backend. Defaults to ``None``.
+
+    Returns:
+        float: Total variation distance measure.
+    """
+    backend = _check_backend(backend)
+
+    if isinstance(prob_dist_p, list):
+        prob_dist_p = backend.cast(prob_dist_p, dtype=np.float64)
+
+    if isinstance(prob_dist_q, list):
+        prob_dist_q = backend.cast(prob_dist_q, dtype=np.float64)
+
+    if validate:
+        if (any(prob_dist_p < 0) or any(prob_dist_p > 1.0)) or (
+            any(prob_dist_q < 0) or any(prob_dist_q > 1.0)
+        ):
+            raise_error(
+                ValueError,
+                "All elements of the probability array must be between 0. and 1..",
+            )
+        if backend.np.abs(backend.np.sum(prob_dist_p) - 1.0) > PRECISION_TOL:
+            raise_error(ValueError, "First probability array must sum to 1.")
+
+        if backend.np.abs(backend.np.sum(prob_dist_q) - 1.0) > PRECISION_TOL:
+            raise_error(ValueError, "Second probability array must sum to 1.")
+
+    tvd = backend.calculate_vector_norm(prob_dist_p - prob_dist_q, order=1)
+
+    return tvd / 2
+
+
 def haar_integral(
     nqubits: int,
     power_t: int,
@@ -357,7 +397,7 @@ def haar_integral(
             Defaults to ``None``.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be
             used in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
         array: Estimation of the Haar integral.
@@ -396,11 +436,15 @@ def haar_integral(
             rand_unit_density, dtype=rand_unit_density.dtype
         )
         for _ in range(samples):
-            haar_state = np.reshape(random_statevector(dim, backend=backend), (-1, 1))
+            haar_state = backend.np.reshape(
+                random_statevector(dim, backend=backend), (-1, 1)
+            )
 
-            rho = haar_state @ np.conj(np.transpose(haar_state))
+            rho = haar_state @ backend.np.conj(haar_state).T
 
-            rand_unit_density += reduce(np.kron, [rho] * power_t)
+            rand_unit_density = rand_unit_density + reduce(
+                backend.np.kron, [rho] * power_t
+            )
 
         integral = rand_unit_density / samples
 
@@ -414,14 +458,16 @@ def haar_integral(
     ]
 
     identity = np.eye(dim**power_t, dtype=float)
-    identity = backend.cast(identity, dtype=identity.dtype)
     identity = np.reshape(identity, (dim,) * (2 * power_t))
+    identity = backend.cast(identity, dtype=identity.dtype)
 
     integral = np.zeros((dim**power_t, dim**power_t), dtype=float)
     integral = backend.cast(integral, dtype=integral.dtype)
     for indices in permutations_list:
-        integral += np.reshape(np.transpose(identity, indices), (-1, dim**power_t))
-    integral *= normalization
+        integral = integral + backend.np.reshape(
+            backend.np.transpose(identity, indices), (-1, dim**power_t)
+        )
+    integral = integral * normalization
 
     return integral
 
@@ -440,7 +486,7 @@ def pqc_integral(circuit, power_t: int, samples: int, backend=None):
         samples (int): number of samples to estimate the integral.
         backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be
             used in the execution. If ``None``, it uses
-            :class:`qibo.backends.GlobalBackend`. Defaults to ``None``.
+            the current backend. Defaults to ``None``.
 
     Returns:
         ndarray: Estimation of the integral.
@@ -469,27 +515,170 @@ def pqc_integral(circuit, power_t: int, samples: int, backend=None):
 
         rho = backend.execute_circuit(circuit).state()
 
-        rand_unit_density += reduce(np.kron, [rho] * power_t)
+        rand_unit_density = rand_unit_density + reduce(np.kron, [rho] * power_t)
 
     integral = rand_unit_density / samples
 
     return integral
 
 
-def _hadamard_transform_1d(array):
+def _hadamard_transform_1d(array, backend=None):
     # necessary because of tf.EagerTensor
     # does not accept item assignment
-    array_copied = np.copy(array)
+    backend = _check_backend(backend)
+    array_copied = backend.np.copy(array)
 
     indexes = [2**k for k in range(int(np.log2(len(array_copied))))]
     for index in indexes:
         for k in range(0, len(array_copied), 2 * index):
             for j in range(k, k + index):
                 # copy necessary because of cupy backend
-                elem_1 = np.copy(array_copied[j])
-                elem_2 = np.copy(array_copied[j + index])
+                elem_1 = backend.np.copy(array_copied[j])
+                elem_2 = backend.np.copy(array_copied[j + index])
                 array_copied[j] = elem_1 + elem_2
                 array_copied[j + index] = elem_1 - elem_2
         array_copied /= 2.0
 
     return array_copied
+
+
+def _cycles_from_perm(sigma: List[int]):
+    """Extract the cycles from a permutation as follows:
+    - Treat the permutation as a directed graph of arrows i->sigma(i).
+    - Depth‑first walk from every unvisited vertex; each walk closes at the start -> a cycle.
+    - Disjoint cycles partition the set {0,...,n-1} and commute, so we can factorize them independently.
+
+    Args:
+        sigma (list[int] or tuple[int]): permutation description on {0,...,n-1}.
+
+    Returns:
+        list[list[tuple[int, int]]]: :math:`t` list of disjoint cycles.
+    """
+    n, seen, cycles = len(sigma), [False] * len(sigma), []
+    for i in range(n):
+        # Depth‑first walk from every unvisited vertex
+        if seen[i]:
+            continue
+        cur, cyc = i, []
+        # Disjoint cycles
+        while not seen[cur]:
+            seen[cur] = True
+            cyc.append(cur)
+            cur = sigma[cur]
+        if len(cyc) > 1:
+            cycles.append(cyc)
+    return cycles
+
+
+def _star_matchings(cyc: list[int]):
+    """
+    Given a cycle :math:`(a_{1}, \\, a_{2}, \\, \\cdots, \\, a_{k})` with :math:`k \\geq 2`,
+    the star factorization expresses the cycle as the ordered product of :math:`(k-1)`
+    disjoint transpositions that all share the first vertex:
+
+    .. math::
+        (a_{1}, \\, a_{2}), \\, (a_{1}, \\, a_{3}), \\, \\cdots, \\, (a_{1}, \\, a_{k}) \\, .
+
+    Applied right‑to‑left, this product reproduces the original cycle.
+
+    Args:
+        cyc (list[list[tuple[int, int]]]): :math:`t` list of disjoint cycles.
+
+    Returns:
+        list[list[tuple[int, int]]]: :math:`t` list of pairwise transpositions.
+    """
+    hub = cyc[0]
+    return [[(min(hub, v), max(hub, v))] for v in cyc[1:]]
+
+
+def _greedy_pack(matchings: List[List[Tuple[int, int]]], m: int):
+    """
+    Add a matching to the current layer if
+        - it shares no vertex with swaps already in the layer, and
+        - new layer size #swaps stays a power of two <= m.
+    Otherwise flush the layer and start a new one.
+    It works since disjointness keeps swaps commutative, and the power‑of‑two size rule
+    aligns exactly with layer constraints.
+
+    Args:
+        matchings (list[list[tuple[int, int]]]): :math:`t` list of pairwise transpositions.
+
+    Returns:
+        list[list[tuple[int, int]]]: :math:`t` layers of pairwise transpositions.
+    """
+    layers: list[list[tuple[int, int]]] = []
+    cur: list[tuple[int, int]] = []
+    used = set()
+
+    def _flush():
+        nonlocal cur, used
+        if cur:
+            layers.append(cur)
+        cur, used = [], set()
+
+    def _verts(layer: list[tuple[int, int]]):
+        for a, b in layer:
+            yield a
+            yield b
+
+    for M in matchings:
+        x = len(cur) + len(M)
+        # is power of 2 and x<=m
+        legal = (x > 0 and (x & (x - 1)) == 0) and x <= m
+        # shares no vertex with swaps already in the layer
+        if cur and (any(v in used for v in _verts(M)) or not legal):  # pragma: no cover
+            _flush()
+        cur.extend(M)
+        used.update(_verts(M))
+        if len(cur) == m:
+            _flush()
+    _flush()
+    return layers
+
+
+def decompose_permutation(
+    sigma: Union[List[int], Tuple[int, ...]], m: int, backend=None
+):
+    """
+     Given permutation ``sigma`` on :math:`\\{0, \\, 1, \\, \\dots, \\, d-1\\}`
+    and a power‑of‑two budget ``m``, this function factors ``sigma``
+    into the fewest layers :math:`\\sigma_{1}, \\, \\sigma_{2}, \\, \\cdots, \\, \\sigma_{t}` such that:
+        - each layer has at most :math:`m` disjoint transpositions
+        - each layer moves a power‑of‑two number of indices.
+
+    We do this as follows:
+        1) Cycle extraction – split sigma into disjoint cycles.
+        2) Star factorisation – a k‑cycle becomes (k-1) hub–spoke swaps.
+        3) Greedy packing – merge swaps into layers while keeping rules.
+
+    Args:
+        sigma (list[int] or tuple[int]): permutation description on :math:`\\{0, \\, 1, \\, \\dots, \\, d-1\\}`.
+        m (int): power‑of‑two budget.
+        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
+            in the execution. If ``None``, it uses the current backend. Defaults to ``None``.
+
+    Returns:
+        list[list[tuple[int, int]]]: :math:`t` layers of pairwise transpositions.
+
+    """
+    backend = _check_backend(backend)
+
+    if isinstance(sigma, tuple):
+        sigma = list(sigma)
+
+    if not isinstance(sigma, (list, tuple)):
+        raise_error(
+            TypeError, f"Permutation sigma must be ``list`` or ``tuple`` of ``int``s."
+        )
+
+    if sum([abs(s - i) for s, i in zip(sorted(sigma), range(len(sigma)))]) != 0:
+        raise_error(
+            ValueError, "Permutation sigma must contain all indices {0,...,n-1}"
+        )
+
+    if m > 0 and (m & (m - 1)) != 0:
+        raise_error(ValueError, f"budget m must be a power‑of‑two")
+
+    matchings = [l for cyc in _cycles_from_perm(sigma) for l in _star_matchings(cyc)]
+
+    return _greedy_pack(matchings, m)
