@@ -27,7 +27,7 @@ class QuantumState:
 
     Args:
         state (np.ndarray): Input quantum state as np.ndarray.
-        backend (qibo.backends.AbstractBackend): Backend used for the calculations. If not provided the :class:`qibo.backends.GlobalBackend` is going to be used.
+        backend (qibo.backends.AbstractBackend): Backend used for the calculations. If not provided, the current backend is going to be used.
     """
 
     def __init__(self, state, backend=None):
@@ -65,6 +65,10 @@ class QuantumState:
     def state(self, numpy: bool = False):
         """State's tensor representation as a backend tensor.
 
+        .. note::
+            If the state has Hamming weight :math:`k` and is computed using the ``HammingWeightBackend``, its dimension
+            is :math:`d = \\binom{n}{k}`, where :math:`n` is the number of qubits.
+
         Args:
             numpy (bool, optional): If ``True`` the returned tensor will be a ``numpy`` array,
                 otherwise it will follow the backend tensor type.
@@ -158,7 +162,7 @@ class MeasurementOutcomes:
     Args:
         measurements (:class:`qibo.gates.M`): Measurement gates.
         backend (:class:`qibo.backends.AbstractBackend`): Backend used for the calculations.
-            If ``None``, then :class:`qibo.backends.GlobalBackend` is used. Defaults to ``None``.
+            If ``None``, then the current backend is used. Defaults to ``None``.
         probabilities (np.ndarray): Use these probabilities to generate samples and frequencies.
         samples (np.darray): Use these samples to generate probabilities and frequencies.
         nshots (int): Number of shots used for samples, probabilities and frequencies generation.
@@ -186,33 +190,32 @@ class MeasurementOutcomes:
             for m in measurements:
                 indices = [self.measurement_gate.qubits.index(q) for q in m.qubits]
                 m.result.register_samples(samples[:, indices])
-        else:
-            for gate in self.measurements:
-                gate.result.reset()
 
     def frequencies(self, binary: bool = True, registers: bool = False):
         """Returns the frequencies of measured samples.
 
         Args:
-            binary (bool, optional): Return frequency keys in binary or decimal form.
+            binary (bool, optional): If ``True``, returns frequency keys in binary form.
+                If ``False``, returns them in decimal form. Defaults to ``True``.
             registers (bool, optional): Group frequencies according to registers.
+                Defaults to ``False``.
 
         Returns:
-            A `collections.Counter` where the keys are the observed values
+            A :class:`collections.Counter` where the keys are the observed values
             and the values the corresponding frequencies, that is the number
             of times each measured value/bitstring appears.
 
             If ``binary`` is ``True``
-                the keys of the `Counter` are in binary form, as strings of
-                :math:`0`s and :math`1`s.
+                the keys of the :class:`collections.Counter` are in binary form,
+                as strings of :math:`0` and :math`1`.
             If ``binary`` is ``False``
-                the keys of the ``Counter`` are integers.
+                the keys of the :class:`collections.Counter` are integers.
             If ``registers`` is ``True``
-                a `dict` of `Counter` s is returned where keys are the name of
-                each register.
+                a `dict` of :class:`collections.Counter` is returned where keys are
+                the name of each register.
             If ``registers`` is ``False``
-                a single ``Counter`` is returned which contains samples from all
-                the measured qubits, independently of their registers.
+                a single :class:`collections.Counter` is returned which contains samples
+                from all the measured qubits, independently of their registers.
         """
         qubits = self.measurement_gate.qubits
 
@@ -247,7 +250,7 @@ class MeasurementOutcomes:
                             if int(bitstring[qubit_map.get(q)]):
                                 idx += 2 ** (len(rqubits) - i - 1)
                         rfreqs[idx] += freq
-                    gate.result.register_frequencies(rfreqs, self.backend)
+                    gate.result.register_frequencies(rfreqs)
             else:
                 self._frequencies = self.backend.calculate_frequencies(
                     self.samples(binary=False)
@@ -273,25 +276,27 @@ class MeasurementOutcomes:
         nqubits = len(self.measurement_gate.qubits)
         if qubits is None:
             qubits = range(nqubits)
-        else:
-            if not set(qubits).issubset(self.measurement_gate.qubits):
-                raise_error(
-                    RuntimeError,
-                    f"Asking probabilities for qubits {qubits}, but only qubits {self.measurement_gate.qubits} were measured.",
-                )
+        elif set(qubits).issubset(self.measurement_gate.qubits):
             qubits = [self.measurement_gate.qubits.index(q) for q in qubits]
+        else:
+            raise_error(
+                RuntimeError,
+                f"Asking probabilities for qubits {qubits}, but only qubits {self.measurement_gate.qubits} were measured.",
+            )
 
         if self._probs is not None and not self.measurement_gate.has_bitflip_noise():
             return self.backend.calculate_probabilities(
                 np.sqrt(self._probs), qubits, nqubits
             )
 
-        probs = [0 for _ in range(2**nqubits)]
+        probs = [0] * 2**nqubits
         for state, freq in self.frequencies(binary=False).items():
             probs[state] = freq / self.nshots
         probs = self.backend.cast(probs)
         self._probs = probs
-        return self.backend.calculate_probabilities(np.sqrt(probs), qubits, nqubits)
+        return self.backend.calculate_probabilities(
+            self.backend.np.sqrt(probs), qubits, nqubits
+        )
 
     def has_samples(self):
         """Check whether the samples are available already.
@@ -326,7 +331,8 @@ class MeasurementOutcomes:
         if self._samples is None:
             if self.measurements[0].result.has_samples():
                 self._samples = self.backend.np.concatenate(
-                    [gate.result.samples() for gate in self.measurements], axis=1
+                    [gate.result.samples() for gate in self.measurements],
+                    axis=1,
                 )
             else:
                 if self._frequencies is not None:
@@ -336,16 +342,20 @@ class MeasurementOutcomes:
                         [np.repeat(x, f) for x, f in frequencies.items()]
                     )
                     np.random.shuffle(samples)
+                    samples = self.backend.cast(samples, dtype=self.backend.np.int64)
                 else:
                     # generate new samples
                     samples = self.backend.sample_shots(self._probs, self.nshots)
                 samples = self.backend.samples_to_binary(samples, len(qubits))
                 if self.measurement_gate.has_bitflip_noise():
                     p0, p1 = self.measurement_gate.bitflip_map
-                    bitflip_probabilities = [
-                        [p0.get(q) for q in qubits],
-                        [p1.get(q) for q in qubits],
-                    ]
+                    bitflip_probabilities = self.backend.cast(
+                        [
+                            [p0.get(q) for q in qubits],
+                            [p1.get(q) for q in qubits],
+                        ],
+                        dtype=self.backend.np.float64,
+                    )
                     samples = self.backend.apply_bitflips(
                         samples, bitflip_probabilities
                     )
@@ -353,12 +363,10 @@ class MeasurementOutcomes:
                 qubit_map = {
                     q: i for i, q in enumerate(self.measurement_gate.target_qubits)
                 }
-                self._samples = self.backend.cast(samples, "int32")
+                self._samples = samples
                 for gate in self.measurements:
                     rqubits = tuple(qubit_map.get(q) for q in gate.target_qubits)
-                    gate.result.register_samples(
-                        self._samples[:, rqubits], self.backend
-                    )
+                    gate.result.register_samples(self._samples[:, rqubits])
 
         if registers:
             return {
@@ -479,7 +487,7 @@ class CircuitResult(QuantumState, MeasurementOutcomes):
     Args:
         final_state (np.ndarray): Input quantum state as np.ndarray.
         measurements (qibo.gates.M): The measurement gates containing the measurements.
-        backend (qibo.backends.AbstractBackend): Backend used for the calculations. If not provided the :class:`qibo.backends.GlobalBackend` is going to be used.
+        backend (qibo.backends.AbstractBackend): Backend used for the calculations. If not provided, then the current backend is going to be used.
         probabilities (np.ndarray): Use these probabilities to generate samples and frequencies.
         samples (np.darray): Use these samples to generate probabilities and frequencies.
         nshots (int): Number of shots used for samples, probabilities and frequencies generation.
