@@ -1,12 +1,10 @@
 from qibo.transpiler.optimizer import Preprocessing, Rearrange
 from qibo.transpiler.pipeline import Passes
 from qibo.transpiler.placer import (
-    Custom,
     Random,
     ReverseTraversal,
     StarConnectivityPlacer,
     Subgraph,
-    Trivial,
 )
 from qibo.transpiler.router import Sabre, ShortestPaths, StarConnectivityRouter
-from qibo.transpiler.unroller import NativeGates
+from qibo.transpiler.unroller import NativeGates, Unroller
